/*

DISCLAIMER 

InSitu is provided as is, without any kind of guarantees or warranty. There is 
no documentation. The author does not provide any kind of support to users of 
this source code and is not responsible for any kind of problems or damages 
resulting from its use. 

The source code is the result of a long iterative design process for a research 
prototype and was never intended to be used in any kind of product. In general, 
the code cannot be expected to compile in any straightforward manner on your 
computer system, since the robot on which it was developed, has a custom design. 

*/
/* Rene Schaad                                                    */
/* Departement of Computer Science, Univ. of Zurich, Switzerland  */
/* schaad@ifi.unizh.ch                                            */


#include "APsup.h"
#include <math.h>

void FloorRay::Set(float x_, float y_, float at_)
{
	x	=	x_;				// store initial values
	y	=	y_;
	at	=	at_;
	xm	=	x + cos(at);	// compute another point along the ray dist==1
	ym	=	y + sin(at);
	a	=	y - ym;			// compute line parameters
	b	=	xm - x;
	c	=	-( a*x + b*y);
}

void FloorLine::Set(float x1_, float y1_, float x2_, float y2_)
{
	x	=	x1_;			// store initial values
	y	=	y1_;
	at	=	atan2(y2_-y1_,x2_-x1_);
	xm	=	x2_;			// store second point
	ym	=	y2_;
	a	=	y - ym;			// compute line parameters
	b	=	xm - x;
	c	=	-( a*x + b*y);
}

////////////////////////////////////////////////////////////////////////
//
//    Go Along a Ray 
//

Data	*GoAlongRay::Step(void)
{
	// set the parameters for computing the speeds
	//
	// constraint: PI2*kd + rd*kd == 0 (to avoid spinning of robot)
	const float	rd	=	5.0;	// reach of influence of distance (5m)
	const float	kh	=	-0.4;	// prop. constant for delta heading 
	const float	kd	=	PI2/rd; // prop. constant for distance
	const float	mr	=	0.4;	// max. rotational speed

	if (!robot) return NULL;
	FloorRay *the_ray;
	GETVAL(FloorRay,the_ray,floorray);	// get the ray, return if not possible
	if (the_ray==NULL) {
		robot->SetRobotVelocity(0.0,0.0);
		done.Set(FUZZTRUE);
		return NULL;
	} else {
		float rx,ry,ra,x,y,a;
		the_ray->Get(rx,ry,ra);						// get ray data	
		robot->GetDRPosition(x,y,a);				// get robot data
		float dd = the_ray->Dist(x,y);				// distance to ray
		float dh = ra - a - Limit(kd*dd,-PI2,PI2);	// diff. heading to ray
		while (dh<-PI) dh+=2*PI;					// keep angle in bounds
		while (dh>PI)  dh-=2*PI;
		float vr = Limit(kh*dh,-mr,mr);				// rotational velocity
		float vt = Limit((1.0-fabs(vr)/mr)*mt,0.0,mt); // translational velocity
		robot->SetRobotVelocity(vt,vr);				// set it!
	}
	
	return &done;
}

void	GoAlongRay::Reset(void)
{
	done.Set(FUZZFALSE);
	if (floorray) floorray->Reset();
}

void	GoAlongRay::Abort(void)
{
	done.Set(FUZZTRUE);
	if (floorray) floorray->Abort();
}


////////////////////////////////////////////////////////////////////////
//
//    Extend a Ray 
//


Data	*ExtendRay::Step(void)
{
	if (!robot) return NULL;
	
	// if not yet locked, do so now
	if (done.False()) {
		Number *the_datt;
		GETVAL(Number,the_datt,dattitude);	// get the ray, return if not possible

		if (the_datt==NULL) {
			done.Set(FUZZTRUE);
			the_ray.SetLock(FALSE);
		} else {
			float x,y,a;
			robot->GetDRPosition(x,y,a);				// get robot data
			the_ray.Set(x,y,AngleMod(a+the_datt->Get()));// set ray data	
			the_ray.SetLock(TRUE);
		}
	}
	
	// return the (now locked) ray
	return &the_ray;
}

void	ExtendRay::Reset(void)
{
	the_ray.SetLock(FALSE);
	done.Set(FUZZFALSE);
	if (dattitude) dattitude->Reset();
}

void	ExtendRay::Abort(void)
{
	done.Set(FUZZTRUE);
	if (dattitude) dattitude->Abort();
}



////////////////////////////////////////////////////////////////////////
//
//    Waypoint to FloorPoint
//


Data *WaypointToFloorPoint::Step(void)
{
	if (! waypoint) return NULL;
	
	Waypoint *wp;
	GETVAL(Waypoint,wp,waypoint);
	if (wp) {
		float x,y,a;
		boolean ds;
		wp->Get(x,y,a,ds);
		the_fpoint.Set(x,y);
		return &the_fpoint;
	} else {
		return NULL;
	}
}

