/*

DISCLAIMER 

InSitu is provided as is, without any kind of guarantees or warranty. There is 
no documentation. The author does not provide any kind of support to users of 
this source code and is not responsible for any kind of problems or damages 
resulting from its use. 

The source code is the result of a long iterative design process for a research 
prototype and was never intended to be used in any kind of product. In general, 
the code cannot be expected to compile in any straightforward manner on your 
computer system, since the robot on which it was developed, has a custom design. 

*/

#ifndef _APsup_h_
#define _APsup_h_

/* Rene Schaad                                                    */
/* Departement of Computer Science, Univ. of Zurich, Switzerland  */
/* schaad@ifi.unizh.ch                                            */

#include "Monsup.h"
#include "Navsup.h"
#include "PrgStruct.h"


////////////////////////////////////////////////////////////////////
//
//  Intermediate Objects (such as markers, etc.)
//

class FloorMarker : public Marker {
public:
	FloorMarker(float x_, float y_)	
		:x(x_),y(y_),Marker(DT_MARKER_MIN){}; // dummy DataType
protected:
	float	x,y; // cartesian coordinates of location (with attitude)
};

class FloorPoint : public FloorMarker {
public:
	FloorPoint(float x_, float y_)	:FloorMarker(x_,y_){};
	
	// manipulation
	void	Set(float x_, float y_)		{x=x_;y=y_;};
	void	Set(FloorPoint *newwp)		{if (newwp) {x=newwp->x;y=newwp->y;};};
	void	Get(float &x_, float &y_)	{x_=x;y_=y;};
				
	virtual boolean	SetData(Data *org_){
		if (!org_) return FALSE;
		if (org_->type!=type) return FALSE;
		Set((FloorPoint*)org_);
		return TRUE;
	};

	virtual Data	*Duplicate(void){return new FloorPoint(x,y);};
	virtual Data	*Print(void){printf("FloorPoint(%f,%f)",x,y);return this;};
};

class FloorRay : public FloorMarker {
public:
	FloorRay(float x_, float y_, float at_)	
		:at(at_),FloorMarker(x_,y_)
		{type=DT_FloorRay;Set(x,y,at);};
		
	// ray specific stuff
	void	GetEqu(float &a_, float &b_, float &c_)
												{a_=a;b_=b;c_=c;}
	float	Dist(float x_, float y_)			{return (a*x_ + b*y_ + c);}
	
	// manipulation
	void	Set(float x_, float y_, float at_);
	void	Set(FloorRay *newwp)	{if (newwp) {Set(newwp->x,newwp->y,newwp->at);};};
	void	Get(float &x_, float &y_, float &at_){x_=x;y_=y;at_=at;};

				
	virtual boolean	SetData(Data *org_){
		if (!org_) return FALSE;
		if (org_->type!=type) return FALSE;
		Set((FloorRay*)org_);
		return TRUE;
	};

	virtual Data	*Duplicate(void){return new FloorRay(x,y,at);};
	virtual Data	*Print(void){printf("FloorRay(%f,%f,%f)",x,y,at);return this;};
protected:
	float	at;	// x,y are cartesian coordinates
					// a is the angle of the ray going out from point x/y
	float	xm,ym;	// some assumed point along the ray
	float	a,b,c;	// line parameters of the ray
};

// A line is not a FloorRect, even though most of the impl. could've been reused

class FloorLine : public FloorRay {
public:
	FloorLine(float x1_,float y1_,float x2_,float y2_)	
		:FloorRay(x1_,y1_,0.0)
		{type=DT_FloorLine;Set(x1_,y1_,x2_,y2_);};
	
	// manipulation
	void	Set(float x1_, float y1_, float x2_, float y2_);
	void	Set(FloorLine *newwp)	{if (newwp) Set(newwp->x,newwp->y,newwp->xm,newwp->ym);};
	void	Get(float &x1_, float &y1_, float &x2_, float &y2_){x1_=x;y1_=y;x2_=xm;y2_=ym;};
				
	virtual boolean	SetData(Data *org_){
		if (!org_) return FALSE;
		if (org_->type!=type) return FALSE;
		Set((FloorLine*)org_);
		return TRUE;
	};

	virtual Data	*Duplicate(void){return new FloorLine(x,y,xm,ym);};
	virtual Data	*Print(void){printf("FloorLine(%f,%f,%f,%f)",x,y,xm,ym);return this;};
};


class FloorCircle : public FloorMarker {
public:
	FloorCircle(float x_, float y_, float r_)	
		:r(r_),FloorMarker(x_,y_){type=DT_FloorCircle;};
	
	// manipulation
	void	Set(float x_, float y_, float r_)	{x=x_;y=y_;r=r_;};
	void	Set(FloorCircle *newwp)	{if (newwp) {x=newwp->x;y=newwp->y;r=newwp->r;};};
	void	Get(float &x_, float &y_, float &r_){x_=x;y_=y;r_=r;};
				
	virtual boolean	SetData(Data *org_){
		if (!org_) return FALSE;
		if (org_->type!=type) return FALSE;
		Set((FloorCircle*)org_);
		return TRUE;
	};

	virtual Data	*Duplicate(void){return new FloorCircle(x,y,r);};
	virtual Data	*Print(void){printf("FloorCircle(%f,%f,%f)",x,y,r);return this;};
private:
	float	r;	// x,y are cartesian coordinates
					// a is the angle of the ray going out from point x/y
};

class FloorSector : public FloorMarker {
public:
	FloorSector(float x_,float y_,float a1_,float a2_)	
		:a1(a1_),a2(a2_),FloorMarker(x_,y_){type=DT_FloorSector;};
	
	// manipulation
	void	Set(float x_, float y_, float a1_, float a2_)	{x=x_;y=y_;a1=a1_;a2=a2_;};
	void	Set(FloorSector *newwp)	{if (newwp) {x=newwp->x;y=newwp->y;a1=newwp->a1;a2=newwp->a2;};};
	void	Get(float &x_, float &y_, float &a1_, float &a2_){x_=x;y_=y;a1_=a1;a2_=a2;};
				
	virtual boolean	SetData(Data *org_){
		if (!org_) return FALSE;
		if (org_->type!=type) return FALSE;
		Set((FloorSector*)org_);
		return TRUE;
	};

	virtual Data	*Duplicate(void){return new FloorSector(x,y,a1,a2);};
	virtual Data	*Print(void){printf("FloorSector(%f,%f,%f,%f)",x,y,a1,a2);return this;};
private:
	float	a1,a2;	// x,y are cartesian coordinates
						// a1,a2 are the bounding rays' angles
};

class FloorRect : public FloorMarker {
public:
	FloorRect(float x_,float y_,float x2_,float y2_)	
		:x2(x2_),y2(y2_),FloorMarker(x_,y_){type=DT_FloorRect;};
	
	// manipulation
	void	Set(float x_, float y_, float x2_, float y2_)	{x=x_;y=y_;x2=x2_;y2=y2_;};
	void	Set(FloorRect *newwp)	{if (newwp) {x=newwp->x;y=newwp->y;x2=newwp->x2;y2=newwp->y2;};};
	void	Get(float &x_, float &y_, float &x2_, float &y2_){x_=x;y_=y;x2_=x2;y2_=y2;};
				
	virtual boolean	SetData(Data *org_){
		if (!org_) return FALSE;
		if (org_->type!=type) return FALSE;
		Set((FloorRect*)org_);
		return TRUE;
	};

	virtual Data	*Duplicate(void){return new FloorRect(x,y,x2,y2);};
	virtual Data	*Print(void){printf("FloorRect(%f,%f,%f,%f)",x,y,x2,y2);return this;};
private:
	float	x2,y2;	// x,y are cartesian coordinates
};


////////////////////////////////////////////////////////////////////
//
//  Perceptual Routines: Processing Perceptual Information by Acting
//

class GoAlongRay : public Action {
public:
	GoAlongRay(RobotProxy *robot_, Steppable *floorray_, float mt_=0.15)
		:robot(robot_),floorray(NULL),mt(mt_)
		{assert(robot);done.Set(FUZZFALSE);floorray=Getport(floorray_);};
	virtual	Data	*Step(void); // gimme next waypoint
	virtual	void	Reset(void);
	virtual	void	Abort(void);
protected:	
	RobotProxy *robot;
	Steppable *floorray;
	float		mt;		// max trans velocity
};

class ExtendRay : public Action {
public:
	ExtendRay(RobotProxy *robot_, Steppable *dattitude_)
		:robot(robot_),dattitude(NULL),the_ray(0,0,0)
		{assert(robot);Reset();dattitude=Getport(dattitude_);};
	virtual	Data	*Step(void); // gimme next waypoint
	virtual	void	Reset(void);
	virtual	void	Abort(void);
protected:	
	RobotProxy 	*robot;
	Steppable	*dattitude;
	FloorRay	the_ray;
};

class WaypointToFloorPoint : public Action {
public:
	WaypointToFloorPoint(Steppable *waypoint_)
		:waypoint(waypoint_),the_fpoint(0.0,0.0)
		{waypoint=Getport(waypoint_);assert(waypoint_);};
	virtual	Data	*Step(void);
	virtual	void	Reset(void)		{if (waypoint) waypoint->Reset();};
	virtual	void	Abort(void)		{if (waypoint) waypoint->Abort();};
	virtual Fuzz	*IsDone(void)	{if (waypoint) return waypoint->IsDone(); else return &fuzz_true;}
protected:	
	Steppable	*waypoint;
	FloorPoint	the_fpoint;
};


#endif	/* _APsup_h_ */
