/*

DISCLAIMER 

InSitu is provided as is, without any kind of guarantees or warranty. There is 
no documentation. The author does not provide any kind of support to users of 
this source code and is not responsible for any kind of problems or damages 
resulting from its use. 

The source code is the result of a long iterative design process for a research 
prototype and was never intended to be used in any kind of product. In general, 
the code cannot be expected to compile in any straightforward manner on your 
computer system, since the robot on which it was developed, has a custom design. 

*/
/* Rene Schaad                                                    */
/* Departement of Computer Science, Univ. of Zurich, Switzerland  */
/* schaad@ifi.unizh.ch                                            */


#include <stdio.h>
#include <math.h>
#include <assert.h>
#include <limits.h>

#include "Graph.h"

#define SAME_POINT_DIST 0.2
#define XFIG_SCALE      450.0

//---- GraphPath ---------------------------------------------------------

GraphPath *GraphPath::Duplicate(void)
{
	GraphPath *npath = new GraphPath;
	npath->length = length;		// copy length
	npath->end_edge=end_edge;	// copy end edge
	npath->edges = edges;		// copy all edge pointers (the path)
	return npath;
}

void GraphPath::Print(void)
{
	if (edges.Size()==0) {
		printf("Graph Path contains NULL edges\n");
		return;
	} 
	GraphEdge *eptr;
	edges.Reset();
	eptr = (GraphEdge*)edges.Iterate(); // first element
	assert(eptr);	// if size>0, eptr must !=NULL !!
	printf("\n--------- Graph Path with %d edges over %f meters --------------\n",
		edges.Size(),length);
	printf("\nFrom (%f, %f)\n",eptr->to_node->x,eptr->to_node->y);
	while (NULL!=(eptr = (GraphEdge*)edges.Iterate())) {
		printf("-------> (%f, %f)\n",eptr->to_node->x,eptr->to_node->y);
	}
}


//---- Graph ---------------------------------------------------------

// This function uses the A* algorithm

GraphPath	*UnGraph::GetShortestPath(GraphNode *fnode, GraphNode *tnode)
{
	assert(fnode&&tnode);
	if (!(fnode&&tnode)) return NULL;
	
	// initialize all variables and the priority queue
	gx = tnode->x;						// coordinates of goal node
	gy = tnode->y;
	PriorityQueue	paths;				// sorted list of paths found so far
	
	start_edge->to_node=fnode;			// dummy start edge to the start node
	GraphPath	*start_path = new GraphPath;// dummy start path to the start node
	start_path->edges.Append(start_edge);	// append the dummy edge to it
	start_path->length=0.0;					// zero length
	start_path->end_edge=start_edge;		// end edge is to start node
	
	paths.Enqueue(start_path,LowerBound(fnode)); // add to the p-queue w/ lower bound
	
	GraphPath *curr_path=NULL;
	// check whether goal is reached
	while ((paths.Size()>0) && 
			((curr_path = (GraphPath*)paths.FindFirst())->end_edge->to_node != tnode)) {
		// expand this element by one level
		GraphEdge *new_out_edge;
		assert(curr_path->end_edge);						// do some checks
		assert(curr_path->end_edge->to_node);
		GraphNode *lastnode = curr_path->end_edge->to_node; // get last node from where to expand
		lastnode->out_edges.Reset();						// start iterating thru new edges
		
		// for each new outgoing node
		while (NULL!=(new_out_edge = (GraphEdge*)(lastnode->out_edges.Iterate()))) {
			GraphNode *new_to_node = new_out_edge->to_node;
			float 	new_length = curr_path->length + new_out_edge->distance;
			// search p-queue to see whether this node was reached before
			GraphPath *pptr;
			paths.Reset();
			boolean	better_path_found=FALSE;
			while (NULL!=(pptr=(GraphPath*)paths.Iterate())) {
				float check_dist;
				if (NodeInPath(new_to_node,pptr,check_dist)) {	// if other path reached here ..
					if (check_dist>new_length) {
						// if old path is longer, remove it from pqueue
						delstack.Push(pptr);
					} else {
						better_path_found=TRUE;
						break;
					}
				} 
			}
			// delete paths that are longer than the new one
			while (NULL!=(pptr=(GraphPath*)delstack.Pop())) 
				delete ((GraphPath*)paths.Remove(pptr));
			// if no better path was found, add it to p-queue (else ignore)
			if (! better_path_found) {
				GraphPath *new_ext_path = curr_path->Duplicate(); // copy
				new_ext_path->edges.Append(new_out_edge); 		  // extend it
				new_ext_path->length = new_length;
				new_ext_path->end_edge = new_out_edge;
				paths.Enqueue(new_ext_path,new_length+LowerBound(new_to_node));
			}
		}
		// remove the current path
		GraphPath *r = (GraphPath*)paths.Remove(curr_path);
		assert(r);
		delete curr_path; // delete the old path
	}
	
	// deallocate priority queue and return result
	assert(delstack.Size()==0);	// all marked paths must be deleted
	float final_dist;
	GraphPath *result = (GraphPath*)paths.Dequeue(final_dist);
	if (result) {
		assert(fabs(result->length-final_dist)<0.01); // estimate must be == len
	}
	paths.Reset();
	GraphPath *pptr;
	while (NULL!=(pptr=(GraphPath*)paths.Iterate())) delete pptr;

	return result;
}

boolean	 UnGraph::Forbidden(float x, float y)
{
	ForbiddenArea *f;	// pointer to forbidden area
	forbid.Reset();
	while (NULL!=(f=(ForbiddenArea*)forbid.Iterate())) 
		if (x>f->xmin && x<f->xmax && y>f->ymin && y<f->ymax) return TRUE;
	return FALSE;
}


float	UnGraph::Dist(float x1, float y1, float x2, float y2)
{
	float dx = x1 - x2;
	float dy = y1 - y2;
	return sqrt(dx*dx+dy*dy);
}

float	UnGraph::LowerBound(GraphNode *fnode)
{
	assert(fnode);
	return Dist(fnode->x,fnode->y,gx,gy);
}

boolean UnGraph::NodeInPath(GraphNode *tnode, GraphPath *tpath, float &dist)
{
	dist = 0.0;
	List *pe = &(tpath->edges);
	GraphEdge *eptr;
	pe->Reset();
	while (NULL!=(eptr=(GraphEdge*)pe->Iterate())) {
		dist += eptr->distance;
		if (eptr->to_node==tnode) return TRUE;
	}
	return FALSE;
}

GraphNode *UnGraph::GetNode(float x_, float y_)
{
	GraphNode *nptr=NULL;
	nodes.Reset();
	while (NULL!=(nptr = (GraphNode*)nodes.Iterate())) {
		if ( (fabs(nptr->x - x_)<SAME_POINT_DIST) &&
			 (fabs(nptr->y - y_)<SAME_POINT_DIST) ) {
			return nptr;	// found the point, return it!
		}
	}
	return NULL;
}

GraphNode *UnGraph::GetClosestNode(float &dist_,float x_, float y_, float max_)
{
	GraphNode 	*nptr=NULL;
	GraphNode 	*best_ptr=NULL;
	float		best_dist=max_;
	nodes.Reset();
	while (NULL!=(nptr = (GraphNode*)nodes.Iterate())) {
		float dist = Dist(nptr->x,nptr->y,x_,y_);
		if (dist<=best_dist) {
			best_dist= dist;
			best_ptr = nptr;	// found the point, return it!
		}
	}
	dist_ = best_dist;
	return best_ptr;
}

void UnGraph::Clear(void)
{
	// clear start edge 
	start_edge->to_node=NULL;
	start_edge->distance=0.0;
	
	// clear the stack and all node and edge stores
	delstack.Clear();		// objects IN stack are not deallocated
	edges.DeleteAll();		// UnGraph is OWNER of these edges and nodes
	nodes.DeleteAll();
	wedges.DeleteAll();		// deleta all edges
	wnodes.DeleteAll();
	forbid.DeleteAll();		// delete all forbidden areas
}

UnGraph *UnGraph::Duplicate() 
{
	UnGraph *nug = new UnGraph;
	
	GraphEdge *eptr;
	GraphNode *nptr;
	GraphNode *from_node;
	
	// for each node, add all links to outnodes
	nodes.Reset();
	while (NULL!=(nptr = (GraphNode*)nodes.Iterate())) {
		from_node = nug->InsertNode(nptr->x,nptr->y);
		// for each outgoing edge of that node
		nptr->out_edges.Reset();
		while (NULL!=(eptr = (GraphEdge*)nptr->out_edges.Iterate())) {
			GraphNode *to_node = nug->InsertNode(eptr->to_node->x,eptr->to_node->y);
			nug->InsertEdge(from_node,to_node);
		}
	}
	
	// do the same for all walls
	// for each node, add all links to outnodes
	wnodes.Reset();
	while (NULL!=(nptr = (GraphNode*)wnodes.Iterate())) {
		from_node = nug->InsertWNode(nptr->x,nptr->y);
		// for each outgoing edge of that node
		nptr->out_edges.Reset();
		while (NULL!=(eptr = (GraphEdge*)nptr->out_edges.Iterate())) {
			GraphNode *to_node = nug->InsertWNode(eptr->to_node->x,eptr->to_node->y);
			nug->InsertWEdge(from_node,to_node);
		}
	}
	
	return nug;
}


void UnGraph::Print(void)
{
	// print all nodes
	GraphNode *nptr;
	nodes.Reset();
	printf("----- %d nodes --------\n", nodes.Size());
	while (NULL!=(nptr = (GraphNode*)nodes.Iterate())) {
		printf("Node at (%f,%f) with %d edges.\n",
			nptr->x,nptr->y,nptr->out_edges.Size());
	}
	
	// print all edges
	GraphEdge *eptr;
	edges.Reset();
	printf("----- %d edges --------\n", edges.Size());
	while (NULL!=(eptr = (GraphEdge*)edges.Iterate())) {
		printf("Edge, length %f ",eptr->distance);
		printf("to point at (%f,%f)\n",eptr->to_node->x,eptr->to_node->y);
	}
}

GraphNode *UnGraph::InsertNode(float x_, float y_)
{
	// see whether it already exists
	GraphNode *nptr;
	if (NULL!=(nptr = GetNode(x_,y_))) return nptr;
	
	// if not, allocate a new node and add it to the list
	nodes.Append(nptr = new GraphNode(x_,y_));
	return nptr;
}

GraphEdge *UnGraph::InsertEdge(GraphNode *from_, GraphNode *to_)
{
	assert(from_ && to_);
	
	// allocate object and add to list 
	GraphEdge *eptr = new GraphEdge(to_, Dist(from_->x,from_->y,to_->x,to_->y));
	edges.Append(eptr);
	
	// add this link to source node
	from_->AppendOutEdge(eptr);
	return eptr;
}


GraphNode *UnGraph::InsertWNode(float x_, float y_)
{
	GraphNode *nptr;
	wnodes.Append(nptr = new GraphNode(x_,y_));
	return nptr;
}

GraphEdge *UnGraph::InsertWEdge(GraphNode *from_, GraphNode *to_)
{
	assert(from_);
	assert(to_);
	
	// allocate object and add to list 
	GraphEdge *eptr = new GraphEdge(to_, Dist(from_->x,from_->y,to_->x,to_->y));
	wedges.Append(eptr);
	
	// add this link to source node
	from_->AppendOutEdge(eptr);
	return eptr;
}


#define FERROR(S)	{printf(S);fclose(graphfile);return FALSE;}

boolean UnGraph::ReadFile(void)
{
	if (! filename) {
		printf("UnGraph: could not open map file, file name was NULL\n");
		return FALSE;
	};
	FILE *graphfile;
	
	// input fields from file
	char  	fig[10];	// #FIG header
	float 	figversion;	// version of xfig
	char  	orient[30];	// orientation of paper
	char	center[30];	// centered?
	char 	metric[30];	// metric?
	int		res;		// resolution (1200)
	int		two;		// a two (what is this?)
	
	// input fields on each header line
	int		type1;		// type information 
	int		type2;		// more type information
	int		lwidth;		// line width
	float	npoints;	// number of points in construct
	int		d0;			// global dummy 
	float	d1,d2;		// global dummies for all kinds of stuff
	
	// open the file
	printf("Attempting to open xfig Graph file (%s)..",filename);
	if ((graphfile=fopen(filename,"r"))==NULL) 
		FERROR("..NOT GOOD, it failed! (in UnGraph::ReadFile)\n");
	printf("..ok\n");
	
	// read in initial fields and check them
	fscanf(graphfile," %s %f %s %s %s %d %d ",
		fig,&figversion,orient,center,metric,&res,&two);
	if (fabs(figversion-3.1)>0.01 || res!=1200 || two!=2) 
		FERROR("ERROR: error in header of xfig file\n");
	
	// read in all the lines
	while (1) {
		// read in line header
		if (2 != fscanf(graphfile," %d %d ", &type1, &type2)) {
			break;
		}
		
		//////////// polyline /////////////////////////////
		if (type1==2 && type2==1) {	
			// get line width
			if (2 != fscanf(graphfile," %d %d ", &d0, &lwidth))
				FERROR("ERROR: could not get line width (line)\n");
			// get numer of points	
			for (int i=0;i<12;i++) 
				if (1 != fscanf(graphfile," %f ", &npoints))
					FERROR("ERROR: line too short\n");
			if (npoints<2.0) 
					FERROR("ERROR: cannot process unconnected points\n");
			unsigned int pts = (unsigned int)npoints;
			boolean first=TRUE,wfirst=TRUE;
			GraphNode *prevnode=NULL;
			GraphNode *wprevnode=NULL;
			// read all points
			for (unsigned int j=0;j<pts;j++) {
				int x,y;
				if (2!=fscanf(graphfile," %d %d ", &x,&y)) 
					FERROR("ERROR: could not read all data points\n");
				y = -y;		// everything is reversed in map !!
				if (lwidth==1) {	// path segments
					if (first) {
						prevnode = InsertNode(x/XFIG_SCALE,y/XFIG_SCALE);
						first=FALSE;
					} else {
						GraphNode *currnode = InsertNode(x/XFIG_SCALE,y/XFIG_SCALE);
						InsertEdge(prevnode,currnode);	// both directions
						InsertEdge(currnode,prevnode);
						prevnode = currnode;
					}
				} else {			// wall segments
					if (wfirst) {
						wprevnode = InsertWNode(x/XFIG_SCALE,y/XFIG_SCALE);
						wfirst=FALSE;
					} else {
						GraphNode *currnode = InsertWNode(x/XFIG_SCALE,y/XFIG_SCALE);
						InsertWEdge(wprevnode,currnode);	// both directions
						wprevnode = currnode;
					}
				}
			}
				
		//////////// circle ////////////////////////////////
		} else if (type1==1 && type2==3) {		// a circle
			float fx,fy;
			// get line width
			if (2 != fscanf(graphfile," %d %d ", &d0, &lwidth))
				FERROR("ERROR: could not get line width (circle)\n");
			// get x/y coordinates
			for (int i=0;i<9;i++) 
				if (1 != fscanf(graphfile," %f ", &fx))
					FERROR("ERROR: line too short scanning circle (fx)\n");
			if (1 != fscanf(graphfile," %f ", &fy))
				FERROR("ERROR: line too short scanning circle (fy)\n");
			fy = -fy; // everything is reversed in map
			for (int j=0;j<3;j++) 
				if (2 != fscanf(graphfile," %f %f ", &d1, &d2))
					FERROR("ERROR: line too short scanning circle (dummies)\n");
			
		//////////// FORBIDDEN SQUARES /////////////////////////////
		} else 	if (type1==2 && type2==2) {		// a square
			// get line width
			if (2 != fscanf(graphfile," %d %d ", &d0, &lwidth))
				FERROR("ERROR: could not get line width (line)\n");
			if (lwidth!=2) 
					FERROR("ERROR: forbidden area must have line width of 2!\n");
					
			// get numer of points	
			for (int i=0;i<12;i++) 
				if (1 != fscanf(graphfile," %f ", &npoints))
					FERROR("ERROR: line too short\n");
			if (npoints!=5) 
					FERROR("ERROR: forbidden area must have exactly 5 points!\n");

			// read all points
			int xmin=INT_MAX,xmax=INT_MIN,ymin=INT_MAX,ymax=INT_MIN;
			for (unsigned int j=0;j<npoints;j++) {
				int x,y;
				if (2!=fscanf(graphfile," %d %d ", &x,&y)) 
					FERROR("ERROR: could not read all data points of forbidden area\n");
				y = -y;		// everything is reversed in map !!
				if (x<xmin) xmin = x;
				if (x>xmax) xmax = x;
				if (y<ymin) ymin = y;
				if (y>ymax) ymax = y;
			}
			forbid.Append(new ForbiddenArea(xmin/XFIG_SCALE,xmax/XFIG_SCALE,
											ymin/XFIG_SCALE,ymax/XFIG_SCALE));
		
		//////////// ERROR /////////////////////////////
		} else 								
			FERROR("ERROR: wrong type information in line of xfig file\n");
	}
	
	assert((unsigned int)(edges.Size()/2)*2 == edges.Size()); // must have even # of edges
	fclose(graphfile);
	printf("Closed graph file after reading:\n");
	printf("   Hiway: %d nodes and %d edges\n", nodes.Size(),edges.Size());
	printf("   Walls: %d nodes and %d edges\n", wnodes.Size(),wedges.Size());
	printf("   Areas: %d forbidden areas\n\n\n", forbid.Size());
	return TRUE;
}

