/*

DISCLAIMER 

InSitu is provided as is, without any kind of guarantees or warranty. There is 
no documentation. The author does not provide any kind of support to users of 
this source code and is not responsible for any kind of problems or damages 
resulting from its use. 

The source code is the result of a long iterative design process for a research 
prototype and was never intended to be used in any kind of product. In general, 
the code cannot be expected to compile in any straightforward manner on your 
computer system, since the robot on which it was developed, has a custom design. 

*/

#ifndef _Graph_h_
#define _Graph_h_

/* Rene Schaad                                                    */
/* Departement of Computer Science, Univ. of Zurich, Switzerland  */
/* schaad@ifi.unizh.ch                                            */

#include "Lists.h"
#include "defines.h"

//---- Graph Operations -----------------------------------------------------------

class GraphNode;

class GraphEdge : public ISobject {
public:
	GraphEdge(GraphNode *tonode_, float distance_)
		:to_node(tonode_),distance(distance_){};
	GraphNode	*to_node;
	float		distance;
};

class GraphPath;

class GraphNode : public ISobject {
public:
	GraphNode(float x_, float y_):x(x_),y(y_){};
	void	AppendOutEdge(GraphEdge *newedge){out_edges.Append(newedge);};
	List	out_edges;	// outgoing edges, is kept ordered according to weight
	float	x,y;		// coordinates of node
};

class GraphPath : public ISobject {
public:
	GraphPath():length(0.0){};
	GraphPath	*Duplicate();// make copy of myself
	void		Print();	// print the path
	List		edges;		// list of edges
	float		length;		// summed lengths of edges
	GraphEdge	*end_edge;	// last edge in list of edges
};

class ForbiddenArea : public ISobject {
public:
	ForbiddenArea(ForbiddenArea &t)
		:xmin(t.xmin),xmax(t.xmax),ymin(t.ymin),ymax(t.ymax){};
	ForbiddenArea(float xmin_, float xmax_, float ymin_, float ymax_)
		:xmin(xmin_),xmax(xmax_),ymin(ymin_),ymax(ymax_){};
	void		Print(){printf("Forbidden area: x(%f->%f),y(%f->%f)",
						xmin,xmax,ymin,ymax);};
	float xmin,xmax,ymin,ymax;	// bounding box of forbidden area
};

class UnGraph : public ISobject {
public:
	UnGraph(char *filename_)
		:filename(filename_)
		{start_edge=new GraphEdge(NULL,0.0);ReadFile();};
	UnGraph():filename(NULL){start_edge=new GraphEdge(NULL,0.0);};
	~UnGraph(){Clear();delete start_edge;};
	void	Print(void);
	
	// path management
	GraphNode	*GetNode(float x_, float y_);
	GraphNode	*GetClosestNode(float &dist_,float x_, float y_, float max_); // max
	GraphPath	*GetShortestPath(GraphNode *fnode, GraphNode *tnode);
	
	// forbidden area management
	boolean		Forbidden(float x, float y);
	
	// memory management
	void		Clear();			// set back to empty graph
	UnGraph		*Duplicate(void);	// return copy of myself (in no particular state)
	
	// these lists register ALL the dynamically allocated objects
	// UnGraph will delete them, too
	// _EXCEPT_ for the solution from GetShortestPath, which is not
	// deallocated by UnGraph
	List	edges;	// edges of graph
	List	nodes;	// nodes (==vertexes) in the graph
	List	wedges;	// wall edges 
	List	wnodes;	// wall nodes
	List	forbid;	// forbidden area (square area) 
	
private:
	// stuff required for shortest path computation
	float		gx,gy;				// goal node coords
	float		Dist(float x1, float y1, float x2, float y2); // dist between points
	float		LowerBound(GraphNode *fnode);	// compute lower bound estimate of remaining cost
	boolean		NodeInPath(GraphNode *tnode, GraphPath *tpath, float &dist);
	Stack		delstack;	// paths to delete
	GraphEdge	*start_edge;// edge to start point
	
	// other stuff
	boolean 	ReadFile(void);
	GraphNode	*InsertNode(float x_, float y_);	// add node to list
	GraphEdge	*InsertEdge(GraphNode *from_, GraphNode *to_);
	GraphNode	*InsertWNode(float x_, float y_);	// add node to list
	GraphEdge	*InsertWEdge(GraphNode *from_, GraphNode *to_);
	
	char *filename;
};


#endif	/* _Graph_h_ */
