/*

DISCLAIMER 

InSitu is provided as is, without any kind of guarantees or warranty. There is 
no documentation. The author does not provide any kind of support to users of 
this source code and is not responsible for any kind of problems or damages 
resulting from its use. 

The source code is the result of a long iterative design process for a research 
prototype and was never intended to be used in any kind of product. In general, 
the code cannot be expected to compile in any straightforward manner on your 
computer system, since the robot on which it was developed, has a custom design. 

*//* Rene Schaad                                                    */
/* Departement of Computer Science, Univ. of Zurich, Switzerland  */
/* schaad@ifi.unizh.ch                                            */


#include "IFsup.h"
#include <iostream.h>
#include <math.h>

#define ACH_T_THRES	0.001
#define ACH_A_THRES	0.005

//---- Interface SUPs ---------------------------------------------------------

Data *BlockedQ::Step(void)
{
	boolean is_blocked = (robot->GetBumpers()>0);
	is_blocked	= is_blocked || (robot->GetSonar(1) < block_dist );
	is_blocked	= is_blocked || (robot->GetSonar(2) < block_dist );
	is_blocked	= is_blocked || (robot->GetSonar(8) < block_dist );

	if (is_blocked) return &fuzz_true;
	else return &fuzz_false;
}

Data *BumpedQ::Step(void)
{
	word new_bump = robot->GetBumpers();

	switch (sensor) {
		case 1:return ((new_bump&0x0808)>0?&fuzz_true:&fuzz_false); // front
		case 2:return ((new_bump&0x0404)>0?&fuzz_true:&fuzz_false);
		case 3:return ((new_bump&0x0202)>0?&fuzz_true:&fuzz_false); // right
		case 4:return ((new_bump&0x0101)>0?&fuzz_true:&fuzz_false);
		case 5:return ((new_bump&0x8080)>0?&fuzz_true:&fuzz_false); // back
		case 6:return ((new_bump&0x4040)>0?&fuzz_true:&fuzz_false);
		case 7:return ((new_bump&0x2020)>0?&fuzz_true:&fuzz_false); // left
		case 8:return ((new_bump&0x1010)>0?&fuzz_true:&fuzz_false);

		case 0:
		default: return (new_bump?&fuzz_true:&fuzz_false);
	};
}

Data *SonarQ::Step(void)
{
	retval = (double) robot->GetSonar(sensor);
	return &retval;
}

Data *LightQ::Step(void)
{
	retval = (double) robot->GetAmbient(sensor);
	return &retval;
}

Data *RadarQ::Step(void)
{
	if (robot->GetRadar()) return &fuzz_true;
	else return &fuzz_false;
}

Data *PirQ::Step(void)
{
	if (robot->GetPIR()) return &fuzz_true;
	else return &fuzz_false;
}


Data *LightACQ::Step(void)
{
	retval = (double) robot->Get50Hz();
	return &retval;
}

Data *TransvQ::Step(void)
{
	float t,r;
	robot->GetRobotVelocity(t,r);
	retval = t;
	return &retval;
}

Data *RotvQ::Step(void)
{
	float t,r;
	robot->GetRobotVelocity(t,r);
	retval = r;
	return &retval;
}

Data *PosxQ::Step(void)
{
	float x,y,a;
	robot->GetDRPosition(x,y,a);
	retval = x;
	return &retval;
}

Data *PosyQ::Step(void)
{
	float x,y,a;
	robot->GetDRPosition(x,y,a);
	retval = y;
	return &retval;
}

Data *AttitudeQ::Step(void)
{
	float x,y,a;
	robot->GetDRPosition(x,y,a);
	retval = a;
	return &retval;
}

Data *CompassQ::Step(void)
{
	retval = robot->GetCompass();
	return &retval;
}

Data *PanQ::Step(void)
{
	float p,t;
	robot->GetCameraPos(p,t);
	retval = p;
	return &retval;
}

Data *TiltQ::Step(void)
{
	float p,t;
	robot->GetCameraPos(p,t);
	retval = t;
	return &retval;
}

Data *ModeQ::Step(void)
{
	retval = (double)robot->GetMode();
	return &retval;
}

Data *SoundQ::Step(void)
{
	retval = (double)robot->GetSound();
	return &retval;
}

/////////////////////////////////////////////////////////////
//
//    Actuator Interface SUPs
//

Data *SetSound::Step(void) 
{
	sound_t s;
	
	
	if (arg) {
		Data *dr = arg->Step();
		if (dr) {
			if (dr->type==DT_Number) {
				s=(sound_t)((Number*)dr)->Get();
			} else {
				s=NOSOUND;
			}
		} else {
			s=NOSOUND;
		}
	} else {
		s=sound_no;
	}
	
	retval=(float)s;
	robot->SetSound(s);
	return &retval;
}

Say::Say(RobotProxy *robot_, char *txt_):ActuatorAction(robot_,NULL),text(NULL)
{
	word len=0;
	if (txt_) {
		if ((len=(word)strlen(txt_))<SPK_TXTLEN) {
			text = new char [len+4];
			strcpy(text,txt_);
			return;
		} 
	}
	
	text = "Error speaking text, in constructor of SAY. ";
}

Data *Say::Step(void) 
{
	char *t;
	
	if (arg) {
		Data *dr = arg->Step();
		if (dr) {
			if (dr->type==DT_Text) {
				t=((Text*)dr)->Get();
			} else {
				t="Error speaking text, because of invalid argument type. ";
			}
		} else {
			t="Error speaking text, because of NULL argument. ";
		}
	} else {
		t=text;
	}
	
	robot->SetSpeech(t);
	retval=0.0;
	return &retval;
}

Data *SetVel::Step(void) 
{
	
	if (arg) {
		// there is a steppable, delivering input
		Data *dr = arg->Step();
		if (dr) {
			if (dr->type==DT_DataList) {
				Data *tp, *ap;
				DataList *l = (DataList*)dr;
				l->ResetList();
				tp = l->Iterate();
				ap = l->Iterate();
				if (tp) if (tp->type==DT_Number) t=((Number*)tp)->Get();
				if (ap) if (ap->type==DT_Number) a=((Number*)ap)->Get();
			}
		} 
	}
	
	retval=0.0;
	robot->SetRobotVelocity(t,a);
	return &retval;
}

Data *AchieveVel::Step(void) 
{
	
	if (arg) {
		// there is a steppable, delivering input
		Data *dr = arg->Step();
		if (dr) {
			if (dr->type==DT_DataList) {
				Data *tp, *ap;
				DataList *l = (DataList*)dr;
				l->ResetList();
				tp = l->Iterate();
				ap = l->Iterate();
				if (tp) if (tp->type==DT_Number) t=((Number*)tp)->Get();
				if (ap) if (ap->type==DT_Number) a=((Number*)ap)->Get();
			}
		} 
	}
	
	retval=0.0;
	robot->SetRobotVelocity(t,a);
	
	// set done-value correctly
	float rt,ra;	// real velocities
	robot->GetRobotVelocity(rt,ra);
	if ( (fabs(t-rt)<ACH_T_THRES) && (fabs(a-ra)<ACH_A_THRES) ) done.Set(FUZZTRUE);
	else														done.Set(FUZZFALSE);
	
	return &retval;
}

Data *SetPos::Step(void) 
{
	
	if (arg) {
		// there is a steppable, delivering input
		Data *dr = arg->Step();
		if (dr) {
			if (dr->type==DT_DataList) {
				Data *xp,*yp,*ap;
				DataList *l = (DataList*)dr;
				l->ResetList();
				xp = l->Iterate();
				yp = l->Iterate();
				ap = l->Iterate();
				if (xp) if (xp->type==DT_Number) x=((Number*)xp)->Get();
				if (yp) if (yp->type==DT_Number) y=((Number*)yp)->Get();
				if (ap) if (ap->type==DT_Number) a=((Number*)ap)->Get();
			}
		} 
	}
	
	robot->SetDRPosition(x,y,a);
	retval=0.0;
	return &retval;
}

Data *SetDist::Step(void) 
{
	
	if (arg) {
		// there is a steppable, delivering input
		Data *dr = arg->Step();
		if (dr) {
			if (dr->type==DT_DataList) {
				Data *tp,*ap;
				DataList *l = (DataList*)dr;
				l->ResetList();
				tp = l->Iterate();
				ap = l->Iterate();
				if (tp) if (tp->type==DT_Number) drt=((Number*)tp)->Get();
				if (ap) if (ap->type==DT_Number) dra=((Number*)tp)->Get();
			}
		} 
	}
	
	robot->SetDRDist(drt,dra);
	retval=0.0;
	return &retval;
}

Data *SetCam::Step(void) 
{
	
	if (arg) {
		// there is a steppable, delivering input
		Data *dr = arg->Step();
		if (dr) {
			if (dr->type==DT_DataList) {
				Data *panp,*tiltp;
				DataList *l = (DataList*)dr;
				l->ResetList();
				panp = l->Iterate();
				tiltp = l->Iterate();
				if (panp) if (panp->type==DT_Number) pan=((Number*)panp)->Get();
				if (tiltp) if (tiltp->type==DT_Number) tilt=((Number*)tiltp)->Get();
			}
		} 
	}
	
	retval=0.0;
	robot->SetCameraPos(pan,tilt);
	return &retval;
}

Data *SetMode::Step(void) 
{
	
	if (arg) {
		// there is a steppable, delivering input
		Data *dr = arg->Step();
		if (dr) 
			if (dr->type==DT_Number) mode=(byte)((Number*)dr)->Get();
	}
	
	retval=0.0;
	robot->SetMode(mode);
	return &retval;
}

Data *SetISize::Step(void) 
{
	
	if (arg) {
		// there is a steppable, delivering input
		Data *dr = arg->Step();
		if (dr) {
			if (dr->type==DT_DataList) {
				Data *wp,*hp;
				DataList *l = (DataList*)dr;
				l->ResetList();
				wp = l->Iterate();
				hp = l->Iterate();
				if (wp) if (wp->type==DT_Number) width=((Number*)wp)->Get();
				if (hp) if (hp->type==DT_Number) height=((Number*)hp)->Get();
			}
		} 
	}
	
	retval=0.0;
	robot->SetCameraImageSize(width,height);
	return &retval;
}




