/*

DISCLAIMER 

InSitu is provided as is, without any kind of guarantees or warranty. There is 
no documentation. The author does not provide any kind of support to users of 
this source code and is not responsible for any kind of problems or damages 
resulting from its use. 

The source code is the result of a long iterative design process for a research 
prototype and was never intended to be used in any kind of product. In general, 
the code cannot be expected to compile in any straightforward manner on your 
computer system, since the robot on which it was developed, has a custom design. 

*/

#ifndef _IFsup_h_
#define _IFsup_h_

/* Rene Schaad                                                    */
/* Departement of Computer Science, Univ. of Zurich, Switzerland  */
/* schaad@ifi.unizh.ch                                            */

#include "PrgStruct.h"
#include "RobotProxy.h"

/////////////////////////////////////////////////////////////
//
//    Sensor Interface SUPs
//

class SensorQuery : public Action {
public:
	SensorQuery(RobotProxy *robot_, byte sensor_=0)
	   :robot(robot_),sensor(sensor_)
	   {done.Set(FUZZTRUE);};
protected:
	RobotProxy *robot;
	byte		sensor;
};

class BlockedQ : public SensorQuery {
public:
	BlockedQ(RobotProxy *robot_, float block_dist_=0.5)
		:block_dist(block_dist_),SensorQuery(robot_)
		{assert(block_dist<2.0 && block_dist>0.4);};
	virtual	Data	*Step(void);
private:
	float	block_dist;
};

class BumpedQ : public SensorQuery {
public:
	BumpedQ(RobotProxy *robot_, byte sensor_=0):SensorQuery(robot_,sensor_){};
	virtual	Data	*Step(void);
};

class SonarQ : public SensorQuery {
public:
	SonarQ(RobotProxy *robot_, byte sensor_=0):SensorQuery(robot_,sensor_){};
	virtual	Data	*Step(void);
private:
	Number		retval;
};

class LightQ : public SensorQuery {
public:
	LightQ(RobotProxy *robot_, byte sensor_=0):SensorQuery(robot_,sensor_){};
	virtual	Data	*Step(void);
private:
	Number		retval;
};

class RadarQ : public SensorQuery {
public:
	RadarQ(RobotProxy *robot_):SensorQuery(robot_){};
	virtual	Data	*Step(void);
};

class PirQ : public SensorQuery {
public:
	PirQ(RobotProxy *robot_):SensorQuery(robot_){};
	virtual	Data	*Step(void);
};

class LightACQ : public SensorQuery {
public:
	LightACQ(RobotProxy *robot_):SensorQuery(robot_){};
	virtual	Data	*Step(void);
private:
	Number		retval;
};

class TransvQ : public SensorQuery {
public:
	TransvQ(RobotProxy *robot_):SensorQuery(robot_){};
	virtual	Data	*Step(void);
private:
	Number		retval;
};

class RotvQ : public SensorQuery {
public:
	RotvQ(RobotProxy *robot_):SensorQuery(robot_){};
	virtual	Data	*Step(void);
private:
	Number		retval;
};

class PosxQ : public SensorQuery {
public:
	PosxQ(RobotProxy *robot_):SensorQuery(robot_){};
	virtual	Data	*Step(void);
private:
	Number		retval;
};

class PosyQ : public SensorQuery {
public:
	PosyQ(RobotProxy *robot_):SensorQuery(robot_){};
	virtual	Data	*Step(void);
private:
	Number		retval;
};

class AttitudeQ : public SensorQuery {
public:
	AttitudeQ(RobotProxy *robot_):SensorQuery(robot_){};
	virtual	Data	*Step(void);
private:
	Number		retval;
};

class CompassQ : public SensorQuery {
public:
	CompassQ(RobotProxy *robot_):SensorQuery(robot_){};
	virtual	Data	*Step(void);
private:
	Number		retval;
};

class PanQ : public SensorQuery {
public:
	PanQ(RobotProxy *robot_):SensorQuery(robot_){};
	virtual	Data	*Step(void);
private:
	Number		retval;
};

class TiltQ : public SensorQuery {
public:
	TiltQ(RobotProxy *robot_):SensorQuery(robot_){};
	virtual	Data	*Step(void);
private:
	Number		retval;
};

class ModeQ : public SensorQuery {
public:
	ModeQ(RobotProxy *robot_):SensorQuery(robot_){};
	virtual	Data	*Step(void);
private:
	Number		retval;
};

class SoundQ : public SensorQuery {
public:
	SoundQ(RobotProxy *robot_):SensorQuery(robot_){};
	virtual	Data	*Step(void);
private:
	Number		retval;
};

/////////////////////////////////////////////////////////////
//
//    Actuator Interface SUPs
//

class ActuatorAction : public Action {
public:
	ActuatorAction(RobotProxy *robot_, Steppable *arg_):robot(robot_),retval(0.0){
		arg = Getport(arg_);
		done.Set(FUZZTRUE);
	};
	virtual	void	Abort(void){if (arg) arg->Abort();};
	virtual void	Reset(void){if (arg) arg->Reset();};
	virtual Fuzz	*IsDone(void) {return ((arg)?arg->IsDone():&done);};
protected:
	RobotProxy	*robot;
	Steppable	*arg;
	Number		retval;
};

class SetSound : public ActuatorAction {
public:
	SetSound(RobotProxy *robot_, float sound_no_=NOSOUND):ActuatorAction(robot_,NULL),sound_no((sound_t)sound_no_){};
	SetSound(RobotProxy *robot_, Steppable *arg_):ActuatorAction(robot_,arg_),sound_no(NOSOUND){};
	virtual	Data	*Step(void);
private:
	sound_t		sound_no;
};

class Say : public ActuatorAction {
public:
	Say(RobotProxy *robot_, char *txt_);
	Say(RobotProxy *robot_, Steppable *arg_):ActuatorAction(robot_,arg_),text(NULL){};
	~Say(){delete [] text;};
	virtual	Data	*Step(void);
private:
	char	*text;
};

class SetVel : public ActuatorAction {
public:
	SetVel(RobotProxy *robot_, float t_=0.0, float a_=0.0 )
		:ActuatorAction(robot_,NULL),t(t_),a(a_){};
	SetVel(RobotProxy *robot_, Steppable *arg_)
		:ActuatorAction(robot_,arg_),t(0.0),a(0.0){};
	virtual	Data	*Step(void);
protected:
	float t,a;
};

class AchieveVel : public SetVel {
public:
	AchieveVel(RobotProxy *robot_, float t_=0.0, float a_=0.0 )
		:SetVel(robot_,t_,a_){done.Set(FUZZFALSE);};
	AchieveVel(RobotProxy *robot_, Steppable *arg_)
		:SetVel(robot_,arg_){done.Set(FUZZFALSE);};
	virtual	Data	*Step(void);
};

class SetPos : public ActuatorAction {
public:
	SetPos(RobotProxy *robot_, float x_, float y_, float a_):ActuatorAction(robot_,NULL),x(x_),y(y_),a(a_){};
	SetPos(RobotProxy *robot_, Steppable *arg_):ActuatorAction(robot_,arg_),x(0.0),y(0.0),a(0.0){};
	virtual	Data	*Step(void);
private:
	float x,y,a;
};

class SetDist : public ActuatorAction {
public:
	SetDist(RobotProxy *robot_, float drt_, float dra_):ActuatorAction(robot_,NULL),drt(drt_),dra(dra_){};
	SetDist(RobotProxy *robot_, Steppable *arg_):ActuatorAction(robot_,arg_),drt(0.0),dra(0.0){};
	virtual	Data	*Step(void);
private:
	float drt,dra;
};

class SetCam : public ActuatorAction {
public:
	SetCam(RobotProxy *robot_, float pan_, float tilt_):ActuatorAction(robot_,NULL),pan(pan_),tilt(tilt_){};
	SetCam(RobotProxy *robot_, Steppable *arg_):ActuatorAction(robot_,arg_),pan(0.0),tilt(0.0){};
	virtual	Data	*Step(void);
private:
	float pan,tilt;
};

class SetMode : public ActuatorAction {
public:
	SetMode(RobotProxy *robot_, byte mode_):ActuatorAction(robot_,NULL),mode(mode_){};
	SetMode(RobotProxy *robot_, Steppable *arg_):ActuatorAction(robot_,arg_),mode(0){};
	virtual	Data	*Step(void);
private:
	byte mode;
};

class SetISize : public ActuatorAction {
public:
	SetISize(RobotProxy *robot_, word width_, word height_):ActuatorAction(robot_,NULL),width(width_),height(height_){};
	SetISize(RobotProxy *robot_, Steppable *arg_):ActuatorAction(robot_,arg_),width(0),height(0){};
	virtual	Data	*Step(void);
private:
	word width,height;
};



#endif	/* _IFsup_h_ */
