/*

DISCLAIMER 

InSitu is provided as is, without any kind of guarantees or warranty. There is 
no documentation. The author does not provide any kind of support to users of 
this source code and is not responsible for any kind of problems or damages 
resulting from its use. 

The source code is the result of a long iterative design process for a research 
prototype and was never intended to be used in any kind of product. In general, 
the code cannot be expected to compile in any straightforward manner on your 
computer system, since the robot on which it was developed, has a custom design. 

*//* Rene Schaad                                                    */
/* Departement of Computer Science, Univ. of Zurich, Switzerland  */
/* schaad@ifi.unizh.ch                                            */

#include "Interpreter.h"
#include "defines.h"
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <sys/time.h>
#include <assert.h>
#include "SupMgr.h"
#include "PrgStruct.h"
#include "IFsup.h"
#include "TBehavior.h"
#include "RufusPhys.h"
#include "Navsup.h"

#define CAR part.pair.car
#define CDR part.pair.cdr


//---- Token ---------------------------------------------------------

Token::Token(OperatorProxy *op_,Token *org)
	:op(op_),token_id(TOK_ERROR),number(0.0),ste(NULL),string(NULL)
{
	Copy(org);				// copy if not NULL
}

void Token::Copy(Token *org)
{
	if (org!=NULL) {					// NOOP if org==NULL
		token_id	=org->token_id;
		number		=org->number;
		ste			=org->ste;
		SetString(org->string);
	}
}

void Token::SetString(char *txt)
{
	word len;
	if (txt==NULL) return;
	if ((len=(word)strlen(txt))>MAX_NAME_LEN) {
		op->ISERR3(iserr_stringsize,txt);
		return;
	}
	delete string;
	string = new char [len+1];
	strcpy(string,txt);
}

//////////////////////////////////////////////////////////////////////////////////////////
//  CONS CELLS


ConsCell::ConsCell(OperatorProxy *op_,Token *org):op(op_)
{
	if (org!=NULL) {
		switch (org->token_id) {
			case TOK_SYMBOL:
				type = T_SYMBOL;
				part.ste=org->ste;
				break;
			case TOK_STRING:
				type = T_STRING;
				part.string=NULL;
				SetString(org->string);
				break;
			case TOK_NUMBER:
				type = T_NUMBER;
				part.number = org->number;
				break;
			default:
				type = T_VOID;
				op->ISERR3(iserr_toktocons,(int)org->token_id);
				break;
		};
	} else {
		type = T_VOID;
	}
}

ConsCell::ConsCell(OperatorProxy *op_,ConsCell *car, ConsCell *cdr):op(op_)
{
	type = T_LIST;
	CAR = car;
	CDR = cdr;
}

void ConsCell::SetString(char *txt)
{
	if (type==T_STRING) delete part.string;
	type=T_STRING;
	word len;
	if (txt==NULL) return;
	if ((len=(word)strlen(txt))>MAX_NAME_LEN) {
		op->ISERR3(iserr_stringsize,txt);
		return;
	}
	part.string = new char [len+1];
	strcpy(part.string,txt);
}

void ConsCell::PrintIndent(int indent)
{
	*op << indent;
	for (int i=0;i<indent;i++) *op << "    ";
}

void ConsCell::PrintLn(int indent, CCRole role)
{
	Print(indent,role);
	*op << "\n";
}

void ConsCell::Print(int indent, CCRole role)
{
	// depending on type, print contents
	switch (type) {
		case T_VOID:
			*op << "VOID";
			break;
		case T_NUMBER:
			*op << part.number;
			break;
		case T_STRING:
			*op << "\"" << part.string << "\"";
			break;
		case T_LIST:
			if (role==CC_CAR) {
				*op << "(";
				indent++;
			}
			if (CAR) {
				CAR->Print(indent,CC_CAR);
			}
			if (CDR) {
				*op << " ";
				CDR->Print(indent,CC_CDR);
			}
			if (role==CC_CAR) {
				*op << ")";
			}
			break;
		case T_CONSTCELL:
			*op << part.ste->lexem;
			break;
		case T_SYMBOL:
			*op << part.ste->lexem;
			break;
		case T_SUP:
			*op << "SubUP";
			break;
		case T_SBUILTIN:
			*op << "SBUILTIN";
			break;
		case T_UBUILTIN:
			*op << "UBUILTIN";
			break;
		case T_EBUILTIN:
			*op << "EBUILTIN";
			break;
		default:
			*op << "UNKNOWN: " << type;
			break;
	};
}

//---- Parser ---------------------------------------------------------


Parser::Parser(RobotProxy *robot_,OperatorProxy *op_,SupMgr *new_manager)
	:robot(robot_),op(op_),sup_mgr(new_manager),
	 nesting(0),line_no(0),cmd_buffer(NULL)
{
	symtab = new SymbolTable(op);
	
	returned_val	= new Token(op);
	error_token		= new Token(op);	error_token->token_id = TOK_ERROR;
	eof_token		= new Token(op);  	eof_token->token_id = TOK_EOF;
	
	// CONSTANT ConsCells
	const_nil			= new ConsCell(op); const_nil->type	= T_CONSTCELL;	// add a cell
	const_nilsym		= symtab->Append("nil", T_CONST, const_nil);	// add a symbol
	const_nil->part.ste = const_nilsym;								// link cell to symbol

	const_true			= new ConsCell(op); const_true->type = T_CONSTCELL;	// add a cell
	const_truesym		= symtab->Append("t", T_CONST, const_true);	// add a symbol
	const_true->part.ste = const_truesym;							// link cell to symbol
	
	const_lambda		= new ConsCell(op); const_lambda->type = T_CONSTCELL;	// add a cell
	const_lambdasym		= symtab->Append("lambda", T_CONST, const_lambda);
	const_lambda->part.ste = const_lambdasym;

	// Add evaluating built-ins to Symbol Table
	symtab->AppendBuiltin("*",EBI_mul, T_EBUILTIN);
	symtab->AppendBuiltin("+",EBI_add, T_EBUILTIN);
	symtab->AppendBuiltin("/",EBI_div, T_EBUILTIN);
	symtab->AppendBuiltin("-",EBI_sub, T_EBUILTIN);
	symtab->AppendBuiltin("eval",EBI_eval, T_EBUILTIN);
	symtab->AppendBuiltin("print",EBI_print, T_EBUILTIN);
	symtab->AppendBuiltin("list",EBI_list, T_EBUILTIN);
	symtab->AppendBuiltin("listp",EBI_listp, T_EBUILTIN);
	symtab->AppendBuiltin("atomp",EBI_atomp, T_EBUILTIN);
	symtab->AppendBuiltin("length",EBI_length, T_EBUILTIN);
	symtab->AppendBuiltin("car",EBI_car, T_EBUILTIN);
	symtab->AppendBuiltin("cdr",EBI_cdr, T_EBUILTIN);
	
	// Add non-evaluating built-ins to Symbol Table
	symtab->AppendBuiltin("print-symtab",UBI_print_symtab, T_UBUILTIN);
	symtab->AppendBuiltin("print-docklist",UBI_print_docklist, T_UBUILTIN);
	symtab->AppendBuiltin("quote",UBI_quote, T_UBUILTIN);
	symtab->AppendBuiltin("setq",UBI_setq, T_UBUILTIN);
	symtab->AppendBuiltin("defun",UBI_defun, T_UBUILTIN);
	symtab->AppendBuiltin("load",UBI_load, T_UBUILTIN);
	symtab->AppendBuiltin("exit",UBI_exit, T_UBUILTIN);
	
	// Built-in data types
	symtab->AppendBuiltin("Waypoint",SBI_waypoint, T_SBUILTIN);
	
	// Add built-in universal plans to Symbol Table
	symtab->AppendBuiltin("Achieve",SBI_achieve, T_SBUILTIN);
	symtab->AppendBuiltin("Do",SBI_do, T_SBUILTIN);
	symtab->AppendBuiltin("Stop",SBI_stop, T_SBUILTIN);
	symtab->AppendBuiltin("Every",SBI_every, T_SBUILTIN);
	symtab->AppendBuiltin("If",SBI_if, T_SBUILTIN);
	symtab->AppendBuiltin("Ramp",SBI_ramp, T_SBUILTIN);
	symtab->AppendBuiltin("Taction",SBI_taction, T_SBUILTIN);
	symtab->AppendBuiltin("Ttester",SBI_ttester, T_SBUILTIN);
	symtab->AppendBuiltin("Pulsetrain",SBI_pulsetrain, T_SBUILTIN);
	symtab->AppendBuiltin("Loop",SBI_loop, T_SBUILTIN);
	symtab->AppendBuiltin("Noop",SBI_noop, T_SBUILTIN);
	symtab->AppendBuiltin("Parallel",SBI_parallel, T_SBUILTIN);
	symtab->AppendBuiltin("Sequence",SBI_sequence, T_SBUILTIN);
	symtab->AppendBuiltin("Switch",SBI_switch, T_SBUILTIN);

	symtab->AppendBuiltin("Dock",SBI_dock, T_SBUILTIN);
	symtab->AppendBuiltin("And",SBI_and, T_SBUILTIN);
	symtab->AppendBuiltin("Or",SBI_or, T_SBUILTIN);

//	symtab->AppendBuiltin("LocalMap",SBI_localmap, T_SBUILTIN);
//	symtab->AppendBuiltin("STGraph",SBI_stgraph, T_SBUILTIN);

	// Add built-in interface universal plans to Symbol Table
	symtab->AppendBuiltin("Bumped?",SBI_bumpedQ, T_SBUILTIN);
	symtab->AppendBuiltin("Sonar?",SBI_sonarQ, T_SBUILTIN);
	symtab->AppendBuiltin("Light?",SBI_lightQ, T_SBUILTIN);
	symtab->AppendBuiltin("Radar?",SBI_radarQ, T_SBUILTIN);
	symtab->AppendBuiltin("Pir?",SBI_pirQ, T_SBUILTIN);
	symtab->AppendBuiltin("LightAC?",SBI_lightacQ, T_SBUILTIN);
	symtab->AppendBuiltin("Transv?",SBI_transvQ, T_SBUILTIN);
	symtab->AppendBuiltin("Rotv?",SBI_rotvQ, T_SBUILTIN);
	symtab->AppendBuiltin("Posx?",SBI_posxQ, T_SBUILTIN);
	symtab->AppendBuiltin("Posy?",SBI_posyQ, T_SBUILTIN);
	symtab->AppendBuiltin("Attitude?",SBI_attitudeQ, T_SBUILTIN);
	symtab->AppendBuiltin("Compass?",SBI_compassQ, T_SBUILTIN);
	symtab->AppendBuiltin("Pan?",SBI_panQ, T_SBUILTIN);
	symtab->AppendBuiltin("Tilt?",SBI_tiltQ, T_SBUILTIN);
	symtab->AppendBuiltin("Mode?",SBI_modeQ, T_SBUILTIN);
	symtab->AppendBuiltin("Sound?",SBI_soundQ, T_SBUILTIN);

	// Add built-in interface universal plans to Symbol Table
	symtab->AppendBuiltin("SetSound",SBI_setsound, T_SBUILTIN);
	symtab->AppendBuiltin("Say",SBI_say, T_SBUILTIN);
	symtab->AppendBuiltin("SetVel",SBI_setvelocity, T_SBUILTIN); // trans, ang
	symtab->AppendBuiltin("SetPos",SBI_setposition, T_SBUILTIN); // x,y,a
	symtab->AppendBuiltin("SetCam",SBI_setcam, T_SBUILTIN);	     // pan, tilt
	symtab->AppendBuiltin("SetMode",SBI_setmode, T_SBUILTIN);     
	symtab->AppendBuiltin("SetISize",SBI_setimagesize, T_SBUILTIN);// x,y
}

Parser::~Parser()
{
	delete const_nil;
	delete const_true;
	delete symtab;
}

/* Find next token in input stream */
// TODO:  1. eof handlen 

Token *Parser::Scan(char *inbuffer, unsigned int &position)
{
	unsigned int 	state=2;
	char	 		lexindex=0;
	int				end_encountered=FALSE;
			
	while (1) {
		/* handle end of input */
		if (inbuffer[position]==0)
			if (end_encountered==FALSE) {
				end_encountered=TRUE;
				inbuffer[position]   = ' ';
				inbuffer[position+1] = 0;
			} else
				return eof_token;
			
			
		switch (state) {
			case 1:	break;
			
			/* main state, recognizes starting characters of lexems */
			case 2:	if (Isnewline(inbuffer[position])) {	// NEW LINE
						position++;
						line_no++;
					} else
					if (Isblank(inbuffer[position])) {		// BLANK
						position++;
					} else
					if (Iscomment(inbuffer[position])) {	// COMMENT
						position++;
						state = 10;
					} else
					if (inbuffer[position]=='(') {			// L BRACKET
						position++;
						returned_val->token_id = TOK_LBRACKET;
						return returned_val;
					} else
					if (inbuffer[position]==')') {			// R BRACKET
						position++;
						returned_val->token_id = TOK_RBRACKET;
						return returned_val;
					} else
					if (inbuffer[position]=='+' ||
						inbuffer[position]=='-' ||
						Isdigit(inbuffer[position])) {		// START OF NUMBER
						lexindex=0;
						lexbuf[0]=inbuffer[position];
						lexindex++;
						position++;
						state = 5;
					} else
					if (Isalpha(inbuffer[position])) {		// ALPHA CHAR
						lexindex=0;
						lexbuf[0]=inbuffer[position];
						lexindex++;
						position++;
						state = 3;
					} else
					if (inbuffer[position]=='"') {			// QUOTE
						lexindex=0;
						position++;
						state = 4;
					} else
					if (inbuffer[position]=='.') {			// DECIMAL COMMA
						lexindex=0;
						lexbuf[0]=inbuffer[position];
						lexindex++;
						position++;
						state = 6;
					} else {
						op->ISERR2(iserr_scanner);
						position++;
						return error_token;
					}
					break;
			
			/* process rest of an identifier */
			case 3:	if (Isalpha(inbuffer[position]) ||
						Isdigit(inbuffer[position])) {		// ALPHA / DIGIT
						lexbuf[lexindex]=inbuffer[position];
						lexindex++;
						position++;
					} else 
					if (Isblank(inbuffer[position]) ||
						Iscomment(inbuffer[position]) ||
						(inbuffer[position]=='(') ||
						(inbuffer[position]==')')) {		// END OF IDENTIFIER
						lexbuf[lexindex]=0; /* terminate string */
						returned_val->token_id = TOK_SYMBOL;
						returned_val->ste      = symtab->Append(lexbuf);;
						returned_val->SetString(lexbuf);
						return returned_val;
					} else {
						position++;
						lexbuf[lexindex]=0;
						return error_token;
					}
					break;
				
			/* process rest of a string */	
			case 4:	if (inbuffer[position]=='"') {			// END OF STRING (QUOTE)
						lexbuf[lexindex]=0; /* terminate string */
						returned_val->token_id = TOK_STRING;
						returned_val->SetString(lexbuf);
						position++;
						return returned_val;
					} else {
						lexbuf[lexindex]=inbuffer[position];
						lexindex++;
						position++;
					}
					break;
					
			/* process rest of a numeric value */
			case 5:	if (inbuffer[position]=='.') {			// NUMERICAL COMMA
						lexbuf[lexindex]=inbuffer[position];
						lexindex++;
						position++;
						state = 6;
					} else
					if (Isdigit(inbuffer[position])) {		// DIGIT
						lexbuf[lexindex]=inbuffer[position];
						lexindex++;
						position++;
					} else 
					if (Isalpha(inbuffer[position])) {		// Oops, was a identifier, after all
						lexbuf[lexindex]=inbuffer[position];
						lexindex++;
						position++;
						state = 3;
					} else 
					if (Isblank(inbuffer[position]) ||
						Iscomment(inbuffer[position]) ||
						(inbuffer[position]==')')) {		// END OF DIGIT
						lexbuf[lexindex]=0; /* terminate string */
						returned_val->token_id = TOK_NUMBER;
						sscanf(lexbuf,
							   "%lf",
							   &(returned_val->number));
						returned_val->SetString(lexbuf);
						return returned_val;
					} else {
						position++;
						lexbuf[lexindex]=0;
						return error_token;
					}
					break;
					
			/* process first digit of a number after dot */
			case 6: if (Isdigit(inbuffer[position])) {		// DIGIT
						lexbuf[lexindex]=inbuffer[position];
						lexindex++;
						position++;
						state = 7;
					} else 
					if (Isalpha(inbuffer[position])) {		// Oops, was a identifier, after all
						lexbuf[lexindex]=inbuffer[position];
						lexindex++;
						position++;
						state = 3;
					} else {
						position++;
						lexbuf[lexindex]=0;
						return error_token;
					}
					break;
					
			/* process rest of number after dot */
			case 7:	if (Isdigit(inbuffer[position])) {		// DIGIT
						lexbuf[lexindex]=inbuffer[position];
						lexindex++;
						position++;
					} else
					if (Isalpha(inbuffer[position])) {		// Oops, was a identifier, after all
						lexbuf[lexindex]=inbuffer[position];
						lexindex++;
						position++;
						state = 3;
					} else
					if (Isblank(inbuffer[position]) ||
						Iscomment(inbuffer[position]) ||
						(inbuffer[position]==')')) {		// END OF DIGIT
						lexbuf[lexindex]=0; /* terminate string */
						returned_val->token_id = TOK_NUMBER;
						sscanf(lexbuf,
							   "%lf",
							   &(returned_val->number));
						returned_val->SetString(lexbuf);
						return returned_val;
					} else {
						position++;
						lexbuf[lexindex]=0;
						return error_token;
					}
					break;
					
			/* This processes the rest of a comment */
			case 10:if (Isnewline(inbuffer[position])) {	// NEW LINE
						state = 2;
					} else {
						position++;
					}
					break;
			
			default: state = 2;
					 break;
		}
	}
}


/******************************************************************/
//  Print a parsed parse tree

void Parser::PrintTree(void)
{
	*op << ">";
	for (int i=1;i<nesting;i++) *op << ">";
	returned_list->Print();
	*op << "\n";
}

/******************************************************************/
/* save and restore parser state for nesting                      */

void Parser::Save(char *inbuffer)
{
	ParserState *nps =  new ParserState(
							lookahead,
							line_no,
							cmd_buffer,
							scanpos,
							returned_val
						);
						
	lookahead 		= NULL;			// no lookahead
	line_no   		= 1;			// first line
	cmd_buffer		= inbuffer;		// new command buffer
	scanpos   		= 0;			// before first character
	returned_val	= new Token(op);	// a new returned token
	
	nesting++;
	ps_stack.Push(nps);
}

void Parser::Restore(void)
{
	assert(nesting>0);	// make sure no superflues unnests !
	
	ParserState *ops = (ParserState*) ps_stack.Pop();	// retrieve old state
	delete returned_val; 								// get rid of unneeded Token
	delete [] cmd_buffer;								// get rid of command text
		
	lookahead 		= ops->lookahead;					// restore old state
	line_no   		= ops->line_no;
	cmd_buffer		= ops->cmd_buffer;
	scanpos   		= ops->scanpos;
	returned_val	= ops->returned_val;
	
	nesting--;
	delete ops;											// delete old state packet
}

/******************************************************************/
/* constructs the parse tree using a lookahead parser             */

ConsCell *Parser::Parse(char *inbuffer)
{
	if (NULL==inbuffer) return NULL;
	
	Errorcode err=no_err;
	Save(inbuffer);

	returned_list = NULL;
	lookahead = Scan(cmd_buffer,scanpos);
	
	while ((err==no_err)&&(lookahead->token_id!=TOK_EOF)) {
		if 		(lookahead->token_id==TOK_LBRACKET) err = P_function();
		else if (lookahead->token_id==TOK_STRING)	err = P_atom();
		else if (lookahead->token_id==TOK_NUMBER)	err = P_atom();
		else if (lookahead->token_id==TOK_SYMBOL)	err = P_atom();
		else err = syntax_err;
		PrintTree();
		Eval(returned_list)->Print();
		*op << "\n";
	}
	
	if ((err==no_err)&&(lookahead->token_id==TOK_EOF)) {
		Match(TOK_EOF);
		Restore();
		return returned_list;;
	} else {
		op->ISERR2(iserr_syntax);
		*op << "\nNote: error " << (int)err << " in line " << (int)line_no;
		*op << " after:`" << lexbuf << "', at character: #";
		*op << (int)cmd_buffer[scanpos-1] << "\n", 
		Restore();
		return NULL;
	}
}

Errorcode Parser::P_function(void)
{
	int step=1;
	Errorcode err=internal_err;
	
	ConsCell	*llist=NULL;	// pointer to LISP list that is created by this function
	ConsCell	*e_llist=NULL;	// end of llist (point of insertion)
		
	while (1) {
		switch (step) {
			// process left bracket
			case 1:	err = Match(TOK_LBRACKET); break;
					
			// process interior between brackets
			case 2: while (err==no_err) {	/* these are arguments: keywords or objects */
						if      (lookahead->token_id==TOK_SYMBOL)	   	err = P_atom();
						else if (lookahead->token_id==TOK_NUMBER)   	err = P_atom();
						else if (lookahead->token_id==TOK_STRING)   	err = P_atom();
						else if (lookahead->token_id==TOK_LBRACKET) 	err = P_function(); 
						else if (lookahead->token_id==TOK_RBRACKET) 	break;
						else err = cmd_notrec_err;
						// Add new elements to the binary tree
						if (returned_list!=NULL) 
							if (e_llist==NULL) {
								llist = e_llist = new ConsCell(op,returned_list);
							} else {
								ConsCell *new_elem = new ConsCell(op,returned_list);
								e_llist->CDR = new_elem;
								e_llist = new_elem;
							}
					}
					break;
					
			// process the right bracket
			case 3: err = Match(TOK_RBRACKET);
					if (llist==NULL) returned_list = const_nil;	// was an empty list, same as nil
					else returned_list = llist;
					return err;
			default:err = internal_err; break;
		}
		
		if (err!=no_err) return err;
		step++;
	}
}


Errorcode Parser::P_atom(void)
{
	returned_list = new ConsCell(op,lookahead);
	Errorcode err=Match(lookahead->token_id); /* Match anything */
	return err;
}

Errorcode Parser::Match(TokenType tok)
{
	if (tok==lookahead->token_id) {
		lookahead = Scan(cmd_buffer,scanpos);
		if (lookahead->token_id==TOK_ERROR) return scan_err;
		return no_err;
	} else
		return match_err;
}

ConsCell	*Parser::Eval(ConsCell *expr)
{
	// check type of ConsCell to be evaluated
	returned_list=const_nil;
	if (expr==NULL) return returned_list;
	switch (expr->type) {
		case T_VOID:
			op->ISERR3(iserr_noeval," T_VOID");
			break;
		case T_NUMBER:				// number evaluates to itself
			returned_list=expr;
			break;
		case T_STRING:				// string evaluates to itself
			returned_list=expr;
			break;
		case T_LIST:				// use EvalCall to evaluate function call
			assert(expr->CAR);
			returned_list = EvalCall(expr->CAR,expr->CDR);
			break;
		case T_SYMBOL:				// find binding of symbol
			assert(expr->part.ste);
			if (expr->part.ste->type==T_UNBOUND) {
				op->ISERR3(iserr_noeval,expr->part.ste->lexem);
				returned_list = const_nil;
			} else {
				returned_list = expr->part.ste->binding;
			}
			break;
		case T_CONSTCELL:
			returned_list = expr;	// built-in constants evaluate to themselves
			break;
		case T_SUP:
			op->ISERR3(iserr_noeval," SUP not in head of a list");
			break;
		case T_SBUILTIN:
		case T_UBUILTIN:
		case T_EBUILTIN:
			op->ISERR3(iserr_noeval," BUILTIN not in head of a list");
			break;
		default:
			op->ISERR3(iserr_noeval," unknown thingy");
			break;
	};
	return returned_list;
}

ConsCell	*Parser::EvalCall(ConsCell *fct, ConsCell *args)
{
	// check some preconditions
	assert(fct);				// must not be NULL
	
	// initialize some local variables
	returned_list = const_nil;	// defaults to nil
	ConsCell *retval=NULL;
	ConsCell *argptr = args;	// running index to arguments
	ConsCell *evald_args=NULL;	// list of evaluated arguments
	ConsCell *prev_arg=NULL;	// previous arg in list
	Steppable *pt=NULL;
	TreeBase  *tb=NULL;
	unsigned int i,size;
	
	// determine type of head
	switch (fct->type) {
		case T_LIST:			//  ((...)  arg1 .. argn)
			assert(fct->CAR);
			if (fct->CAR->type==T_SYMBOL && fct->CAR->part.ste==const_lambdasym) {
				returned_list = EvalLambda(fct->CDR,args);
			} else {
				op->ISERR3(iserr_notyet," nested function calls");
			}
			break;
		case T_SUP:				//  (sup_head  arg1 .. argn)
			retval = fct;
			pt = fct->part.sup;
			assert(pt);
			size=pt->GetNoParameters();
			if (size>0) { // if it's a TreeBase
				// if it has arguments, then instantiate them now
				tb = (TreeBase*) pt;	// cast the pointer
				i=0;
				argptr = args;
				while (argptr&&(argptr!=const_nil)) {
					assert(argptr->type==T_LIST);
					if (i>=size) break;
					ConsCell *ncc = Eval(argptr->CAR);
					if (! ncc) {
						op->ISERR3(iserr_argtype," expecting a SUP / PSUP");
						retval = const_nil;
						break;
					}
					// now set the parameters, auto-convert if number or string
					if (ncc->type==T_SUP) 			tb->SetParameter(i, ncc->part.sup);
					else if (ncc->type==T_NUMBER) 	tb->SetParameter(i, new Number(ncc->part.number));
					else if (ncc->type==T_STRING) 	tb->SetParameter(i, new Text(ncc->part.string));
					else {
						op->ISERR3(iserr_argtype," expecting a SUP / PSUP");
						retval = const_nil;
						break;
					}
					i++;
					argptr=argptr->CDR;
				}
			}
			returned_list = retval;
			break;
		case T_SYMBOL:			//  (symbol  arg1 .. argn)
			assert(fct->part.ste);
			if (fct->part.ste->type==T_UNBOUND) {
				op->ISERR3(iserr_undeffct,fct->part.ste->lexem);
			} else {
				returned_list = EvalCall(fct->part.ste->binding,args);	// call with its binding as function
			}
			break;
		case T_SBUILTIN:		//  (sbi_head  arg1 .. argn), evaluating function
			// evaluate the arguments
			argptr = args;
			while (argptr&&(argptr!=const_nil)) {
				assert(argptr->type==T_LIST);
				ConsCell *ncc = new ConsCell(op,Eval(argptr->CAR),NULL); // replace with eval'd
				if (prev_arg) {
					prev_arg->CDR = ncc;
				} else {
					evald_args=ncc;
				}
				prev_arg = ncc;
				argptr=argptr->CDR;
			}
			// call the function
			assert(fct->part.builtin);
			returned_list = (*(fct->part.builtin))(this,evald_args);
			break;
		case T_UBUILTIN:		//  (ubi_head  arg1 .. argn), unevaluating function
			// call function with arguments unevaluated
			assert(fct->part.builtin);
			returned_list = (*(fct->part.builtin))(this,args);
			break;
		case T_EBUILTIN:		//  (ebi_head  arg1 .. argn), evaluating function
			// evaluate all arguments
			argptr = args;
			while (argptr&&(argptr!=const_nil)) {
				assert(argptr->type==T_LIST);
				ConsCell *ncc = new ConsCell(op,Eval(argptr->CAR),NULL); // replace with eval'd
				if (prev_arg) {
					prev_arg->CDR = ncc;
				} else {
					evald_args=ncc;
				}
				prev_arg = ncc;
				argptr=argptr->CDR;
			}
			// call function
			assert(fct->part.builtin);
			returned_list = (*(fct->part.builtin))(this,evald_args);
			break;
		default:
			op->ISERR3(iserr_nonfct,(int)fct->type);
			break;
	};
	
	return returned_list;
}

ConsCell	*Parser::EvalLambda(ConsCell *largs, ConsCell *args)
{
	assert(largs);
	
	// some inits and some checks
	ConsCell *locvar = largs->CAR;
	ConsCell *body   = largs->CDR;
	assert(locvar);
	if (locvar->type!=T_LIST && locvar!=const_nil) {
		op->ISERR3(iserr_lambda," first argument must be list of local vars");
		return const_nil;
	}
	returned_list = const_nil;	// return value defaults to nil
	
	// make list of evaluated arguments
	ConsCell *argptr = args;	// running index to arguments
	ConsCell *evald_args=NULL;	// list of evaluated arguments
	ConsCell *prev_arg=NULL;	// previous arg in list
	while (argptr&&(argptr!=const_nil)) {
		assert(argptr->type==T_LIST);
		ConsCell *ncc = new ConsCell(op,Eval(argptr->CAR),NULL); // replace with eval'd
		if (prev_arg) prev_arg->CDR = ncc;
		else		  evald_args=ncc;
		prev_arg = ncc;
		argptr=argptr->CDR;
	}
	
	// save current variable bindings and bind evaluated arguments to variables
	ConsCell *locvarptr = locvar;	// running index to local variables
	argptr = evald_args;			// running index to evaluated arguments
	while (argptr && locvarptr && argptr!=const_nil && locvarptr!=const_nil) {
		assert(locvarptr->CAR);
		assert(locvarptr->CAR->part.ste);
		assert(argptr->CAR);
		assert(locvarptr->CAR->type==T_SYMBOL);
		symtab->Push(locvarptr->CAR->part.ste, T_VARIABLE, argptr->CAR);
		argptr=argptr->CDR;
		locvarptr=locvarptr->CDR;
	}
	
	// check for same amount of arguments/local variables
	if ((argptr!=NULL&&argptr!=const_nil) ||
		(locvarptr!=NULL&&locvarptr!=const_nil)) {
		op->ISERR3(iserr_lambda," wrong # of args");
		return const_nil;
	}
		
	// evaluate each form in body
	ConsCell *bodyptr = body;
	while (bodyptr&&bodyptr!=const_nil) {
		returned_list = Eval(bodyptr->CAR);
		bodyptr=bodyptr->CDR;
	}
	
	// restore variable bindings
	locvarptr = locvar;	// running index to local variables
	while (locvarptr && locvarptr!=const_nil) {
		symtab->Pop(locvarptr->CAR->part.ste);
		locvarptr=locvarptr->CDR;
	}
	
	return returned_list;
}

/******************************************************************/
/******************************************************************/
/*                      Symbol table entries                      */

SymbolTableEntry::SymbolTableEntry(OperatorProxy *op_,char *lexem_, SymbolType t_, ConsCell *binding_):op(op_)
{
	Set(lexem_,t_,binding_);
}

SymbolTableEntry::SymbolTableEntry(OperatorProxy *op_,SymbolTableEntry *org):op(op_)
{
	Set(org);
}

void SymbolTableEntry::Push(SymbolType t_, ConsCell *binding_)
{
	pushed.Push(new SymbolTableEntry(op,this));
	Set(NULL,t_,binding_); // push without altering lexem
}

boolean SymbolTableEntry::Pop(SymbolType &t_, ConsCell *&binding_)
{
	if (empty) {			// if empty, dont pop
		t_=T_ERROR;
		binding_=NULL;
	} else {
		t_=type;			// set returned values
		binding_=binding;
		SymbolTableEntry *popped;
		if (pushed.Size()>0) 	{
			Set(popped=(SymbolTableEntry*)pushed.Pop());
			delete popped;
		} else
			empty=TRUE;
	}
	return empty;
}

void SymbolTableEntry::Set(SymbolTableEntry *org)
{
	if (org==NULL) 
		Set(NULL,T_ERROR,NULL,TRUE);
	else
		Set(org->lexem,org->type,org->binding,org->empty);
}

void SymbolTableEntry::Set(char *lexem_, SymbolType t_, ConsCell *binding_, boolean empty_)
{
	binding=binding_;
	type = t_;
	
	if (NULL==binding) 
		type = T_UNBOUND;	// if no binding, assume unbound in any case
	if (type==T_UNBOUND) {
		binding = new ConsCell(op);
		binding->type = T_SYMBOL;
		binding->part.ste = this;
	}
	
	if (lexem_) {
		lexem=NULL;
		SetLexem(lexem_);
	}
	empty=empty_;
}

SymbolTableEntry::~SymbolTableEntry()
{
	delete lexem;
	assert(binding);
	if (type==T_SYMBOL) delete binding;
	SymbolTableEntry *c;
	pushed.Reset();
	while (NULL!=(c=(SymbolTableEntry*)pushed.Iterate())) delete c;
}

void SymbolTableEntry::SetLexem(char *txt)
{
	word len;
	if (txt==NULL) return;
	if ((len=(word)strlen(txt))>MAX_NAME_LEN) {
		op->ISERR2(iserr_stringsize);
		return;
	}
	
	delete lexem;
	lexem = new char [len+1];
	strcpy(lexem,txt);
}

/******************************************************************/
/******************************************************************/
/*                      Symbol table code                         */

SymbolTable::~SymbolTable(void)
{
	symbols.Reset();
	SymbolTableEntry *ste;
	while ((ste=(SymbolTableEntry*)symbols.Iterate())!=NULL) delete ste;
}

SymbolTableEntry *SymbolTable::Append(char *lextext, SymbolType t_, ConsCell *binding_)
{
	SymbolTableEntry *found;
	if ((found=Lookup(lextext))!=NULL) return found;	// check: does it already exist?
	SymbolTableEntry *nptr= new SymbolTableEntry(op,lextext,t_,binding_); // generate new
	symbols.Append(nptr); // add to table
	return nptr;
}

SymbolTableEntry *SymbolTable::Replace(char *lextext, SymbolType t_, ConsCell *binding_)
{
	SymbolTableEntry *found;
	if ((found=Lookup(lextext))==NULL) return NULL;	// check: does it already exist?
	
	found->type		= t_;
	found->binding	= binding_;
	return found;
}

SymbolTableEntry *SymbolTable::Push(char *lextext, SymbolType t_, ConsCell *binding_)
{
	SymbolTableEntry *found=Lookup(lextext);
	
	if (found!=NULL) {	// push existing
		found->Push(t_,binding_);
		return found;
	} else {			// create new one
		SymbolTableEntry *nptr= new SymbolTableEntry(op,lextext,t_,binding_); // generate new
		symbols.Append(nptr); // add to table
		return nptr;
	}
}

SymbolTableEntry *SymbolTable::Push(SymbolTableEntry *ste_, SymbolType t_, ConsCell *binding_)
{
	if (ste_==NULL) return NULL;
	
	ste_->Push(t_,binding_);
	return ste_;
}

SymbolTableEntry *SymbolTable::Pop(SymbolTableEntry *ste_)
{
	SymbolType	rec_type;
	ConsCell	*rec_binding;
	return Pop(ste_,rec_type,rec_binding);
}

SymbolTableEntry *SymbolTable::Pop(SymbolTableEntry *ste_, SymbolType &t_, ConsCell *&binding_)
{
	if (ste_==NULL) return NULL;	// does exist?
	
	if (ste_->Pop(t_,binding_)) {	// retrieve data, if TRUE returned, then..
		delete (SymbolTableEntry *)symbols.Remove(ste_);// ..remove if no longer valid
		return NULL;
	} else {
		return ste_;
	}
}

SymbolTableEntry *SymbolTable::Pop(char *lextext, SymbolType &t_, ConsCell *&binding_)
{
	return Pop(Lookup(lextext),t_,binding_);
}

SymbolTableEntry *SymbolTable::AppendBuiltin(char *lextext, Builtin builtin_, ConsCellType t_)
{
	SymbolTableEntry *found,*nptr;
	if ((found=Lookup(lextext))!=NULL) return found;	// check: does it already exist?
	if (t_!=T_SBUILTIN && t_!=T_EBUILTIN && t_!=T_UBUILTIN) {	// if it is not a builtin function, append it as normal symbol, UNBOUND
		return Append(lextext);
	} else {
		ConsCell *ncc = new ConsCell(op);					// make new binding
		ncc->type = t_;									// set type
		assert(builtin_);					
		ncc->part.builtin = builtin_;					// link to binary function
		nptr= new SymbolTableEntry(op,lextext,T_CONST,ncc);// generate new entry
		symbols.Append(nptr); 							// add to table
		return nptr;
	}
	
}

SymbolTableEntry *SymbolTable::AppendTree(char *lextext, TreeBase *tree_)
{
	assert(tree_);	
	SymbolTableEntry *found,*nptr;
	if ((found=Lookup(lextext))!=NULL) return found;	// check: does it already exist?
	ConsCell *ncc = new ConsCell(op);		// make new binding
	ncc->type = T_SUP;							// set type
	ncc->part.sup = tree_;					// link to allocated Steppable tree
	nptr= new SymbolTableEntry(op,lextext,T_CONST,ncc);// generate new entry
	symbols.Append(nptr); 							// add to table
	return nptr;
}

SymbolTableEntry *SymbolTable::Lookup(char *lextext)
{
	symbols.Reset();
	SymbolTableEntry *ste;
	while ((ste=(SymbolTableEntry*)symbols.Iterate())!=NULL) {
		if (strcmp(lextext,ste->lexem)==0) return ste;
	}
	return NULL;
}

void SymbolTable::PrintAll(void)
{
	int ctr=0;
	
	*op << "\nSymbol Table Contents\n";
	*op <<   "=====================\n\n";
	*op <<   "   No   Type       Lexem  /  Value <ConsCell Type>\n";
	*op <<   "-----------------------------------------------------------------------------\n";
	
	symbols.Reset();
	SymbolTableEntry	*ste;
	while ((ste=(SymbolTableEntry*)symbols.Iterate())!=NULL) {
		*op << ctr++;
		switch (ste->type) {
		case T_UNBOUND:
			*op << " UNBOUND   ";
			*op << ste->lexem;
			break;
		case T_CONST:
			*op << " CONSTANT  ";
			*op << ste->lexem << "  /  ";
			ste->binding->Print();
			*op << " <" << ste->binding->type << ">";
			break;
		case T_VARIABLE:
			*op << " VARIABLE  ";
			*op << ste->lexem << "  /  ";
			ste->binding->Print();
			*op << " <" << ste->binding->type << ">";
			break;
		default:
			*op << " UNKNOWN,  STE TYPE: " << ste->type;
			break;
		};
		*op << "\n";
	}

	*op <<   "-----------------------------------------------------------------------------\n";
}


////////////////////////////////////////////////////////////////////////////
//
//   Argument checkers
//
////////////////////////////////////////////////////////////////////////////

#define ARGNO(n,mi,ma)  ArgNo(__FILE__,__LINE__,pars,(n),args,(mi),(ma))
#define ARGTYPE(n,nr,t) ArgType(__FILE__,__LINE__,pars,(n),args,(nr),(t))
#define ARGTYPE_N(nr,t) ArgTypeN(__FILE__,__LINE__,pars,args,(nr),(t))

int	Parser::ArgNo(char *sf, int sln, Parser *pars, char *name, ConsCell *args, int min, int max)
{
	int count = 0;
	ConsCell *argptr=args;
	while (argptr) {
		if (argptr->type!=T_LIST) {
			if (name) pars->op->ReportError(sf,sln,iserr_argformat,name);
			else      pars->op->ReportError(sf,sln,iserr_argformat,"<unknown>");
			break;
		}
		if (argptr->CAR) count++;
		else {
			if (name) pars->op->ReportError(sf,sln,iserr_argnull,name);
			else      pars->op->ReportError(sf,sln,iserr_argnull,"<unknown>");
			break;
		}
		argptr = argptr->CDR;
	}
	if (count>=min) {
		if (max>=0 && count>max) {
			if (name) pars->op->ReportError(sf,sln,iserr_arg2many,name);
			else      pars->op->ReportError(sf,sln,iserr_arg2many,"<unknown>");
			return -1;
		}
		return count;
	}
	if (name) pars->op->ReportError(sf,sln,iserr_arg2few,name);
	else      pars->op->ReportError(sf,sln,iserr_arg2few,"<unknown>");
	return -1;
}

ConsCell *Parser::ArgType(char *sf, int sln, Parser *pars, char *name, ConsCell *args, int no, ConsCellType t)
{
	int count=0;
	ConsCell *argptr=args;
	while (argptr) {
		if (argptr->type!=T_LIST) {
			if (name) pars->op->ReportError(sf,sln,iserr_argformat,name);
			else      pars->op->ReportError(sf,sln,iserr_argformat,"<unknown>");
			return NULL;
		}
		count++;
		if (count>=no) {
			if (argptr->CAR==NULL) {
				if (name) pars->op->ReportError(sf,sln,iserr_argnull,name);
				else      pars->op->ReportError(sf,sln,iserr_argnull,"<unknown>");
				return NULL;
			}
			if (t==T_LIST) {
				// if a list is wanted, check for empty list == nil  (which is ok)
				if (pars->const_nil==argptr->CAR) return argptr->CAR;
				// check further without returning
			}
			if (argptr->CAR->type==t) {
				return argptr->CAR;
			} else {
				if (name) {pars->op->ReportError(sf,sln,iserr_argtype,name);
				*(pars->op) << "Wanted type " << (double)t << " got type " << (double)argptr->CAR->type << "\n";}
				else      pars->op->ReportError(sf,sln,iserr_argtype,"<unknown>");
				return NULL;
			}
		}
		argptr = argptr->CDR;
	}
	if (name) pars->op->ReportError(sf,sln,iserr_argnoex,name);
	else      pars->op->ReportError(sf,sln,iserr_argnoex,"<unknown>");
	return NULL;
}

// Same function as above, but no reports to the operator

ConsCell *Parser::ArgTypeN(char *sf, int sln, Parser *pars, ConsCell *args, int no, ConsCellType t)
{
	int count=0;
	sln = 0;		// this is to silence the compiler warning about not using sln
	sf  = NULL;		//   " dito "
	
	ConsCell *argptr=args;
	while (argptr) {
		if (argptr->type!=T_LIST) {
			return NULL;
		}
		count++;
		if (count>=no) {
			if (argptr->CAR==NULL) {
				return NULL;
			}
			if (t==T_LIST) {
				// if a list is wanted, check for empty list == nil  (which is ok)
				if (pars->const_nil==argptr->CAR) return argptr->CAR;
				// check further without returning
			}
			if (argptr->CAR->type==t) {
				return argptr->CAR;
			} else {
				return NULL;
			}
		}
		argptr = argptr->CDR;
	}
	return NULL;
}

////////////////////////////////////////////////////////////////////////////
//
//   Built-ins go here
//
////////////////////////////////////////////////////////////////////////////
ConsCell *Parser::EBI_mul(Parser *pars, ConsCell *args)
{
	double res=1.0;	// result of computation
	ConsCell *argptr = args;
	
	// multiply all arguments
	while (argptr) {
		assert(argptr->CAR);
		if (argptr->CAR->type==T_NUMBER) {	// check whether is NUMBER
			res *= argptr->CAR->part.number;
		} else {										// error, no number !
			pars->op->ISERR3(iserr_argtype," in multiplication");
			return pars->const_nil;
		}
		argptr = argptr->CDR;
	}
	ConsCell *retcell = new ConsCell(pars->op);
	retcell->type = T_NUMBER;
	retcell->part.number=res;
	return retcell;
}

ConsCell *Parser::EBI_add(Parser *pars, ConsCell *args)
{
	double res=0.0;	// result of computation
	ConsCell *argptr = args;
	
	// add all arguments
	while (argptr) {
		assert(argptr->CAR);
		if (argptr->CAR->type==T_NUMBER) {	// check whether is NUMBER
			res += argptr->CAR->part.number;
		} else {										// error, no number !
			pars->op->ISERR3(iserr_argtype," in addition");
			return pars->const_nil;
		}
		argptr = argptr->CDR;
	}
	ConsCell *retcell = new ConsCell(pars->op);
	retcell->type = T_NUMBER;
	retcell->part.number=res;
	return retcell;
}

ConsCell *Parser::EBI_sub(Parser *pars, ConsCell *args)
{
	double res=0.0;	// result of computation
	ConsCell *argptr = args;
	
	// check for existence of first argument
	if (argptr==NULL) {
		pars->op->ISERR3(iserr_arg2few," in subtraction");
		return pars->const_nil;
	}
	
	// set first argument
	assert(argptr->CAR);
	if (argptr->CAR->type==T_NUMBER) {	// check whether is NUMBER
		res = argptr->CAR->part.number;	// get the first number
	} else {										// error, no number !
		pars->op->ISERR3(iserr_argtype," in subtraction (first arg)");
		return pars->const_nil;
	}
	argptr = argptr->CDR;
	
	// subtract all other arguments
	while (argptr) {
		assert(argptr->CAR);
		if (argptr->CAR->type==T_NUMBER) {	// check whether is NUMBER
			res -= argptr->CAR->part.number;
		} else {										// error, no number !
			pars->op->ISERR3(iserr_argtype," in subtraction");
			return pars->const_nil;
		}
		argptr = argptr->CDR;
	}
	ConsCell *retcell = new ConsCell(pars->op);
	retcell->type = T_NUMBER;
	retcell->part.number=res;
	return retcell;
}

ConsCell *Parser::EBI_div(Parser *pars, ConsCell *args)
{
	double res=1.0;	// result of computation
	ConsCell *argptr = args;
	
	// check for existence of first argument
	if (argptr==NULL) {
		pars->op->ISERR3(iserr_arg2few," in division");
		return pars->const_nil;
	}
	
	// set first argument
	assert(argptr->CAR);
	if (argptr->CAR->type==T_NUMBER) {	// check whether is NUMBER
		if (argptr->CAR->part.number==0.0) {
			pars->op->ISERR2(iserr_div0);
			return pars->const_nil;
		}
		res = argptr->CAR->part.number;
	} else {										// error, no number !
		pars->op->ISERR3(iserr_argtype," in division (first arg)");
		return pars->const_nil;
	}
	argptr = argptr->CDR;
	
	// divide by all other arguments
	while (argptr) {
		assert(argptr->CAR);
		if (argptr->CAR->type==T_NUMBER) {	// check whether is NUMBER
			if (argptr->CAR->part.number==0.0) {
				pars->op->ISERR2(iserr_div0);
				return pars->const_nil;
			}
			res /= argptr->CAR->part.number;
		} else {										// error, no number !
			pars->op->ISERR3(iserr_argtype," in division");
			return pars->const_nil;
		}
		argptr = argptr->CDR;
	}
	ConsCell *retcell = new ConsCell(pars->op);
	retcell->type = T_NUMBER;
	retcell->part.number=res;
	return retcell;
}

ConsCell *Parser::EBI_eval(Parser *pars, ConsCell *args)
{
	if (ARGNO("eval",1,1)<0)	return pars->const_nil;
	return pars->Eval(args->CAR);
}

ConsCell *Parser::EBI_print(Parser *pars, ConsCell *args)
{
	if (ARGNO("print",1,1)<0)	return pars->const_nil;
	assert(args->CAR);
	args->CAR->PrintLn();
	return args->CAR;
}

ConsCell *Parser::EBI_list(Parser *pars, ConsCell *args)
{
	if (args==NULL) return pars->const_nil;
	return args;
}

ConsCell *Parser::EBI_listp(Parser *pars, ConsCell *args)
{
	if (ARGNO("listp",1,1)<0)	return pars->const_nil;

	if (args->CAR==pars->const_nil) 
		return pars->const_true;

	if (args->CAR->type==T_LIST)
		return pars->const_true;
	else
		return pars->const_nil;
}

ConsCell *Parser::EBI_atomp(Parser *pars, ConsCell *args)
{
	if (ARGNO("atomp",1,1)<0)	return pars->const_nil;

	if (args->CAR==pars->const_nil) 
		return pars->const_true;

	if (args->CAR->type==T_LIST)
		return pars->const_nil;
	else
		return pars->const_true;
}

ConsCell *Parser::EBI_length(Parser *pars, ConsCell *args)
{
	ConsCell *argptr1=NULL;
	if (ARGNO("length",1,1)<0)	return pars->const_nil;
	if (NULL==(argptr1=ARGTYPE("length",1,T_LIST))) return pars->const_nil;

	unsigned int cntr=0;
	ConsCell *curelem = argptr1;
	
	while (curelem && curelem!=pars->const_nil) {
		if (curelem->type!=T_LIST) break;
		cntr++;
		curelem = curelem->CDR;
	}
	
	return new ConsCell(pars->op,(double)cntr);
}

ConsCell *Parser::EBI_car(Parser *pars, ConsCell *args)
{
	ConsCell *argptr1=NULL;
	if (ARGNO("car",1,1)<0)	return pars->const_nil;
	if (NULL==(argptr1=ARGTYPE("car",1,T_LIST))) return pars->const_nil;

	if (argptr1==pars->const_nil) return argptr1;
	
	if (argptr1->CAR)
		return argptr1->CAR;
	else
		return pars->const_nil;
}

ConsCell *Parser::EBI_cdr(Parser *pars, ConsCell *args)
{
	ConsCell *argptr1=NULL;
	if (ARGNO("cdr",1,1)<0)	return pars->const_nil;
	if (NULL==(argptr1=ARGTYPE("cdr",1,T_LIST))) return pars->const_nil;
	
	if (argptr1==pars->const_nil) return argptr1;
	
	if (argptr1->CDR)
		return argptr1->CDR;
	else
		return pars->const_nil;
}


////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////

ConsCell *Parser::UBI_print_symtab(Parser *pars, ConsCell *args)	
{
	if (ARGNO("print-symtab",0,0)<0)	return pars->const_nil;
	pars->symtab->PrintAll();
	return pars->const_true;
}

ConsCell *Parser::UBI_print_docklist(Parser *pars, ConsCell *args)	
{
	if (ARGNO("print-docklist",0,0)<0)	return pars->const_nil;
	pars->sup_mgr->PrintDocks();
	return pars->const_true;
}

ConsCell *Parser::UBI_quote(Parser *pars, ConsCell *args)	
{
	ConsCell *argptr1=NULL,*argptr2=NULL;
	if (ARGNO("quote",1,1)<0)	return pars->const_nil;
	return args->CAR;
}

ConsCell *Parser::UBI_setq(Parser *pars, ConsCell *args)
{
	ConsCell *argptr1=NULL;
	if (ARGNO("setq",2,-1)<0)	return pars->const_nil;
	if (NULL==(argptr1=ARGTYPE("setq",1,T_SYMBOL))) return pars->const_nil;

	// check type of first argument (receiver)
	assert(args->CAR->part.ste);
	if (args->CAR->part.ste->type == T_CONST) {
		pars->op->ISERR3(iserr_argtype, " cannot setq constants");
		return pars->const_nil;
	}
	
	// eval and assign second argument
	args->CAR->part.ste->type = T_VARIABLE;
	return (args->CAR->part.ste->binding = 
		pars->Eval(args->CDR->CAR));
}


// (defun name (args) body1 body2 ... ) is translated into equivalent:
// (setq  name (lambda (args) body1 body2...) )
ConsCell *Parser::UBI_defun(Parser *pars, ConsCell *args)
{
	ConsCell *argptr1=NULL,*argptr2=NULL;
	if (ARGNO("defun",3,3)<0)	return pars->const_nil;
	if (NULL==(argptr1=ARGTYPE("defun",1,T_SYMBOL))) return pars->const_nil;
	if (NULL==(argptr2=ARGTYPE("defun",2,T_LIST))) return pars->const_nil;
	
	// check for constant symbols
	assert(args->CAR->part.ste);
	if (args->CAR->part.ste->type == T_CONST) {
		pars->op->ISERR3(iserr_argtype, "  cannot define constants as functions");
		return pars->const_nil;
	}
	
	// prepend a "lambda", assign to second argument
	args->CAR->part.ste->type = T_VARIABLE;
	args->CAR->part.ste->binding = 
		new ConsCell(pars->op,new ConsCell(pars->op,pars->const_lambdasym),args->CDR);
	return args->CAR;
}

// (load "file-name")
ConsCell *Parser::UBI_load(Parser *pars, ConsCell *args)
{
	ConsCell *argptr1=NULL;
	char *cb;

	if (ARGNO("load",1,1)<0)	return pars->const_nil;
	if (NULL==(argptr1=ARGTYPE("load",1,T_STRING))) return pars->const_nil;

	// return nil, if nil filename
	if (args->CAR==pars->const_nil) return pars->const_nil;
	
	// open the file
	if ((cb = pars->op->LoadFile(argptr1->part.string))==NULL) return pars->const_nil;
	
	// now, parse and evaluate the stuff (ATTENTION: Parse deletes the buffer after using it!!)
	return pars->Parse(cb);
}

// (exit)
ConsCell *Parser::UBI_exit(Parser *pars, ConsCell *args)
{
	ConsCell *argptr1=NULL;
	const int retries=10;

	if (ARGNO("load",0,0)<0)	return pars->const_nil;

	// try to stop the robot
	int l;
	for (l=0; l<retries; l++) {
		pars->robot->SetMode(rm_periphery_off);
		if (pars->robot->UpdateSensors()) exit(0); // exit, if sent ok
	}
	exit(0);		// could not stop robot, now exiting anyway
}


////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////

ConsCell *Parser::SBI_waypoint(Parser *pars, ConsCell *args)	
{
	ConsCell *argptr1=NULL,*argptr2=NULL;
	if (ARGNO("Waypoint",2,2)<0)	return pars->const_nil;
	if (NULL==(argptr1=ARGTYPE("Waypoint",1,T_NUMBER))) return pars->const_nil;
	if (NULL==(argptr2=ARGTYPE("Waypoint",2,T_NUMBER))) return pars->const_nil;
	ConsCell *retval 	= new ConsCell(pars->op);
	retval->type 		= T_SUP;
	retval->part.sup 	= new Waypoint((float)argptr1->part.number,(float)argptr2->part.number);
	return retval;
}

////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////

// SUP built-ins take cons-cells with pointers to Steppables as arguments
// (among others) and return a cons-cell with a pointer to a Steppable.
// I.e. they construct a larger UP from smaller ones.
// This is passed to the MERGER which links it with the rest of the UP.


ConsCell *Parser::SBI_noop(Parser *pars, ConsCell *args)	
{
	if (ARGNO("Noop",0,0)<0)	return pars->const_nil;
	ConsCell *retval 	= new ConsCell(pars->op);
	retval->type 		= T_SUP;
	retval->part.sup 	= new Noop;
	return retval;
}

ConsCell *Parser::SBI_achieve(Parser *pars, ConsCell *args)	
{
	int nargs=-1;
	ConsCell *argptr1=NULL,*argptr2=NULL, *argptr3=NULL;
	if ((nargs=ARGNO("Achieve",3,3))<0)	return pars->const_nil;
	if (NULL==(argptr1=ARGTYPE("Achieve",1,T_SUP))) 	return pars->const_nil;
	if (NULL==(argptr2=ARGTYPE("Achieve",2,T_SUP))) 	return pars->const_nil;
	if (NULL==(argptr3=ARGTYPE("Achieve",3,T_STRING))) 	return pars->const_nil;
	if (NULL==pars->sup_mgr->AchieveWithBlock(
			argptr1->part.sup,
			argptr2->part.sup,
			argptr3->part.string)) {
		pars->op->ISERR3(iserr_nodock,"not inserted.");
		return pars->const_nil;
	}
	return args->CAR;
}

ConsCell *Parser::SBI_do(Parser *pars, ConsCell *args)	
{
	int nargs=-1;
	ConsCell *argptr1,*argptr2=NULL;
	if ((nargs=ARGNO("Do",1,2))<0)	return pars->const_nil;
	if (NULL==(argptr1=ARGTYPE("Do",1,T_SUP))) 	return pars->const_nil;
	if (nargs>1) {
		// more than one param: 2nd is name of dock
		if (NULL==(argptr2=ARGTYPE("Do",2,T_STRING))) 	return pars->const_nil;
		if (NULL==pars->sup_mgr->InsertBlock(argptr1->part.sup,argptr2->part.string)) {
			pars->op->ISERR3(iserr_nodock,"not inserted.");
			return pars->const_nil;
		}
	} else {
		// just one param: use default dock
		if (! pars->sup_mgr->InsertBlock(argptr1->part.sup)) {
			pars->op->ReportError(__FILE__,__LINE__,iserr_arg2few,"Do");
			return pars->const_nil;
		}
	}
	return args->CAR;
}

ConsCell *Parser::SBI_stop(Parser *pars, ConsCell *args)	
{
	int nargs=-1; 
	ConsCell *argptr1=NULL;
	if ((nargs=ARGNO("Stop",1,1))<0)	return pars->const_nil;
	if (NULL==(argptr1=ARGTYPE("Stop",1,T_SUP))) return pars->const_nil;
	pars->sup_mgr->RemoveBlock(argptr1->part.sup);
	return args->CAR;
}

ConsCell *Parser::SBI_every(Parser *pars, ConsCell *args)	
{
	ConsCell *argptr1=NULL,*argptr2=NULL;
	if (ARGNO("Every",2,2)<0)	return pars->const_nil;
	if (NULL==(argptr1=ARGTYPE("Every",1,T_NUMBER))) return pars->const_nil;
	if (NULL==(argptr2=ARGTYPE("Every",2,T_SUP))) 	return pars->const_nil;
	ConsCell *retval 	= new ConsCell(pars->op);
	retval->type 		= T_SUP;
	retval->part.sup 	= new Every((unsigned long)argptr1->part.number,argptr2->part.sup);
	return retval;
}

ConsCell *Parser::SBI_if(Parser *pars, ConsCell *args)	
{
	int nargs=-1;
	ConsCell *argptr1,*argptr2,*argptr3=NULL,*argptr4=NULL;
	
	if ((nargs=ARGNO("If",2,4))<0)	return pars->const_nil;
	if (NULL==(argptr1=ARGTYPE("If",1,T_SUP))) 	return pars->const_nil;
	if (NULL==(argptr2=ARGTYPE("If",2,T_SUP))) 	return pars->const_nil;
	assert(argptr1&&argptr2);
	if (nargs>2) 
		if (NULL==(argptr3=ARGTYPE("If",3,T_SUP))) 	return pars->const_nil;
	if (nargs>3) 
		if (NULL==(argptr4=ARGTYPE("If",4,T_NUMBER)))	return pars->const_nil;
	
	ConsCell *retval 	= new ConsCell(pars->op);
	retval->type 		= T_SUP;
	if (nargs>3) {
		assert(argptr3&&argptr4);
		retval->part.sup 	= new If(	argptr1->part.sup,
										argptr2->part.sup,
										argptr3->part.sup,
										argptr4->part.number);
	} else if (nargs>2) {
		assert(argptr3);
		retval->part.sup 	= new If(	argptr1->part.sup,
										argptr2->part.sup,
										argptr3->part.sup);
	} else {
		retval->part.sup 	= new If(	argptr1->part.sup,
										argptr2->part.sup);
	}
	
	return retval;
}

ConsCell *Parser::SBI_ramp(Parser *pars, ConsCell *args)	
{
	int nargs=-1;
	ConsCell *argptr1=NULL;
	
	if ((nargs=ARGNO("Ramp",1,1))<0)	return pars->const_nil;
	if (NULL==(argptr1=ARGTYPE("Ramp",1,T_NUMBER)))	return pars->const_nil;
	assert(argptr1);
	
	ConsCell *retval 	= new ConsCell(pars->op);
	retval->type 		= T_SUP;
	retval->part.sup 	= new Ramp(argptr1->part.number);
	
	return retval;
}

ConsCell *Parser::SBI_taction(Parser *pars, ConsCell *args)	
{
	int nargs=-1;
	ConsCell *argptr1=NULL;
	
	if ((nargs=ARGNO("Taction",1,1))<0)	return pars->const_nil;
	if (NULL==(argptr1=ARGTYPE("Taction",1,T_NUMBER)))	return pars->const_nil;
	assert(argptr1);
	
	ConsCell *retval 	= new ConsCell(pars->op);
	retval->type 		= T_SUP;
	retval->part.sup 	= new TestAction(argptr1->part.number);
	
	return retval;
}

ConsCell *Parser::SBI_ttester(Parser *pars, ConsCell *args)	
{
	int nargs=-1;
	ConsCell *argptr1=NULL;
	
	if ((nargs=ARGNO("Ttester",1,1))<0)	return pars->const_nil;
	if (NULL==(argptr1=ARGTYPE("TTester",1,T_NUMBER)))	return pars->const_nil;
	assert(argptr1);
	
	ConsCell *retval 	= new ConsCell(pars->op);
	retval->type 		= T_SUP;
	retval->part.sup 	= new TestTester(argptr1->part.number);
	
	return retval;
}

ConsCell *Parser::SBI_pulsetrain(Parser *pars, ConsCell *args)	
{
	int nargs=-1;
	ConsCell *argptr1=NULL,*argptr2=NULL;
	
	if ((nargs=ARGNO("Pulsetrain",2,2))<0)	return pars->const_nil;
	if (NULL==(argptr1=ARGTYPE("Pulsetrain",1,T_NUMBER)))	return pars->const_nil;
	if (NULL==(argptr2=ARGTYPE("Pulsetrain",2,T_NUMBER)))	return pars->const_nil;
	assert(argptr1&&argptr2);
	
	ConsCell *retval 	= new ConsCell(pars->op);
	retval->type 		= T_SUP;
	retval->part.sup 	= new Pulsetrain(argptr1->part.number,argptr2->part.number);
	
	return retval;
}

ConsCell *Parser::SBI_loop(Parser *pars, ConsCell *args)	
{
	int nargs=-1;
	ConsCell *argptr1,*argptr2=NULL,*argptr3=NULL,*argptr4=NULL;
	
	if ((nargs=ARGNO("Loop",1,4))<0)	return pars->const_nil;
	if (NULL==(argptr1=ARGTYPE("Loop",1,T_SUP))) 	return pars->const_nil;
	assert(argptr1);
	if (nargs>1)
		if (NULL==(argptr2=ARGTYPE("Loop",2,T_SUP))) 	return pars->const_nil;
	if (nargs>2) 
		if (NULL==(argptr3=ARGTYPE("Loop",3,T_SUP))) 	return pars->const_nil;
	if (nargs>3) 
		if (NULL==(argptr4=ARGTYPE("Loop",4,T_NUMBER)))	return pars->const_nil;
	
	ConsCell *retval 	= new ConsCell(pars->op);
	retval->type 		= T_SUP;
	if (nargs>3) {
		assert(argptr2&&argptr3&&argptr4);
		retval->part.sup 	= new Loop(	argptr1->part.sup,
										argptr2->part.sup,
										argptr3->part.sup,
										argptr4->part.number);
	} if (nargs>2) {
		assert(argptr2&&argptr3);
		retval->part.sup 	= new Loop(	argptr1->part.sup,
										argptr2->part.sup,
										argptr3->part.sup);
	} if (nargs>1) {
		assert(argptr2);
		retval->part.sup 	= new Loop(	argptr1->part.sup,
										argptr2->part.sup);
	} else {
		retval->part.sup 	= new Loop(	argptr1->part.sup);
	}
	
	return retval;
}

ConsCell *Parser::SBI_sequence(Parser *pars, ConsCell *args)	
{
	int nargs=-1;
	ConsCell *argptr1,*argptr2=NULL,*argptr3=NULL,*argptrA=NULL,*argptrB=NULL;
	
	if ((nargs=ARGNO("Sequence",4,-1))<0)	return pars->const_nil;
	
	if (NULL==(argptr1=ARGTYPE("Sequence",1,T_SUP))) 	return pars->const_nil;
	if (NULL==(argptr2=ARGTYPE("Sequence",2,T_SUP))) 	return pars->const_nil;
	if (NULL==(argptr3=ARGTYPE("Sequence",3,T_SUP))) 	return pars->const_nil;
	assert(argptr1&&argptr2&&argptr3);
	
	ConsCell *retval 	= new ConsCell(pars->op);
	retval->type 		= T_SUP;
	retval->part.sup 	= new Sequence(	argptr1->part.sup,
										argptr2->part.sup,
										argptr3->part.sup);
										
	for (int ac=4;ac<=nargs;ac+=2) {
		if (NULL==(argptrA=ARGTYPE("Sequence",ac,T_SUP))) 	return pars->const_nil;
		if (ac<nargs) {
			if (NULL==(argptrB=ARGTYPE("Sequence",ac+1,T_SUP))) return pars->const_nil;
			((Sequence*)retval->part.sup)->Append(argptrA->part.sup,argptrB->part.sup);
		} else {
			((Sequence*)retval->part.sup)->Append(argptrA->part.sup,NULL);
		}
	}
	
	return retval;
}

ConsCell *Parser::SBI_switch(Parser *pars, ConsCell *args)	
{
	int nargs=-1;
	ConsCell *argptr1,*argptr2,*argptrA=NULL,*argptrB=NULL;
	
	if ((nargs=ARGNO("Switch",4,-1))<0)	return pars->const_nil;
	
	if (NULL==(argptr1=ARGTYPE("Switch",1,T_NUMBER))) 	return pars->const_nil;
	if (NULL==(argptr2=ARGTYPE("Switch",2,T_SUP))) 	return pars->const_nil;
	assert(argptr1&&argptr2);
	
	ConsCell *retval 	= new ConsCell(pars->op);
	retval->type 		= T_SUP;
	retval->part.sup 	= new Switch(argptr1->part.number,argptr2->part.sup);
										
	for (int ac=3;ac<=nargs;ac+=2) {
		if (NULL==(argptrA=ARGTYPE("Switch",ac,T_SUP))) 	return pars->const_nil;
		if (ac<nargs) {
			if (NULL==(argptrB=ARGTYPE("Switch",ac+1,T_SUP))) return pars->const_nil;
			((Switch*)retval->part.sup)->Append(argptrA->part.sup,argptrB->part.sup);
		} else {
			((Switch*)retval->part.sup)->Append(argptrA->part.sup,NULL);
		}
	}
	
	return retval;
}

ConsCell *Parser::SBI_parallel(Parser *pars, ConsCell *args)	
{
	int nargs=-1;
	ConsCell *argptr=NULL;
	
	if ((nargs=ARGNO("Parallel",1,-1))<0)	return pars->const_nil;
	
	ConsCell *retval 	= new ConsCell(pars->op);
	retval->type 		= T_SUP;
	retval->part.sup 	= new Parallel();
										
	for (int ac=1;ac<=nargs;ac++) {
		if (NULL==(argptr=ARGTYPE("Parallel",ac,T_SUP))) 	return pars->const_nil;
		((Parallel*)retval->part.sup)->Append(argptr->part.sup);
	}
	
	return retval;
}

ConsCell *Parser::SBI_and(Parser *pars, ConsCell *args)	
{
	int nargs=-1;
	ConsCell *argptr=NULL;
	
	if ((nargs=ARGNO("And",1,-1))<0)	return pars->const_nil;
	
	ConsCell *retval 	= new ConsCell(pars->op);
	retval->type 		= T_SUP;
	retval->part.sup 	= new And(NULL,NULL);
										
	for (int ac=1;ac<=nargs;ac++) {
		if (NULL==(argptr=ARGTYPE("And",ac,T_SUP))) 	return pars->const_nil;
		((And*)retval->part.sup)->Append(argptr->part.sup);
	}
	
	return retval;
}

ConsCell *Parser::SBI_or(Parser *pars, ConsCell *args)	
{
	int nargs=-1;
	ConsCell *argptr=NULL;
	
	if ((nargs=ARGNO("Or",1,-1))<0)	return pars->const_nil;
	
	ConsCell *retval 	= new ConsCell(pars->op);
	retval->type 		= T_SUP;
	retval->part.sup 	= new Or(NULL,NULL);
										
	for (int ac=1;ac<=nargs;ac++) {
		if (NULL==(argptr=ARGTYPE("Or",ac,T_SUP))) 	return pars->const_nil;
		((Or*)retval->part.sup)->Append(argptr->part.sup);
	}
	
	return retval;
}

ConsCell *Parser::SBI_dock(Parser *pars, ConsCell *args)	
{
	int nargs=-1;
	ConsCell *argptr1,*argptr2,*argptr3,*argptrA=NULL,*argptrB=NULL;
	
	if ((nargs=ARGNO("Dock",1,3))<0)	return pars->const_nil;
	
	if (nargs==1) {
		// if only one argument is present, Dock is just created
		if (NULL==(argptr1=ARGTYPE("Dock",1,T_STRING))) 	return pars->const_nil;
		assert(argptr1);
		ConsCell *retval 	= new ConsCell(pars->op);
		retval->type 		= T_SUP;
		retval->part.sup 	= new Dock(pars->sup_mgr,argptr1->part.string);
		return retval;
	} else if (nargs==2) {
		// if two arguments are present
		if (NULL==(argptr1=ARGTYPE("Dock",1,T_STRING))) 	return pars->const_nil;
		if (NULL!=(argptr2=ARGTYPE_N(2,T_SUP))) {
			// 2nd arg is a SUP, allocate it as the exit
			assert(argptr1&&argptr2);
			ConsCell *retval 	= new ConsCell(pars->op);
			retval->type 		= T_SUP;
			retval->part.sup 	= new Dock(pars->sup_mgr,argptr1->part.string,argptr2->part.sup);
			return retval;
		} else {
			// 2nd arg is a number (indicating DEFAULT or not)
			if (NULL==(argptr2=ARGTYPE("Dock",2,T_NUMBER))) 	return pars->const_nil;
			assert(argptr1&&argptr2);
			ConsCell *retval 	= new ConsCell(pars->op);
			retval->type 		= T_SUP;
			retval->part.sup 	= new Dock(pars->sup_mgr,argptr1->part.string,
									NULL,
									(argptr2->part.number >0));
			return retval;			
		}
	} else {
		// if three arguments are present, alloc. exit and set default
		if (NULL==(argptr1=ARGTYPE("Dock",1,T_STRING))) 	return pars->const_nil;
		if (NULL==(argptr2=ARGTYPE("Dock",2,T_SUP))) 	return pars->const_nil;
		if (NULL==(argptr3=ARGTYPE("Dock",3,T_NUMBER))) 	return pars->const_nil;
		
		assert(argptr1 && argptr2 && argptr3);
		ConsCell *retval 	= new ConsCell(pars->op);
		retval->type 		= T_SUP;
		retval->part.sup 	= new Dock(pars->sup_mgr,argptr1->part.string,
								argptr2->part.sup,
								(argptr3->part.number >0));
		return retval;
	}
}


ConsCell *Parser::SBI_bumpedQ(Parser *pars, ConsCell *args)	
{
	ConsCell *argptr1=NULL;
	int nargs=-1, sensor=0;
	
	if ((nargs=ARGNO("Bumped?",0,1))<0)	return pars->const_nil;
	if (nargs>0) {
		if (NULL==(argptr1=ARGTYPE("Bumped",1,T_NUMBER)))return pars->const_nil;
		sensor = argptr1->part.number;
	}
	
	ConsCell *retval 	= new ConsCell(pars->op);
	retval->type 		= T_SUP;
	retval->part.sup 	= new BumpedQ(pars->robot,sensor);
	
	return retval;
}

ConsCell *Parser::SBI_sonarQ(Parser *pars, ConsCell *args)	
{
	ConsCell *argptr1=NULL;
	int nargs=-1, sensor=0;
	
	if ((nargs=ARGNO("Sonar?",1,1))<0)	return pars->const_nil;
	if (NULL==(argptr1=ARGTYPE("Sonar?",1,T_NUMBER)))return pars->const_nil;
	sensor = argptr1->part.number;
	
	ConsCell *retval 	= new ConsCell(pars->op);
	retval->type 		= T_SUP;
	retval->part.sup 	= new SonarQ(pars->robot,sensor);
	
	return retval;
}

ConsCell *Parser::SBI_lightQ(Parser *pars, ConsCell *args)	
{
	ConsCell *argptr1=NULL;
	int nargs=-1, sensor=0;
	
	if ((nargs=ARGNO("Light?",1,1))<0)	return pars->const_nil;
	if (NULL==(argptr1=ARGTYPE("Light?",1,T_NUMBER)))return pars->const_nil;
	sensor = argptr1->part.number;
	
	ConsCell *retval 	= new ConsCell(pars->op);
	retval->type 		= T_SUP;
	retval->part.sup 	= new LightQ(pars->robot,sensor);
	
	return retval;
}

ConsCell *Parser::SBI_radarQ(Parser *pars, ConsCell *args)	
{
	ConsCell *argptr1=NULL;
	int nargs=-1;
	
	if ((nargs=ARGNO("Radar?",0,0))<0)	return pars->const_nil;
	
	ConsCell *retval 	= new ConsCell(pars->op);
	retval->type 		= T_SUP;
	retval->part.sup 	= new RadarQ(pars->robot);
	
	return retval;
}

ConsCell *Parser::SBI_pirQ(Parser *pars, ConsCell *args)	
{
	ConsCell *argptr1=NULL;
	int nargs=-1;
	
	if ((nargs=ARGNO("Pir?",0,0))<0)	return pars->const_nil;
	
	ConsCell *retval 	= new ConsCell(pars->op);
	retval->type 		= T_SUP;
	retval->part.sup 	= new PirQ(pars->robot);
	
	return retval;
}

ConsCell *Parser::SBI_lightacQ(Parser *pars, ConsCell *args)	
{
	ConsCell *argptr1=NULL;
	int nargs=-1;
	
	if ((nargs=ARGNO("LightAC?",0,0))<0)	return pars->const_nil;
	
	ConsCell *retval 	= new ConsCell(pars->op);
	retval->type 		= T_SUP;
	retval->part.sup 	= new LightACQ(pars->robot);
	
	return retval;
}

ConsCell *Parser::SBI_transvQ(Parser *pars, ConsCell *args)	
{
	ConsCell *argptr1=NULL;
	int nargs=-1;
	
	if ((nargs=ARGNO("Transv?",0,0))<0)	return pars->const_nil;
	
	ConsCell *retval 	= new ConsCell(pars->op);
	retval->type 		= T_SUP;
	retval->part.sup 	= new TransvQ(pars->robot);
	
	return retval;
}

ConsCell *Parser::SBI_rotvQ(Parser *pars, ConsCell *args)	
{
	ConsCell *argptr1=NULL;
	int nargs=-1;
	
	if ((nargs=ARGNO("Rotv?",0,0))<0)	return pars->const_nil;
	
	ConsCell *retval 	= new ConsCell(pars->op);
	retval->type 		= T_SUP;
	retval->part.sup 	= new RotvQ(pars->robot);
	
	return retval;
}


ConsCell *Parser::SBI_posxQ(Parser *pars, ConsCell *args)	
{
	ConsCell *argptr1=NULL;
	int nargs=-1;
	
	if ((nargs=ARGNO("Posx?",0,0))<0)	return pars->const_nil;
	
	ConsCell *retval 	= new ConsCell(pars->op);
	retval->type 		= T_SUP;
	retval->part.sup 	= new PosxQ(pars->robot);
	
	return retval;
}


ConsCell *Parser::SBI_posyQ(Parser *pars, ConsCell *args)	
{
	ConsCell *argptr1=NULL;
	int nargs=-1;
	
	if ((nargs=ARGNO("Posy?",0,0))<0)	return pars->const_nil;
	
	ConsCell *retval 	= new ConsCell(pars->op);
	retval->type 		= T_SUP;
	retval->part.sup 	= new PosyQ(pars->robot);
	
	return retval;
}


ConsCell *Parser::SBI_attitudeQ(Parser *pars, ConsCell *args)	
{
	ConsCell *argptr1=NULL;
	int nargs=-1;
	
	if ((nargs=ARGNO("Attitude?",0,0))<0)	return pars->const_nil;
	
	ConsCell *retval 	= new ConsCell(pars->op);
	retval->type 		= T_SUP;
	retval->part.sup 	= new AttitudeQ(pars->robot);
	
	return retval;
}


ConsCell *Parser::SBI_compassQ(Parser *pars, ConsCell *args)	
{
	ConsCell *argptr1=NULL;
	int nargs=-1;
	
	if ((nargs=ARGNO("Compass?",0,0))<0)	return pars->const_nil;
	
	ConsCell *retval 	= new ConsCell(pars->op);
	retval->type 		= T_SUP;
	retval->part.sup 	= new CompassQ(pars->robot);
	
	return retval;
}


ConsCell *Parser::SBI_panQ(Parser *pars, ConsCell *args)	
{
	ConsCell *argptr1=NULL;
	int nargs=-1;
	
	if ((nargs=ARGNO("Pan?",0,0))<0)	return pars->const_nil;
	
	ConsCell *retval 	= new ConsCell(pars->op);
	retval->type 		= T_SUP;
	retval->part.sup 	= new PanQ(pars->robot);
	
	return retval;
}


ConsCell *Parser::SBI_tiltQ(Parser *pars, ConsCell *args)	
{
	ConsCell *argptr1=NULL;
	int nargs=-1;
	
	if ((nargs=ARGNO("Tilt?",0,0))<0)	return pars->const_nil;
	
	ConsCell *retval 	= new ConsCell(pars->op);
	retval->type 		= T_SUP;
	retval->part.sup 	= new TiltQ(pars->robot);
	
	return retval;
}


ConsCell *Parser::SBI_modeQ(Parser *pars, ConsCell *args)	
{
	ConsCell *argptr1=NULL;
	int nargs=-1;
	
	if ((nargs=ARGNO("Mode?",0,0))<0)	return pars->const_nil;
	
	ConsCell *retval 	= new ConsCell(pars->op);
	retval->type 		= T_SUP;
	retval->part.sup 	= new ModeQ(pars->robot);
	
	return retval;
}


ConsCell *Parser::SBI_soundQ(Parser *pars, ConsCell *args)	
{
	ConsCell *argptr1=NULL;
	int nargs=-1;
	
	if ((nargs=ARGNO("Sound?",0,0))<0)	return pars->const_nil;
	
	ConsCell *retval 	= new ConsCell(pars->op);
	retval->type 		= T_SUP;
	retval->part.sup 	= new SoundQ(pars->robot);
	
	return retval;
}

///////////////////////////////////////////////////////////////////////
//
//   Actuator output
//

ConsCell *Parser::SBI_setsound(Parser *pars, ConsCell *args)	
{
	ConsCell *retval=NULL;
	int nargs=-1;
	
	if ((nargs=ARGNO("SetSound",1,1))<0)	return pars->const_nil;
	if (args->CAR->type==T_NUMBER) {
		retval = new ConsCell(pars->op);
		retval->type = T_SUP;
		retval->part.sup = new SetSound(pars->robot,(float)args->CAR->part.number);
	} else {
		if (args->CAR->type==T_SUP) {
			retval = new ConsCell(pars->op);
			retval->type = T_SUP;
			retval->part.sup = new SetSound(pars->robot,args->CAR->part.sup);
		} else {
			return pars->const_nil;
		}
	}
	return retval;
}

ConsCell *Parser::SBI_say(Parser *pars, ConsCell *args)	
{
	ConsCell *retval=NULL;
	int nargs=-1;
	
 	if ((nargs=ARGNO("Say",1,1))<0)	return pars->const_nil;
	if (args->CAR->type==T_STRING) {
		retval = new ConsCell(pars->op);
		retval->type = T_SUP;
		retval->part.sup = new Say(pars->robot,args->CAR->part.string);
	} else {
		if (args->CAR->type==T_SUP) {
			retval = new ConsCell(pars->op);
			retval->type = T_SUP;
			retval->part.sup = new Say(pars->robot,args->CAR->part.sup);
		} else {
			return pars->const_nil;
		}
	}
	return retval;
}

ConsCell *Parser::SBI_setvelocity(Parser *pars, ConsCell *args)	
{
	ConsCell *argptr1,*argptr2,*retval=NULL;
	int nargs=-1;
	
 	if ((nargs=ARGNO("SetVel",1,2))<0)	return pars->const_nil;
 	if (nargs==1) {
		if (NULL==(argptr1=ARGTYPE("SetVel",1,T_SUP)))return pars->const_nil;
		retval = new ConsCell(pars->op);
		retval->type = T_SUP;
		retval->part.sup = new SetVel(pars->robot,argptr1->part.sup);
 	} else {
		if (NULL==(argptr1=ARGTYPE("SetVel",1,T_NUMBER)))return pars->const_nil;
		if (NULL==(argptr2=ARGTYPE("SetVel",2,T_NUMBER)))return pars->const_nil;
		retval = new ConsCell(pars->op);
		retval->type = T_SUP;
		retval->part.sup = new SetVel(pars->robot,argptr1->part.number,argptr2->part.number);
	}
	return retval;
}


ConsCell *Parser::SBI_setposition(Parser *pars, ConsCell *args)	
{
	ConsCell *argptr1,*argptr2,*argptr3,*retval=NULL;
	int nargs=-1;
	
 	if ((nargs=ARGNO("SetPos",1,3))<0)	return pars->const_nil;
 	if (nargs==1) {
		if (NULL==(argptr1=ARGTYPE("SetPos",1,T_SUP)))return pars->const_nil;
		retval = new ConsCell(pars->op);
		retval->type = T_SUP;
		retval->part.sup = new SetPos(pars->robot,argptr1->part.sup);
 	} else {
		if (nargs!=3) {
			ARGNO("SetPos",3,3);	// create error msg
			return pars->const_nil;
		}
		if (NULL==(argptr1=ARGTYPE("SetPos",1,T_NUMBER)))return pars->const_nil;
		if (NULL==(argptr2=ARGTYPE("SetPos",2,T_NUMBER)))return pars->const_nil;
		if (NULL==(argptr3=ARGTYPE("SetPos",3,T_NUMBER)))return pars->const_nil;
		retval = new ConsCell(pars->op);
		retval->type = T_SUP;
		retval->part.sup = new SetPos(pars->robot,
			argptr1->part.number,
			argptr2->part.number,
			argptr3->part.number);
	}
	return retval;
}

ConsCell *Parser::SBI_setcam(Parser *pars, ConsCell *args)	
{
	ConsCell *argptr1,*argptr2,*retval=NULL;
	int nargs=-1;

 	if ((nargs=ARGNO("SetCam",1,2))<0)	return pars->const_nil;
 	if (nargs==1) {
		if (NULL==(argptr1=ARGTYPE("SetCam",1,T_SUP)))return pars->const_nil;
		retval = new ConsCell(pars->op);
		retval->type = T_SUP;
		retval->part.sup = new SetCam(pars->robot,argptr1->part.sup);
 	} else {
		if (NULL==(argptr1=ARGTYPE("SetCam",1,T_NUMBER)))return pars->const_nil;
		if (NULL==(argptr2=ARGTYPE("SetCam",2,T_NUMBER)))return pars->const_nil;
		retval = new ConsCell(pars->op);
		retval->type = T_SUP;
		retval->part.sup = new SetCam(pars->robot,argptr1->part.number,argptr2->part.number);
	}
	return retval;
}

ConsCell *Parser::SBI_setmode(Parser *pars, ConsCell *args)	
{
	ConsCell *retval=NULL;
	int nargs=-1;
	
 	if ((nargs=ARGNO("SetMode",1,1))<0)	return pars->const_nil;
 	if (args->CAR->type==T_SUP) {
		retval = new ConsCell(pars->op);
		retval->type = T_SUP;
		retval->part.sup = new SetMode(pars->robot,args->CAR->part.sup);
 	} else if (args->CAR->type==T_NUMBER) {
		retval = new ConsCell(pars->op);
		retval->type = T_SUP;
		retval->part.sup = new SetMode(pars->robot,args->CAR->part.number);
 	} else {
		if (!(ARGTYPE("SetMode",1,T_NUMBER)))return pars->const_nil; // produce error msg
 	}
	return retval;
}

ConsCell *Parser::SBI_setimagesize(Parser *pars, ConsCell *args)	
{
	ConsCell *argptr1,*argptr2,*retval=NULL;
	int nargs=-1;
	
 	if ((nargs=ARGNO("SetISize",1,2))<0)	return pars->const_nil;
 	if (nargs==1) {
		if (NULL==(argptr1=ARGTYPE("SetISize",1,T_SUP)))return pars->const_nil;
		retval = new ConsCell(pars->op);
		retval->type = T_SUP;
		retval->part.sup = new SetISize(pars->robot,argptr1->part.sup);
 	} else {
		if (NULL==(argptr1=ARGTYPE("SetISize",1,T_NUMBER)))return pars->const_nil;
		if (NULL==(argptr2=ARGTYPE("SetISize",2,T_NUMBER)))return pars->const_nil;
		retval = new ConsCell(pars->op);
		retval->type = T_SUP;
		retval->part.sup = new SetISize(pars->robot,argptr1->part.number,argptr2->part.number);
	}
	return retval;
}








