/*

DISCLAIMER 

InSitu is provided as is, without any kind of guarantees or warranty. There is 
no documentation. The author does not provide any kind of support to users of 
this source code and is not responsible for any kind of problems or damages 
resulting from its use. 

The source code is the result of a long iterative design process for a research 
prototype and was never intended to be used in any kind of product. In general, 
the code cannot be expected to compile in any straightforward manner on your 
computer system, since the robot on which it was developed, has a custom design. 

*/

#ifndef _Interpreter_h_
#define _Interpreter_h_

/* Rene Schaad                                                    */
/* Departement of Computer Science, Univ. of Zurich, Switzerland  */
/* schaad@ifi.unizh.ch                                            */

#include "defines.h"
#include "PrgStruct.h"
#include "OperatorProxy.h"
#include "RobotProxy.h"
#include <stdio.h>

#define INBUF_LEN 200
#define MAX_NAME_LEN 200
#define LEXEM_BUF_SIZE 1000
#define MAX_OBJECT_TYPES 100
#define IP_CMD_BUFLEN 5000

class InterpreterObject;


//---- Error codes -----------------------------------------------

enum Errorcode {no_err, 			// 0
				internal_err, 		// 1
				match_err, 			// 2
				cmd_notrec_err, 	// 3
				scan_err, 			// 4
				syntax_err, 		// 5
				method_err,			// 6
				wrongarg_err, 		// 7
				wrongnarg_err,		// 8
				notexist_err		// 9
				};

//---- Symbol Table Management -----------------------------------------------

enum SymbolType		{				// types of symbols
					T_UNBOUND,		// unbound symbol
					T_CONST, 		// constant - cannot be changed
					T_VARIABLE,	 	// bound variable
					T_ERROR			// invalid symbol
					};
					
enum ConsCellType	{				// types of cons-cells	(used by eval..)
					T_VOID,			// void object
					T_NUMBER,		// number (only doubles)
					T_STRING,		// string
					T_LIST,			// list
					T_SYMBOL,		// an unevaluated symbol
					T_CONSTCELL,	// a built-in constant
					T_SUP,			// sub - universal plan
					T_SBUILTIN,		// built-in non-evaluating functions
					T_UBUILTIN,		// built-in non-evaluating functions
					T_EBUILTIN		// built-in evaluating functions
					};
					
enum TokenType		{				// types of tokens returned by scanner
					TOK_ERROR,		// an error occured
					TOK_SYMBOL,		// a symbol was found
					TOK_LBRACKET,	// a left bracket was found
					TOK_RBRACKET,	// a right bracket was found
					TOK_STRING, 	// a string was found ("....")
					TOK_NUMBER,		// a number was found (doubles only)
					TOK_EOF			// an end-of-file was found
					};
					
enum CCRole {CC_CAR, CC_CDR};

class SymbolTableEntry;
class ValObject; 
class Token;
class ConsCell;
class Parser;
class SupMgr;
class Steppable;
typedef ConsCell *(*Builtin)(Parser *, ConsCell *);

class ConsCell : public ISobject {
public:
	ConsCell(OperatorProxy *op_,double nr):op(op_),type(T_NUMBER){part.number=nr;};
	ConsCell(OperatorProxy *op_,SymbolTableEntry *ste_):op(op_),type(T_SYMBOL){part.ste=ste_;};
	ConsCell(OperatorProxy *op_,Token *org=NULL);
	ConsCell(OperatorProxy *op_,ConsCell *car, ConsCell *cdr=NULL);
	void	PrintLn(int indent=0, CCRole role=CC_CAR);
	void	Print(int indent=0, CCRole role=CC_CAR);
	void	PrintIndent(int indent=0);
	void	SetString(char *txt=NULL);
	
	ConsCellType	type;
	OperatorProxy	*op;		// operator proxy for IO
	
	// depending on type, these are the contents
	union {
		double	number;				// for T_NUMBER
		char	*string;			// for T_STRING
		SymbolTableEntry *ste;		// for T_SYMBOL, T_CONSTCELL
		Builtin			builtin;	// for T_BUILDIN
		Steppable		 *sup;		// for T_SUP		
		struct {					// for T_LIST, T_FUNCTION
			ConsCell	*car;	// first element
			ConsCell	*cdr;	// rest
		} pair;
	} part;
};


class SymbolTableEntry : public ISobject {
public:
	SymbolTableEntry(OperatorProxy *op_,char *lexem_, SymbolType t_=T_UNBOUND, ConsCell *binding_=NULL);
	SymbolTableEntry(OperatorProxy *op_,SymbolTableEntry *org);
	~SymbolTableEntry();
	void SetLexem(char	*txt);
	void Print(void) {*op << lexem;};
	void	Push(SymbolType t_=T_UNBOUND, ConsCell *binding_=NULL);
	boolean Pop(SymbolType &t_, ConsCell *&binding_);
	void Set(char *lexem_, SymbolType t_, ConsCell *binding_, boolean empty_=FALSE);
	void Set(SymbolTableEntry *org);
	
	// These are the slots for the symbols
	SymbolType		type;		// what is the type of the symbol < SEPARATOR
	char			*lexem;		// buffer containing symbol string
	ConsCell		*binding;	// what is the symbol bound to
	boolean			empty;		// is this ste empty (==not valid)
	Stack			pushed;		// pushed versions if myself
	OperatorProxy	*op;		// operator proxy for IO
};

class SymbolTable : public ISobject  {	
public:
	SymbolTable(OperatorProxy *op_):op(op_){};
	~SymbolTable();
	
	SymbolTableEntry	*Append(char *lextext, SymbolType t_=T_UNBOUND, ConsCell *binding_=NULL);
	SymbolTableEntry	*Replace(char *lextext, SymbolType t_=T_UNBOUND, ConsCell *binding_=NULL);
	SymbolTableEntry	*Push(SymbolTableEntry *ste_, SymbolType t_=T_UNBOUND, ConsCell *binding_=NULL);
	SymbolTableEntry	*Push(char *lextext, SymbolType t_=T_UNBOUND, ConsCell *binding_=NULL);
	SymbolTableEntry	*Pop(SymbolTableEntry *ste_, SymbolType &t_, ConsCell *&binding_);
	SymbolTableEntry	*Pop(SymbolTableEntry *ste_);
	SymbolTableEntry	*Pop(char *lextext, SymbolType &t_, ConsCell *&binding_);
	SymbolTableEntry	*AppendBuiltin(char *lextext, Builtin  builtin_, ConsCellType t_);
	SymbolTableEntry	*AppendTree(char *lextext, TreeBase *tree_);
	SymbolTableEntry	*Lookup(char *lextext);
	void				PrintAll(void);
private:
	List				symbols;
	OperatorProxy		*op;		// operator proxy for IO
};

class Token : public ISobject  {
public:
	Token(OperatorProxy *op_,Token *org=NULL); 
	~Token() {delete [] string;};
	
	void	Copy(Token *org=NULL);
	void	SetString(char *txt);
		
	TokenType			token_id;
	OperatorProxy		*op;		// operator proxy for IO
	
	// depending on type, these are the contents
	double				number;
	char				*string;
	SymbolTableEntry 	*ste;
};


// stores intermediate state of (nested) parser/scanner
class ParserState : public ISobject {
friend class Parser;
private:
	ParserState(Token *l,unsigned int ln, char *cb, unsigned int sp, Token *rv) :
		lookahead(l),line_no(ln),cmd_buffer(cb),scanpos(sp),returned_val(rv){};
	Token			*lookahead;		// lookahead token pointer
	unsigned int 	line_no;		// current line number
	char			*cmd_buffer;	// buffered input
	unsigned int 	scanpos;		// where in the input file is the scanner
	Token			*returned_val;	// value returned by the scanner
};
				
//---- Parser -----------------------------------------------------------


class Parser : public Action {
public:
	Parser(RobotProxy *robot_,OperatorProxy *op_,SupMgr *new_manager=NULL);
	~Parser();
	
	virtual Data	*Step(void) {
		(void)Parse(op->GetCommand());
		return NULL;		
	};
	
	Token 		*Scan(char *inbuffer, unsigned int &position);
	ConsCell	*Parse(char *inbuffer);	// returns a parse tree
	ConsCell	*Eval(ConsCell *expr);		// to be run after the parse
	ConsCell	*EvalCall(ConsCell *fct, ConsCell *args);	
	ConsCell	*EvalLambda(ConsCell *largs, ConsCell *args);	
	
	void	AppendTree(char *lextext, TreeBase *tree_)
		{symtab->AppendTree(lextext,tree_);};
	
	void	PrintTree(void);
	void	Save(char *inbuffer);
	void	Restore(void);
	
	// argument checking functions
	static int		ArgNo(char *sf, int sln, Parser *pars, char *name, ConsCell *args, int min, int max=-1);
	static ConsCell	*ArgType(char *sf, int sln, Parser *pars, char *name, ConsCell *args, int no, ConsCellType t);
	static ConsCell	*ArgTypeN(char *sf, int sln, Parser *pars, ConsCell *args, int no, ConsCellType t);
	
	// built-in LISP evaluating functions
	static ConsCell *EBI_mul(Parser *pars, ConsCell *args);		// multiplication
	static ConsCell *EBI_add(Parser *pars, ConsCell *args);		// addition
	static ConsCell *EBI_div(Parser *pars, ConsCell *args);		// division
	static ConsCell *EBI_sub(Parser *pars, ConsCell *args);		// subtraction
	static ConsCell *EBI_eval(Parser *pars, ConsCell *args);	// evaluation
	static ConsCell *EBI_print(Parser *pars, ConsCell *args);	// print arg
	static ConsCell *EBI_list(Parser *pars, ConsCell *args);	// list construction
	static ConsCell *EBI_listp(Parser *pars, ConsCell *args);	// is arg a list?
	static ConsCell *EBI_atomp(Parser *pars, ConsCell *args);	// is arg an atom?
	static ConsCell *EBI_length(Parser *pars, ConsCell *args);	// length of list
	static ConsCell *EBI_car(Parser *pars, ConsCell *args);		// head/car of list
	static ConsCell *EBI_cdr(Parser *pars, ConsCell *args);		// body/cdr of list

	// built-in LISP non-evaluating functions
	static ConsCell *UBI_print_symtab(Parser *pars, ConsCell *args); // print the symbol table
	static ConsCell *UBI_print_docklist(Parser *pars, ConsCell *args); // print the symbol table
	static ConsCell *UBI_quote(Parser *pars, ConsCell *args);	// quoting
	static ConsCell *UBI_setq(Parser *pars, ConsCell *args);	// setq
	static ConsCell *UBI_defun(Parser *pars, ConsCell *args);	// define function
	static ConsCell *UBI_load(Parser *pars, ConsCell *args);	// define function
	static ConsCell *UBI_exit(Parser *pars, ConsCell *args);	// exit the behavior module
	
	// built-in data types
	static ConsCell *SBI_waypoint(Parser *pars, ConsCell *args);	// waypoint data type
	
	// built-in universal plans
	static ConsCell *SBI_achieve(Parser *pars, ConsCell *args);	// hook temporary sups in scheduler
	static ConsCell *SBI_do(Parser *pars, ConsCell *args);		// hook sups in scheduler
	static ConsCell *SBI_stop(Parser *pars, ConsCell *args);	// remove something from scheduler
	static ConsCell *SBI_every(Parser *pars, ConsCell *args);	
	static ConsCell *SBI_if(Parser *pars, ConsCell *args);		
	static ConsCell *SBI_noop(Parser *pars, ConsCell *args);	
	static ConsCell *SBI_ramp(Parser *pars, ConsCell *args);	
	static ConsCell *SBI_pulsetrain(Parser *pars, ConsCell *args);	
	static ConsCell *SBI_loop(Parser *pars, ConsCell *args);	
	static ConsCell *SBI_sequence(Parser *pars, ConsCell *args);
	static ConsCell *SBI_parallel(Parser *pars, ConsCell *args);
	static ConsCell *SBI_switch(Parser *pars, ConsCell *args);
	static ConsCell *SBI_taction(Parser *pars, ConsCell *args);	
	static ConsCell *SBI_ttester(Parser *pars, ConsCell *args);	
	
	static ConsCell *SBI_dock(Parser *pars, ConsCell *args);	
	static ConsCell *SBI_and(Parser *pars, ConsCell *args);	
	static ConsCell *SBI_or(Parser *pars, ConsCell *args);	

	// built-in interface universal plans: SENSORS

	static ConsCell *SBI_bumpedQ(Parser *pars, ConsCell *args);
	static ConsCell *SBI_sonarQ(Parser *pars, ConsCell *args);
	static ConsCell *SBI_lightQ(Parser *pars, ConsCell *args);
	static ConsCell *SBI_radarQ(Parser *pars, ConsCell *args);
	static ConsCell *SBI_pirQ(Parser *pars, ConsCell *args);
	static ConsCell *SBI_lightacQ(Parser *pars, ConsCell *args);
	static ConsCell *SBI_transvQ(Parser *pars, ConsCell *args);
	static ConsCell *SBI_rotvQ(Parser *pars, ConsCell *args);
	static ConsCell *SBI_posxQ(Parser *pars, ConsCell *args);
	static ConsCell *SBI_posyQ(Parser *pars, ConsCell *args);
	static ConsCell *SBI_attitudeQ(Parser *pars, ConsCell *args);
	static ConsCell *SBI_compassQ(Parser *pars, ConsCell *args);
	static ConsCell *SBI_panQ(Parser *pars, ConsCell *args);
	static ConsCell *SBI_tiltQ(Parser *pars, ConsCell *args);
	static ConsCell *SBI_modeQ(Parser *pars, ConsCell *args);
	static ConsCell *SBI_soundQ(Parser *pars, ConsCell *args);

	// built-in interface universal plans: ACTUATORS

	static ConsCell *SBI_setsound(Parser *pars, ConsCell *args);
	static ConsCell *SBI_say(Parser *pars, ConsCell *args);
	static ConsCell *SBI_setvelocity(Parser *pars, ConsCell *args);
	static ConsCell *SBI_setposition(Parser *pars, ConsCell *args);
	static ConsCell *SBI_setcam(Parser *pars, ConsCell *args);
	static ConsCell *SBI_setmode(Parser *pars, ConsCell *args);
	static ConsCell *SBI_setimagesize(Parser *pars, ConsCell *args);

//	static ConsCell *SBI_localmap(Parser *pars, ConsCell *args);	
//	static ConsCell *SBI_stgraph(Parser *pars, ConsCell *args);
	
	// utility functions for the scanner 
	boolean	Isnewline(char c) { return ((c==13)||(c==10)||(c==11));}
	boolean	Isblank(char c)   { return ((c==9)||(c==' ')||Isnewline(c));}
	boolean	Iscomment(char c) { return (c==';');}
	boolean	Isdigit(char c)   { return ((c>='0')&&(c<='9'));}
	boolean	Issign(char c)   { return ((c=='+')||(c=='-'));}
	boolean	Isalpha(char c)   { return (((c>='a')&&(c<='z'))||
										((c>='A')&&(c<='Z'))||
										(c=='?')||(c=='.')||
										(c=='_')||(c=='*')||
										(c=='+')||(c=='-')||(c=='/'));}
	
	Errorcode P_function(void);
	Errorcode P_atom(void);
	Errorcode Match(TokenType tok);

	Token			*lookahead;		// lookahead token pointer
	char	 		lexbuf[MAX_NAME_LEN];	// lexical buffer used to scan atoms
	unsigned int 	line_no;		// current line number
	char			*cmd_buffer;	// buffered input
	unsigned int 	scanpos;		// where in the input file is the scanner
	SymbolTable		*symtab;		// symbol table
	ConsCell		*returned_list;	// assembled parse tree
	Token			*returned_val;	// value returned by the scanner
	Token   		*error_token;	// used to return error by scanner
	Token			*eof_token;  	// used to return eof by scanner
	
	unsigned int    nesting;		// how many times has parser been nested
	Stack			ps_stack;		// stack to hold state when nested
	
	// constants as used by eval
	ConsCell			*const_nil;
	SymbolTableEntry	*const_nilsym;
	ConsCell			*const_true;
	SymbolTableEntry	*const_truesym;
	ConsCell			*const_lambda;
	SymbolTableEntry	*const_lambdasym;
	
	// stuff used for universal plan evaluation and construction
	SupMgr				*sup_mgr;
	OperatorProxy		*op;		// operator proxy for IO
	RobotProxy			*robot;		// connect to robot (for interface functions)
};

#endif	/* _Interpreter_h_ */




