/*

DISCLAIMER 

InSitu is provided as is, without any kind of guarantees or warranty. There is 
no documentation. The author does not provide any kind of support to users of 
this source code and is not responsible for any kind of problems or damages 
resulting from its use. 

The source code is the result of a long iterative design process for a research 
prototype and was never intended to be used in any kind of product. In general, 
the code cannot be expected to compile in any straightforward manner on your 
computer system, since the robot on which it was developed, has a custom design. 

*//* Rene Schaad                                                    */
/* Departement of Computer Science, Univ. of Zurich, Switzerland  */
/* schaad@ifi.unizh.ch                                            */


#include "LocalMap.h"
#include <math.h>
#include <stdio.h>

//---- LocalMap ---------------------------------------------------------


LocalMap::LocalMap(RobotProxy *robot_p) 
{
	Init(robot_p);
}

LocalMap::~LocalMap()
{
	delete cell;
}

void LocalMap::Init(RobotProxy *robot_p)
{
	assert(robot_p!=NULL);
	for (int x=0; x<LMSIZE_CX; x++)	
		for (int y=0; y<LMSIZE_CY; y++)	{
			ccell[x][y] =
			sqcell[x][y] =
			fcell[x][y] =
			cell[x][y] = INI_CERT;
		}
	robot = robot_p;
	robot->GetDRPosition(cpos_x, cpos_y, cpos_a);
	apos_x = cpos_x;
	apos_y = cpos_y;
	apos_a = cpos_a;
}

void LocalMap::Move(void)
{
	int xf,xt,yf,yt;

	// read in position data and compute difference agentpos-cellpos
	robot->GetDRPosition(apos_x,apos_y,apos_a);
	dx = apos_x - cpos_x;
	dy = apos_y - cpos_y;
	da = apos_a - cpos_a;
	
	// move the array to compensate for motion of agent
	int mx = (int) (-dx/CELLSIZE_X);	// number of cells to move
	int my = (int) (-dy/CELLSIZE_Y);
	cpos_x -= mx * CELLSIZE_X;		// update the position of the array
	cpos_y -= my * CELLSIZE_Y;
	
	if (mx>=0) {xf=mx; xt=LMSIZE_CX;}	// compute loop bounds
	else       {xf=0;  xt=LMSIZE_CX+mx;}
	if (my>=0) {yf=my; yt=LMSIZE_CY;}
	else       {yf=0;  yt=LMSIZE_CY+my;}
	
//	printf("mx(%d), my(%d)\n",mx,my);	
//	printf("apos_x(%f), apos_y(%f), cpos_x(%f), cpos_y(%f)\n",
//		apos_x,apos_y,cpos_x,cpos_y);	
		
	int x,y;
	
	if ((mx>=0)&&(my>=0)) {	// mx>0, my>0
		for ( x=xt-1; x>=xf; x--) {				// loop to set contents
			for ( y=yt-1; y>=yf; y--) {
				cell[x][y] = cell[x-mx][y-my]; 
			}
		}
	} else if (mx>=0) {		// mx>0, my<0
		for ( x=xt-1; x>=xf; x--) {				// loop to set contents
			for ( y=yf; y<yt; y++) {
				cell[x][y] = cell[x-mx][y-my]; 
			}
		}
	} else if (my>=0) {		// mx<0, my>0
		for ( x=xf; x<xt; x++) {				// loop to set contents
			for ( y=yt-1; y>=yf; y--) {
				cell[x][y] = cell[x-mx][y-my]; 
			}
		}
	} else {				// mx<0, my<0
		for ( x=xf; x<xt; x++) {				// loop to set contents
			for ( y=yf; y<yt; y++) {
				cell[x][y] = cell[x-mx][y-my];
			}
		}
	}
	

	
	for (x=0; x<LMSIZE_CX; x++) {		// set 0 elements
		for ( y=0; y<LMSIZE_CY; y++) {
			if ((x<xf)||(x>=xt)||(y<yf)||(y>=yt)) cell[x][y] = INI_CERT;
		}
	}


}

void LocalMap::Fuse(void)
{
	int   sn;		   // sonar number (1=front, 2=front_left...)
	int sx,sy,ex,ey; // sonar distance, start and end points of beam
	float sangle = 0;
	
	// position of robot relative to cell array
	float ax = apos_x - cpos_x;		// let cell array center be
	float ay = apos_y - cpos_y;		// origin
	int   mx = (int)ax;				// robot position in cell coordinates
	int   my = (int)ay;

	word b=robot->GetBumpers();		// read bumper sensor

	// fuse in robot body center 
	ConfPointMIN(mx,my);
		
	// fuse in sonar readings and bumpers
	for (sn=1; sn<=NO_SONARS; sn++) {
		float sdist = robot->GetSonar(sn);
		float siga = da+sangle;			// sum of angles
		float ca = cos(siga);			// cos of sum of angles
		float sa = sin(siga);			// sin of sum of angles
		sx = (int) ((ax + SONAR_RAD * ca)/CELLSIZE_X);	// coords of sonar sensor
		sy = (int) ((ay + SONAR_RAD * sa)/CELLSIZE_Y);
		
		// test sonar range reading
		if ((sdist<MAX_SONAR_R)&&(sdist>MIN_SONAR_R)) {	
			float sigd = SONAR_RAD+sdist;	// sum of radius & distance
			ex = (int) ((ax + sigd * ca)/CELLSIZE_X);		// coords of measured end point
			ey = (int) ((ay + sigd * sa)/CELLSIZE_Y);
			ConfLine(sx,sy,ex,ey);			// draw the line of confidences
		}
			
		// fuse in bumpers
		switch (sn) {
			case 1: // front plate
					if (b&0x0808) ConfPointMAX(sx,sy);
					else ConfPointMIN(sx,sy);
					break;
			case 2: // left front
					if (b&0x1010) ConfPointMAX(sx,sy);
					else ConfPointMIN(sx,sy);
					break;
			case 3: // left
					if (b&0x2020) ConfPointMAX(sx,sy);
					else ConfPointMIN(sx,sy);
					break;
			case 4: // left back
					if (b&0x4040) ConfPointMAX(sx,sy);
					else ConfPointMIN(sx,sy);
					break;
			case 5: // back
					if (b&0x8080) ConfPointMAX(sx,sy);
					else ConfPointMIN(sx,sy);
					break;
			case 6: // back right
					if (b&0x0101) ConfPointMAX(sx,sy);
					else ConfPointMIN(sx,sy);
					break;
			case 7: // right
					if (b&0x0202) ConfPointMAX(sx,sy);
					else ConfPointMIN(sx,sy);
					break;
			case 8: // right front
					if (b&0x0404) ConfPointMAX(sx,sy);
					else ConfPointMIN(sx,sy);
					break;
			default: // error
					printf("ERROR: invalid CASE in LocalMap::Fuse\n");
					break;
		}
		sangle += ANGLE_INC;			// increase sonar sensor angle
	}
}

void LocalMap::ConfLine(int sx, int sy, int ex, int ey)
{
	// draws line of confidence values
	// using Bresenham's line drawing algorithm
	// input parameters' coordinate system is centered in cell array

	// start and end coordinates are in cell units
	// center of coordinate system is (logically) in center of 
	// cell array
	
	int xch,ych,er=0;
	int deltaX = ex-sx;
	int deltaY = ey-sy;
	int x=sx,y=sy;
	
	// compute increments
	
	if (deltaX<0) {	
		xch = -1;
		deltaX = -deltaX;
	} else {
		xch = 1;
	}
	
	if (deltaY<0) {
		ych = -1;
		deltaY = -deltaY;
	} else {
		ych = 1;
	}
	
	// loop through the line
	if (deltaX<deltaY) {
		for (int i=0;i<deltaY;i++) {
			ConfPoint(x,y);
			y+=ych;
			er+=deltaX;
			if (er>deltaY) {
				x+=xch;
				er-=deltaY;
			}
		}
	} else {
		for (int i=0;i<deltaX;i++) {
			ConfPoint(x,y);
			x+=xch;
			er+=deltaY;
			if (er>deltaX) {
				y+=ych;
				er-=deltaX;
			}
		}
	}
	ConfEndPoint(x,y);
}

void LocalMap::ConfPointMIN(int px, int py)
{
	px+=OFFSET_CX;	// move the origin to the center of the array
	py+=OFFSET_CY;
	if ((px>=0)&&(py>=0)&&(px<LMSIZE_CX)&&(py<LMSIZE_CY)) { //check bounds
		cell[px][py] = MIN_CERT; // check bounds
	}
}

void LocalMap::ConfPointMAX(int px, int py)
{
	px+=OFFSET_CX;	// move the origin to the center of the array
	py+=OFFSET_CY;
	if ((px>=0)&&(py>=0)&&(px<LMSIZE_CX)&&(py<LMSIZE_CY)) { //check bounds
		cell[px][py] = MAX_CERT; // check bounds
	}
}

void LocalMap::ConfPoint(int px, int py)
{
	px+=OFFSET_CX;	// move the origin to the center of the array
	py+=OFFSET_CY;
	if ((px>=0)&&(py>=0)&&(px<LMSIZE_CX)&&(py<LMSIZE_CY)) { //check bounds
		int nv = (int)cell[px][py] - LM_DEC; // subtract one confidence unit
		cell[px][py] = (unsigned char) ((nv>=MIN_CERT)?nv:MIN_CERT); // check bounds
	}
}

void LocalMap::ConfEndPoint(int px, int py)
{
	// explanation see above (ConfPoint)
	px+=OFFSET_CX;
	py+=OFFSET_CY;
	
	
	if ((px>=0)&&(py>=0)&&(px<LMSIZE_CX)&&(py<LMSIZE_CY)) {
		// increment the cells
		int nv = (int)cell[px][py] + LM_INC;	// increment cell
		
		// consider the neighbours as well
		int xf=px-1,xt=px+1,yf=py-1,yt=py+1;			// find bounds
		xf = (xf<0)?px:xf;								// check bounds
		xt = (xt>=LMSIZE_CX)?px:xt;
		xf = (yf<0)?py:yf;
		xt = (yt>=LMSIZE_CY)?py:yt;		
		int sum=0;
		for (int x=xf;x<=xt;x++)						// sum up neighbors
			for (int y=xf;y<=yt;y++) sum += cell[x][y];	
		nv += sum/NEIGHBOR_SCALE;						// scale them back
		
		// limit range
		cell[px][py] = (unsigned char) ((nv<=MAX_CERT)?nv:MAX_CERT);
	}
}

void LocalMap::Cluster(void)
{
	int x,y,i,j;
	
	// square all cell values
	for ( x=0;x<LMSIZE_CX;x++)
		for ( y=0;y<LMSIZE_CY;y++) {
			int c = cell[x][y];
			sqcell[x][y] = (unsigned char) (c*c/MAX_CERT);
		}
		
	// average and threshold (very inefficient implementation, to be improved)
	for ( x=1;x<LMSIZE_CX-1;x++)
		for ( y=1;y<LMSIZE_CY-1;y++) {
			int sum = 0;
			for (i=x-1;i<=x+1;i++) 
				for (j=y-1;j<=y+1;j++) sum += sqcell[i][j];
			ccell[x][y] = (((sum>=CLUSTER_THRES)&&(sqcell[x][y]>IND_THRES))?MAX_CERT:MIN_CERT);
		}
}

void LocalMap::FreeSpace(int size)
{
	assert(size>0);
	
	// first find runs of length "size" on X-axis
	// then find runs of length "size" on Y-axis
}

Data *LocalMap::Step(void)
{
	Move();
	Fuse();
	Cluster();
	FreeSpace(3);
	return NULL;
}

