/*

DISCLAIMER 

InSitu is provided as is, without any kind of guarantees or warranty. There is 
no documentation. The author does not provide any kind of support to users of 
this source code and is not responsible for any kind of problems or damages 
resulting from its use. 

The source code is the result of a long iterative design process for a research 
prototype and was never intended to be used in any kind of product. In general, 
the code cannot be expected to compile in any straightforward manner on your 
computer system, since the robot on which it was developed, has a custom design. 

*/

#ifndef _Monsup_h_
#define _Monsup_h_

/* Rene Schaad                                                    */
/* Departement of Computer Science, Univ. of Zurich, Switzerland  */
/* schaad@ifi.unizh.ch                                            */

#include "PrgStruct.h"
#include "RobotProxy.h"
#include <windows/Qwindows.h>

////////////////////////////////////////////////////////////////////
//
//  Base class for all Intermediate Objects (such as markers, etc.)
//

class Marker : public Data {
public:
	Marker(DataType type_,boolean registered_=FALSE)
		:registered(registered_),Data(type_){};
	
	// data stuff, purely virtual
	virtual boolean	SetData(Data *org_)=0;
	virtual Data	*Duplicate(void)=0;
	virtual Data	*Print(void)=0;
		
	// handle locks of Intermediate Objects
	boolean				IsRegistered(void){return registered;};
	void				SetLock(boolean registered_=TRUE)	
							{registered=registered_;};
protected:	
	boolean		registered;	// did the object lock onto something ?
};


// if argument-marker is registered
class RegisteredQ : public Action {
public:
	RegisteredQ(Steppable *mark_=NULL);
	virtual	Data	*Step(void);
	virtual Fuzz	*IsDone(void)
		{if (mark) return mark->IsDone();else return &fuzz_true;};
private:
	Steppable	*mark;			// marker source
};

// register an intermediate object

class Register : public Action {
public:
	Register(Steppable *registrar_,Steppable *checker_)
		:registrar(NULL),checker(NULL),
		 registered(FALSE),most_recent(NULL)
		{done.SetFalse();
		 registrar=Getport(registrar_);checker=Getport(checker_);};
	virtual	Data	*Step(void);
	virtual	void	Reset(void);
	virtual	void	Abort(void);
private:
	Steppable	*registrar;
	Steppable	*checker;
	boolean		registered;
	Marker		*most_recent;	// most recent Marker/Marker
};

class VisualMarker : public Marker {
public:
	VisualMarker(DataType type_, CImage *img_, boolean registered_=FALSE)
		:src_image(img_),Marker(type_,registered_){};
	CImage	*GetImage(void){return src_image;};
protected:
	CImage	*src_image;	// image to which the marker refers
};

class VisualPoint : public VisualMarker {
public:
	VisualPoint(CImage *img_, float pan_=0.0, float tilt_=0.0, boolean registered_=FALSE)
		:pan(pan_),tilt(tilt_),VisualMarker(DT_VisualPoint,img_,registered_){};
	
	// manipulation
	void	Set(CImage *img_,float pan_=0.0, float tilt_=0.0,boolean registered_=FALSE) 
		{src_image=img_;pan=pan_;tilt=tilt_;registered=registered_;};
	void	Set(VisualPoint *newvp)	
		{if (newvp) {src_image=newvp->src_image;pan=newvp->pan;tilt=newvp->tilt;registered=newvp->registered;};};
	void	Get(CImage *&img_,float &pan_, float &tilt_, boolean &registered_)	
		{img_=src_image;pan_=pan;tilt_=tilt;registered_=registered;};

	virtual boolean	SetData(Data *org_){
		if (!org_) return FALSE;
		if (org_->type!=type) return FALSE;
		Set((VisualPoint*)org_);
		return TRUE;
	};

	VisualPoint	&operator += (VisualPoint &v){pan+=v.pan;tilt+=v.tilt;return *this;};
	VisualPoint	&operator -= (VisualPoint &v){pan+=v.pan;tilt+=v.tilt;return *this;};
	VisualPoint	&operator  = (VisualPoint &v){Set(&v);return *this;};
	
	virtual Data	*Duplicate(void){return new VisualPoint(src_image,pan,tilt,registered);};
	virtual Data	*Print(void){printf("VisualPoint(%f,%f,%d)",pan,tilt,registered);return this;};
private:
	float	pan,tilt;	// visual coordinates
};


/////////////////////////////////////////////////////////////////////////////
//
//    ACTIONS related to visual stuff
//

class DrawAll : public Action {
public:
	DrawAll(void){done.Set(FUZZTRUE);};	
	virtual	Data	*Step(void){Draw(); return NULL;};
};

class ShowVisual : public Action {
public:
	ShowVisual(Steppable *vis_marker_, unsigned int cx_, unsigned int cy_)
		:cx(cx_),cy(cy_),vis_marker(NULL)
		{vis_marker=Getport(vis_marker_);assert(vis_marker);done.Set(FUZZFALSE);};
	virtual	Data	*Step(void); // returns a marker to tracked object
	virtual void	Reset(void)
		{if (vis_marker)vis_marker->Reset();done.Set(FUZZFALSE);}
	virtual void	Abort(void)
		{if (vis_marker)vis_marker->Abort();done.Set(FUZZTRUE);}
private:	
	unsigned int cx,cy;
	Steppable *vis_marker;	// pointer to steppable delivering visual marker
	CImage	outbuf;			// buffer for outgoing image
};

class SaccadeTo : public Action {
public:
	SaccadeTo(RobotProxy *robot_, Steppable *vis_point_)
		:robot(robot_),vis_point(NULL)
		{vis_point=Getport(vis_point_);assert(vis_point&&robot);done.Set(FUZZFALSE);};
	virtual	Data	*Step(void); // returns a marker to tracked object
	virtual void	Reset(void)
		{if (vis_point)vis_point->Reset();done.Set(FUZZFALSE);}
	virtual void	Abort(void)
		{if (vis_point)vis_point->Abort();done.Set(FUZZTRUE);}
private:	
	RobotProxy *robot;
	Steppable *vis_point;	// pointer to steppable delivering visual marker
};


class MonitorCam : public Action {
public:
	MonitorCam(RobotProxy *robot_) 
		:robot(robot_),target(&bbox_image){assert(robot);done.Set(FUZZFALSE);};
	virtual	Data	*Step(void); // returns a marker to tracked object
	
	// image handling functions
	CImage *GetPrevImage(void){return &prev_image;}
	CImage *GetCurrImage(void){return &curr_image;}
	CImage *GetDiffImage(void){return &diff_image;}
	CImage *GetBboxImage(void){return &bbox_image;}

private:	
	RobotProxy *robot;
	
	CImage	curr_image;	// current camera image
	CImage	prev_image;	// previous camera image
	CImage	diff_image;	// difference of new and previous image
	CImage	bbox_image;	// image containing bounding box
	
	VisualPoint	target;	// target do be monitored
	
};



#endif	/* _Monsup_h_ */
