/*

DISCLAIMER 

InSitu is provided as is, without any kind of guarantees or warranty. There is 
no documentation. The author does not provide any kind of support to users of 
this source code and is not responsible for any kind of problems or damages 
resulting from its use. 

The source code is the result of a long iterative design process for a research 
prototype and was never intended to be used in any kind of product. In general, 
the code cannot be expected to compile in any straightforward manner on your 
computer system, since the robot on which it was developed, has a custom design. 

*/
/* Rene Schaad                                                    */
/* Departement of Computer Science, Univ. of Zurich, Switzerland  */
/* schaad@ifi.unizh.ch                                            */


#include "Navsup.h"
#include <iostream.h>
#include <math.h>
#include <stdlib.h>

#define TARGET_THRES_STOP  		0.10
#define TARGET_THRES_NOSTOP		0.50
#define TARGET_A_THRES_STOP  	0.0175
#define TARGET_A_THRES_NOSTOP  	100.0

#define SONAR_RADIUS            0.225
#define MIN_WALL_SAMPLES		2

#define FLOAT_ZERO              1E-6

#define MAX_TRANS_V				0.4
#define MAX_TURN_V				0.4
#define TURN_FACT				1.0
#define ORIENT_FACT				3.0
#define ACCEL_T					0.03

#define UNBUMP_V				0.05
#define UNBUMP_A				0.2
#define UNBUMP_A_THRES			0.02

/////////////////////////////////////////////////////////////
//
//    A C T I O N   for   Navigation
//

Data *IterateRoute::Step(void)
{
	if (done.True()) return NULL;

	float dd,dh,da;	// distances (translation, heading, attitude)
	float dt,at;	// thresholds for distance and attituce
	boolean dostop;	// stop at waypoint?
	Route *mr = GetRoute();
	
	if (mr) {
		Waypoint *cwp = mr->GetCurrent();
		while (cwp) {
			FloorVector(myrobot,cwp,dd,dh,da,dostop);
			if (Arrived(dd,da,dostop,dt,at)) {
				cwp = mr->Iterate();
				if (cwp) {
					float x,y,a;
					boolean s;
					cwp->Get(x,y,a,s);
					printf("Approaching waypoint %f/%f/%f/%d\n",x,y,a,s);
				} else
					printf("Last waypoint in route reached.\n");
			} else 
				return cwp;
		}
	}
	
	done.Set(FUZZTRUE);
	return NULL;
}

void IterateRoute::Reset(void)
{
	done.Set(FUZZFALSE);
	Route *mr = GetRoute();
	if (!mr) return;
	(void)mr->Reset();
}

void IterateRoute::Abort(void)
{
	done.Set(FUZZTRUE);
	Route *mr = GetRoute();
	if (!mr) return;
	(void)mr->GetLast();
}

Route *IterateRoute::GetRoute(void)
{
	if (route==NULL) return NULL;	// check steppable
	Data *myroute = route->Step();	// get the actual route
	if (!myroute) return NULL;		// check actual route
	if (myroute->type!=DT_Route) 
		return NULL; 				// check type for route
	return (Route *)myroute;
}

///////////////////////////////////////////////////////////////////////////
//   Plan a route (generate a route from hiwaymap, start, goal points

Data *PlanARoute::Step(void)
{ 
	assert(hiwaymap);
	assert(wp_start);
	assert(wp_goal);

	if ((!hiwaymap) || (!wp_start) || (!wp_goal)) return NULL;

	// get the map
	Data *argval = hiwaymap->Step();
	if (!argval) return NULL;
	if (argval->type != DT_HiwayMap) return NULL;
	HiwayMap *the_map = (HiwayMap*)argval;
	
	// get the starting point
	argval = wp_start->Step();
	if (!argval) return NULL;
	if (argval->type != DT_Waypoint) return NULL;
	Waypoint *the_start = (Waypoint*)argval;

	// get the goal point
	argval = wp_goal->Step();
	if (!argval) return NULL;
	if (argval->type != DT_Waypoint) return NULL;
	Waypoint *the_goal = (Waypoint*)argval;

	// compute a path from start to goal
	return the_map->GetRoute(the_start,the_goal);
}

///////////////////////////////////////////////////////////////////////////
//   Plan a random route (from somewhere to a randomly chosen end point)

Data *PlanARandomRoute::Step(void)
{ 
	assert(hiwaymap);
	assert(wp_start);
	if ((!hiwaymap) || (!wp_start)) return NULL;

	// get the map
	Data *argval = hiwaymap->Step();
	if (!argval) return NULL;
	if (argval->type != DT_HiwayMap) return NULL;
	HiwayMap *the_map = (HiwayMap*)argval;
	
	// get the starting point
	argval = wp_start->Step();
	if (!argval) return NULL;
	if (argval->type != DT_Waypoint) return NULL;
	Waypoint *the_start = (Waypoint*)argval;

	// get the goal point
	Waypoint *the_goal = the_map->GetRandomPoint();
	if (the_goal==NULL) return NULL;

	// compute a path from start to goal
	return the_map->GetRoute(the_start,the_goal);
}

///////////////////////////////////////////////////////////////////////////
//   Follow a wall

#define MIN_ACQUIRE_DIST   2.0
#define FW_MINDIST 0.5

FollowWall::FollowWall(RobotProxy *robot_,float mspeed_,float dist_, byte prefside_) 
		:robot(robot_), wside(NO_WALL), transspeed(mspeed_), prefside(prefside_),
		 prefdist(dist_),olddist(0.0)
{
	// while debugging, catch stupid values 
	assert(transspeed>0.0 && transspeed <0.41);
	assert(prefdist>0.4 && prefdist<2.01);
	
	// limit the allowed input values
	transspeed 	= Limit(transspeed,0.0,0.4); // allow only small forward speeds
	prefdist	= Limit(prefdist, FW_MINDIST, 2.0); // allow only close wall-hugs
	
	Reset();
}

void FollowWall::Reset(void)
{
	done.Set(FUZZFALSE);
	wside=prefside;
}

void FollowWall::Abort(void)
{
	done.Set(FUZZTRUE);
	wside=prefside;
}

Data *FollowWall::Step(void)
{
	done.Set(FUZZTRUE);
	if (!robot) {
		return &done;
	}
	
	// first, acquire wall
	if (wside==NO_WALL) {
		float r[4],mv;
		int tm;
		done.Set(FUZZFALSE);
		r[0]	= robot->GetSonar(2);		// gen sensor values
		r[1]	= robot->GetSonar(3);
		r[2]	= robot->GetSonar(7);
		r[3]	= robot->GetSonar(8);
		MinIndex(r,4,mv,tm);					// find min
		if (mv>MIN_ACQUIRE_DIST) return &done;	// check min. distance
		if (tm>1)	wside = RIGHT_WALL;			// decide wall
		else		wside = LEFT_WALL;
	} 
	
	float rot  = 0.0; 					// default to a straight course
	byte  sside[2][2] = { {2,3},{8,7}}; // sensor number to use for each side

	// tuned parameters for avoidance
	const float kp 			= 1.5 ;	// proportional part
	const float kd 			= 18.0;	// differential part
	const float diagfactor  = 1.4;	// accounts for diagonal sensors' angle
	const float turnmax     = 0.25;	// maximum turning rate
	const float avoidturn	= 0.25; // turning rate to avoid obstacles
	const float maxrange    = 3.0;	// maximum sensor range (in m)
	const float mindist		= FW_MINDIST;	// minimum allowed distance
	

	// read the sensors and subtract preferred distance
	float distm = robot->GetSonar(1);
	float distf = robot->GetSonar(sside[wside][0])-prefdist*diagfactor;	// diff to preferred distance
	float dists = robot->GetSonar(sside[wside][1])-prefdist; 				// diff to preferred distance
	float dist  = (dists<distf)?dists:distf;   				// find minimum distance

	// compute the rotational rate
	if (dist<maxrange) {
		rot = kp * dist + kd * (dist - olddist);
		olddist = dist;
	}
	if (wside==RIGHT_WALL) rot = -rot; // reverse if the right hand side

	// set the speed and rate
	float tm =  Limit(distm-mindist,0.0,transspeed);
	float ts =	Limit((distf+prefdist*diagfactor-mindist),0.0,transspeed);
	float r	=	Limit(-rot,-turnmax,turnmax);
	
	robot->SetRobotVelocity(ts,r);	// actuate!
	
	// actuate, avoid, if middle sensor more activated
//	if (tm>ts) {
//		robot->SetRobotVelocity(ts,r);
//	} else {
//		if (r>0)	robot->SetRobotVelocity(tm,avoidturn);
//		else		robot->SetRobotVelocity(tm,-avoidturn);
//	}
	
	
	

	return &done;
}

///////////////////////////////////////////////////////////////////////////
//   Return current location as waypoint

Here::Here(RobotProxy *robot_, Steppable *wp_)
	:robot(robot_),this_location(NULL),wp(wp_)
{
	assert(robot);
	done.Set(FUZZTRUE);
	this_location = new Waypoint(0.0,0.0);
}

Here::~Here()
{
	delete this_location;
}

boolean Here::Update(void)
{
	if ((!robot) || (!this_location)) return FALSE;
	float x,y,a;				// current coordinates
	robot->GetDRPosition(x,y,a);// get current coordinates
	this_location->Set(x,y,a);	// store current coordinates
	return TRUE;
}

Data *Here::Step(void)
{
	if (Update()) {
		if (wp) {
			Data *res = wp->Step();
			if (res) {
				if (res->type==DT_Waypoint) {
					((Waypoint*)res)->Set(this_location);
				}
			}
		}
		return this_location;
	}
	else return NULL;
}

///////////////////////////////////////////////////////////////////////////
//   Go directly to goal location

void Goto::FloorVector(
		RobotProxy *robot, Waypoint *goal_waypoint,						// inputs
		float &dist, float &heading, float &dattitude, boolean &dostop)	// outputs
{
	assert(robot);
	dist=9999999.999; 	// defensive programming !
	dostop = TRUE;
	if (!goal_waypoint || !robot) return;

	float ix,iy,ia;			// actual current location
	float dx,dy;			// delta to current location
	float goal_x,goal_y,goal_a;	// goal location

	// acquire all relevant data
	robot->GetDRPosition(ix,iy,ia);						// get real data
	goal_waypoint->Get(goal_x,goal_y,goal_a,dostop);	// get goal data
	
	// compute distance to target
	dx=goal_x-ix;			// x-distance to target
	dy=goal_y-iy;			// y-distance to target
	dist = sqrt(dx*dx+dy*dy); // compute distance to target
	
	// compute heading from current position to target position
	if (fabs(dist)<FLOAT_ZERO) heading = 0.0;
	else heading = acos(dx/dist);		// absolute heading to target
	if (dy<0) heading = -heading; 	// find correct quadrant of absolute heading
	heading -= ia;					// subtract current attitude for relative heading
	while (heading>PI) heading-=2*PI;	// correct for over/underflow of angle
	while (heading<-PI) heading+=2*PI;
	
	// compute difference in desired and actual attitude
	if (fabs(goal_a-ANYATTITUDE)<10.0) {
		dattitude=0.0;						// does attitude matter?
	} else {
		dattitude = goal_a - ia;			// attitude difference
		if (dattitude>PI) dattitude-=2*PI;	// correct for over/underflow of angle
		if (dattitude<-PI) dattitude+=2*PI;
	}
}

boolean Goto::Arrived(
	float dist, 	// inputs
	float dattitude, 
	boolean dostop,
	float &dthres,	// output thresholds used, for distance...
	float &athres)	// ...for angle
{
	// take decision
	if (dostop) {
		dthres = TARGET_THRES_STOP;
		athres = TARGET_A_THRES_STOP;
	} else {
		dthres = TARGET_THRES_NOSTOP;
		athres = TARGET_A_THRES_NOSTOP;
	}
	
	if (dist<=dthres && fabs(dattitude)<=athres) return TRUE;
	else return FALSE;
}

Data *Goto::Step(void) 
{
	if (waypoint) {
		Data *therealwaypoint = waypoint->Step();
		if (therealwaypoint) 
			if (therealwaypoint->type==DT_Waypoint) {
				Waypoint *goal_waypoint=(Waypoint*)therealwaypoint;
printf("Going to: ");				
goal_waypoint->Print();
printf("\n");				
				float dd,dh,da,dthres,athres;
				boolean dostop;
				FloorVector(myrobot,goal_waypoint,dd,dh,da,dostop);

				if (Arrived(dd,da,dostop,dthres,athres)) {
					myrobot->SetRobotVelocity(0.0,0.0);
					done.Set(FUZZTRUE);	
				} else {
					if (dd>dthres) {	// not close enough
						float ct,ca;	// current speeds (trans,angular)
						myrobot->GetRobotVelocity(ct,ca);
						float des_t = MAX_TRANS_V * (1-0.1/(dd+0.1)) / (dh*dh*ORIENT_FACT+1);
						des_t = Min(des_t,ct+ACCEL_T); // limit acceleration
						float des_a = Limit(-dh*TURN_FACT,-MAX_TURN_V,MAX_TURN_V);
						myrobot->SetRobotVelocity(des_t,des_a);	
					} else {			// close enough, but not oriented
						// turn to desired attitude
						myrobot->SetRobotVelocity(0,Limit(-da*TURN_FACT,-MAX_TURN_V,MAX_TURN_V));	
					}
					done.Set(FUZZFALSE);
				}
				return goal_waypoint;
			}
	}
	// stop robot and return NULL (because of error)
	//    possible error conditions: no waypoint, no retval, wrong type
	done.Set(FUZZTRUE);	
	myrobot->SetRobotVelocity(0.0,0.0);
	return NULL;
}

void Goto::Reset(void) 
{
	done.Set(FUZZFALSE);
	if (waypoint) waypoint->Reset();
}

void Goto::Abort(void) 
{
	done.Set(FUZZTRUE);
	if (waypoint) waypoint->Abort();
}

/////////////////////////////////////////////////////////////
//  Turn towards a particular point (similar to "Goto")

Data *Turnto::Step(void) 
{
	if (waypoint) {
		Data *therealwaypoint = waypoint->Step();
		if (therealwaypoint) 
			if (therealwaypoint->type==DT_Waypoint) {
				Waypoint *goal_waypoint=(Waypoint*)therealwaypoint;
				float dd,dh,da;
				boolean dostop;
				FloorVector(myrobot,goal_waypoint,dd,dh,da,dostop);

				if (fabs(dh)<TARGET_A_THRES_STOP) 	done.Set(FUZZTRUE);	
				else								done.Set(FUZZFALSE);

				myrobot->SetRobotVelocity(0.0,Limit(-dh*TURN_FACT,-MAX_TURN_V,MAX_TURN_V));	
				return goal_waypoint;
			}
	}
	// stop robot and return NULL (because of error)
	//    possible error conditions: no waypoint, no retval, wrong type
	done.Set(FUZZTRUE);	
	myrobot->SetRobotVelocity(0.0,0.0);
	return NULL;
}

/////////////////////////////////////////////////////////////
//  Turn by this amount of rads

Turnby::Turnby(RobotProxy *robot_, Steppable *da_step_)
	:Moveby(robot_,da_step_)
{
}

Turnby::Turnby(RobotProxy *robot_, float da_)
	:Moveby(robot_,da_)
{
}

Data *Turnby::Step(void) 
{
	if (!robot) return &fuzz_false;
	if (done.True()) {
		robot->SetRobotVelocity(0.0,0.0);	// stop 'cause it's done
		return &fuzz_false;
	}
	
	
	if (dist_step != NULL) {
		Data *ddist = dist_step->Step();
		if (ddist!=NULL) {
			if (ddist->type==DT_Number) {
				dist = ((Number*)ddist)->Get();
printf("Acquiring dist...");				
			}
		}
	}
printf("Turndist: %f\n",dist);	
	
	float x,y,a;
	robot->GetDRPosition(x,y,a);	// get current position
	
	float da = (started_a + dist) - a;	// compute "distance"
	while (da < -PI) da += PI*2.0;		// wrap to circle
	while (da >  PI) da -= PI*2.0;
	
	if (fabs(da) < UNBUMP_A_THRES) {
		done.Set(FUZZTRUE);					// done, stop !
		robot->SetRobotVelocity(0.0,0.0);
		return &fuzz_false;
	} else {
		robot->SetRobotVelocity(0.0,-Limit(da,-UNBUMP_A,UNBUMP_A));
		return &fuzz_true;
	}
}

Data *TurntoAttitude::Step(void) 
{
	if (!robot) return &fuzz_false;
	if (done.True()) {
		robot->SetRobotVelocity(0.0,0.0);	// stop 'cause it's done
		return &fuzz_false;
	}
	
	
	if (dist_step != NULL) {
		Data *ddist = dist_step->Step();
		if (ddist!=NULL) {
			if (ddist->type==DT_Number) {
				dist = Limit(((Number*)ddist)->Get(),-PI,PI);
			}
		}
	}
	
	float x,y,a;
	robot->GetDRPosition(x,y,a);	// get current position
	
	float da = dist - a;			// compute "distance"
	while (da < -PI) da += PI*2.0;	// wrap to circle
	while (da >  PI) da -= PI*2.0;
	
	if (fabs(da) < UNBUMP_A_THRES) {
		done.Set(FUZZTRUE);					// done, stop !
		robot->SetRobotVelocity(0.0,0.0);
		return &fuzz_false;
	} else {
		robot->SetRobotVelocity(0.0,-Limit(da,-UNBUMP_A,UNBUMP_A));
		return &fuzz_true;
	}
}


/////////////////////////////////////////////////////////////
//  Move by this amount of meters

Moveby::Moveby(RobotProxy *robot_, float dist_)
	:robot(robot_),dist_step(NULL),dist(dist_),started_x(0.0),started_y(0.0),started_a(0.0)
{
	done.Set(FUZZFALSE);
}

Moveby::Moveby(RobotProxy *robot_, Steppable *dist_step_)
	:robot(robot_),dist_step(NULL),dist(0.0),started_x(0.0),started_y(0.0),started_a(0.0)
{
	dist_step=Getport(dist_step_);
	assert(dist_step);
	done.Set(FUZZFALSE);
}

void Moveby::Reset(void) 
{
	done.Set(FUZZFALSE);
	if (! robot) return;
	robot->GetDRPosition(started_x,started_y,started_a);
	robot->SetRobotVelocity(0.0,0.0);
}

void Moveby::Abort(void) 
{
	done.Set(FUZZTRUE);
	if (! robot) return;
	robot->SetRobotVelocity(0.0,0.0);
}

Data *Moveby::Step(void) 
{
	if (!robot) return &fuzz_false;
	if (done.True()) {
		robot->SetRobotVelocity(0.0,0.0);	// stop 'cause it's done
		return &fuzz_false;
	}
	
	if (dist_step != NULL) {
		Data *ddist = dist_step->Step();
		if (ddist!=NULL) {
			if (ddist->type==DT_Number) {
				dist = ((Number*)ddist)->Get();
			}
		}
	}
	
	float x,y,a;
	robot->GetDRPosition(x,y,a);	// get current position
	float dx = x - started_x;
	float dy = y - started_y;
	float dd = sqrt(dx*dx+dy*dy);	// compute distance to org
	
	if (dd<fabs(dist)) {
		if (dist<0.0) 	robot->SetRobotVelocity(-UNBUMP_V,0.0); // backward
		else			robot->SetRobotVelocity(UNBUMP_V,0.0);	// forward
		return &fuzz_true;
	} else {
		done.Set(FUZZTRUE);					// done, stop !
		robot->SetRobotVelocity(0.0,0.0);
		return &fuzz_false;
	}
}

/////////////////////////////////////////////////////////////
//
//    D A T A   for   Navigation
//

Route::Route(char *filename_)
	:filename(filename_),Data(DT_Route)
{
	(void)ReadRoute();
	ptlist.Reset();
}

Route::Route(Route *template_)
	:filename(NULL),Data(DT_Route)
{
	Set(template_);
	ptlist.Reset();
}

void Route::Set(Route *template_)
{
	if (! template_) return; // if no template stay empty
	DeleteAll();	// delete all stored waypoints
	Waypoint *wp;
	List *tl = &(template_->ptlist);
	tl->Reset();
	while (NULL!=(wp = (Waypoint*) tl->Iterate())) 
		Append((Waypoint*)wp->Duplicate());		// make new copies !!!
}

void Route::ResetRoute(void)
{
	ptlist.Reset();
}

Waypoint *Route::Iterate(void)
{
	return (Waypoint*)ptlist.Iterate();
}

Waypoint *Route::GetLast(void)
{
	return (Waypoint*)ptlist.FindLast();
}

Waypoint *Route::GetCurrent(void)
{
	return (Waypoint*)ptlist.FindCurrent();
}

Waypoint *Route::Reverse(void)
{
	ptlist.Reverse();
	return (Waypoint*)ptlist.FindCurrent();
}





boolean Route::ReadRoute(void)
{
	if (! filename) return FALSE;
	
	FILE *mapfile;
	char txt[200];
	float x,y,a;
	char s;
	
	cout << "Attempting to open waypoint-file " << filename << "\n";
	if ((mapfile
			=fopen(filename,"r"))==NULL) {
		cerr << "ERROR: could not open hiwaymap file (in Route::ReadRoute)\n";
		return FALSE;
	}
	
	// read in initial fields
	fscanf(mapfile," %s %s %s %s %s %s",txt,txt,txt,txt,txt,txt);
	DeleteAll(); // destroy old list of waypoints
	
	// read in all the lines
	while (1) {
		if (5!=fscanf(mapfile,"%s %f %f %f %c",txt,&x,&y,&a,&s)) break;
		printf("Scanned: %s %f %f %f %s\n", 
			txt, x,y,a,(s=='Y' || s=='y')?"STOP":"NOSTOP");
		Append(new Waypoint(x,y,a,(s=='Y' || s=='y')));
	}
	
	fclose(mapfile);
	printf("Scanned %d lines \n",ptlist.Size());
	if (0<Size()) return TRUE;
	else return FALSE;
}

Data *Route::Print(void) 
{
	printf("Route(with %d points: ",ptlist.Size());
	Reset();
	Waypoint *curr;
	while (NULL!=(curr=Iterate())) curr->Print();
	printf(")");
	return this;
}



/////////////////////////////////////////////////////////////
//
//    HiwayMap   for   Navigation
//

#define HWM_RANDSEED   1

HiwayMap::HiwayMap(char *filename_)
	:the_map(NULL),the_route(NULL),filename(filename_),tot_dist(0.0),random_pt(NULL),Data(DT_HiwayMap)
{
	ReadMap();
}

HiwayMap::HiwayMap(HiwayMap *template_)
	:the_map(NULL),the_route(NULL),filename(NULL),tot_dist(0.0),random_pt(NULL),Data(DT_HiwayMap)
{
	Set(template_);
};

void HiwayMap::ReadMap(void)
{
	assert(filename);
	if (!filename)return;
	
	// try to read file and check whether it worked
	UnGraph *nhwm = new UnGraph(filename);
	if (nhwm->edges.Size()<1) {
		delete nhwm;
		return;
	}
	
	// if eveything went ok, delete old one, and assign new one
	delete the_map;
	the_map = nhwm;
	return;
}

void HiwayMap::Set(HiwayMap *template_)
{
	assert(template_);
	if (! template_) return;
	delete the_map;
	the_map = template_->the_map->Duplicate();
}

Data *HiwayMap::Print(void)
{
	if (! the_map)
		printf("HiwayMap(empty)");
	else
		printf("HighwayMap(%d nodes, %d edges, %d walls\n",
			the_map->nodes.Size(),
			the_map->edges.Size(),
			the_map->wedges.Size());
	return this;
}

Route *HiwayMap::GetRoute(Waypoint *wp_start, Waypoint *wp_goal)
{
	assert(wp_start);
	assert(wp_goal);
	assert(the_map);
	if ((!the_map) || (!wp_start) || (!wp_goal)) return NULL;
	
	// find closest points
	float sx,sy,sa,gx,gy,ga;	// start/goal coordinates
	float sdist, gdist;			// initial and final leg distances
	boolean ss,gs;		// must I stop there?

	wp_start->Get(sx,sy,sa,ss);
	wp_goal->Get(gx,gy,ga,gs);
	GraphNode *psn = the_map->GetClosestNode(sdist,sx,sy,10000.0);
	GraphNode *pgn = the_map->GetClosestNode(gdist,gx,gy,10000.0);
	if (!psn || !pgn) return NULL;
	
	// plan a path
	GraphPath *the_path = the_map->GetShortestPath(psn,pgn);
	if (!the_path) return NULL;
	
	// transform to route
	delete the_route;	// delete the previous route and make new one
	the_route = new Route;
	the_path->edges.Reset();
	GraphEdge *eptr;
	the_route->Append(new Waypoint(sx,sy)); // initial leg
	while (NULL!=(eptr = (GraphEdge*)the_path->edges.Iterate())) {
		assert(eptr->to_node);
		the_route->Append(new Waypoint(eptr->to_node->x,eptr->to_node->y));
printf("Planned to waypoint %f/%f\n",	eptr->to_node->x,eptr->to_node->y);	
	}
	the_route->Append(new Waypoint(gx,gy,ga,gs)); // final leg
	tot_dist = sdist + the_path->length + gdist;	// compute length 
	delete the_path;	// deallocate the new path
	
	// return the stuff
	return the_route;
}

Waypoint	*HiwayMap::GetRandomPoint(void)
{
	const float assume_attitude = ANYATTITUDE;

	if (!the_map) return NULL;
	List	*nodes = &(the_map->nodes);
	unsigned int nomappts = nodes->Size();	// number of points to choose from
	unsigned int draw	  = (rand() % nomappts);
	
	GraphNode *curr;
	nodes->Reset();
	unsigned int cntr=draw;
	while (NULL!=(curr=(GraphNode*)nodes->Iterate())) {
		if (cntr==0) {
printf("\n\nGoint to random point %f %f\n\n",curr->x,curr->y);
			if (random_pt!=NULL)	random_pt->Set(curr->x,curr->y,assume_attitude);
			else					random_pt = new Waypoint(curr->x,curr->y,assume_attitude);
			return random_pt;
		}
		cntr--;
	}

	return NULL;
}

/////////////////////////////////////////////////////////////
//
//    Update position of robot based on estimated wall pos
//

Data *UpdateDRPWall::Step(void)
{
	// get wall estimate
	if (!est) return NULL;
	Data *ret_est = est->Step();
	if (! ret_est) return NULL;
	if (ret_est->type != DT_WallEstimate) return NULL;
	WallEstimate	*rest	= (WallEstimate*)ret_est;
	
	// get wall
	if (! wall) return NULL;
	Data *ret_wall = wall->Step();
	if (! ret_wall) return NULL;
	if (ret_wall->type != DT_Wall) return NULL;
	Wall *rwall = (Wall*)ret_wall;

	// get a,b,c from estimate and this wall
	float ma,mb,mc,ea,eb,ec;	// wall parameters
	rwall->Get(ma, mb, mc);		// parameters of world model wall
	if (! rest->Estimate(ea, eb, ec)) 
		return NULL;			// parameters of estimated (real) wall

printf("ABC MODEL (%f,%f,%f), ABC EST(%f,%f,%f)\n",ma,mb,mc,ea,eb,ec);
	
	// get current position of robot
	float x,y,a;
	robot->GetDRPosition(x,y,a);
	
	// compute angle and distances
	double nm = sqrt(ma*ma+mb*mb);	// norm of normal vector of model wall
	double ne = sqrt(ea*ea+eb*eb);	// norm of normal vector of estimated wall
	double dm = (ma*x+mb*y+mc)/nm;	// distance to model wall
	double de = (ea*x+eb*y+ec)/ne;	// distance to estimated wall
printf("nm %f ne %f dm %f de %f\n",nm,ne,dm,de);
	// compute correction
	double da = fabs(acos((ma*ea+mb*eb)/(nm*ne)));// angle between the normal vectors
	if ((ea*mb-eb*ma)<0.0) da = -da;// determine orientation
	if (fabs(da)>PI2) {				// correct for orientation
		dm = -dm;
		ma = -ma;
		mb = -mb;
	}
	while (da>PI2)  da-=PI;			// correct 180 deg problem
	while (da<-PI2) da+=PI;
	double dd = (de-dm)/nm;			// correction factor
	double dx = ma*dd;				// correction in x
	double dy = mb*dd;				// correction in y
	
printf("da %f dd %f\n",da,dd);	
	double nx = x + dx;				// compute new values
	double ny = y + dy;
	double na = a + da;
	while (na < -PI) na += PI;		// limit range of angle of attitude
	while (na >  PI) na -= PI; 
	
	// update dead reckoning position in robot
	robot->SetDRPosition(nx,ny,na);

printf("Updating from %f,%f,%f by %f,%f,%f to %f,%f,%f deg\n",
	x,y,a/PI*180.0,
	dx,dy,da/PI*180.0,
	nx,ny,na/PI*180.0);
	
	return rest;
}

/////////////////////////////////////////////////////////////
//
//    Estimate location of wall
//

EstimateWall::EstimateWall(RobotProxy *robot_, byte sonarno_, Steppable *est_)
	:robot(robot_),sonarno(sonarno_),est(NULL)
{
	est=Getport(est_);
	assert(robot && est);
	assert(sonarno>0 && sonarno<9);
	inp_ok = (robot && sonarno>0 && sonarno<9 && est);
	Reset();
}

void EstimateWall::Reset(void)
{
	done.Set(FUZZFALSE);
}

void EstimateWall::Abort(void)
{
	done.Set(FUZZTRUE);
}

Data *EstimateWall::Step(void)
{
	done.SetTrue();
	
	if (! inp_ok) return NULL;
	
	// get the data
	Data *retval = est->Step();
	if (! retval) return NULL;
	if (retval->type != DT_WallEstimate) return NULL;
	WallEstimate *rest = (WallEstimate*)retval;
	
	// get sonar reading
	float x,y,a;
	double dist = robot->GetSonar(sonarno) + SONAR_RADIUS;
	robot->GetDRPosition(x,y,a);
	
	// calculate end position of sonar beam and update the estimate
	double angle = a + PI4*(sonarno-1);	// compute global angle of sonar
	rest->Update(x,y,dist,angle);
	
	if (rest->NoSamples()<20) done.SetFalse();
	
	return rest;
}

Data	*ResetWallEstimate::Step(void)
{
	if (!est) return NULL;
	Data *retval = est->Step();
	if (! retval) return NULL;
	if (retval->type != DT_WallEstimate) return NULL;
	WallEstimate *rest = (WallEstimate*)retval;
	rest->Init();
	return rest;
}


/////////////////////////////////////////////////////////////
//
//    Estimate bumped location of wall
//

EstimateBumpWall::EstimateBumpWall(RobotProxy *robot_, Steppable *est_)
	:EstimateWall(robot_,1,est_)
{}

Data *EstimateBumpWall::Step(void)
{
	done.SetTrue();
	
	if (! inp_ok) return NULL;
	
	// get the data
	Data *retval = est->Step();
	if (! retval) return NULL;
	if (retval->type != DT_WallEstimate) return NULL;
	WallEstimate *rest = (WallEstimate*)retval;
	
	// get sonar reading
	float x,y,a;
	robot->GetDRPosition(x,y,a);
	
	// calculate end position of sonar beam and update the estimate
	rest->Update(x,y,SONAR_RADIUS,a);
	
	if (rest->NoSamples()<1) done.SetFalse();
	
	return rest;
}

/////////////////////////////////////////////////////////////
//
//    Detect an edge in a wall
//

EdgeDetected::EdgeDetected(RobotProxy *robot_)
	:robot(robot_),sonarno(33),state(-1),sample_count(-33),
	curr_est(ANYCOORD,ANYCOORD,0),prev_est(ANYCOORD,ANYCOORD,0)
{
	assert(robot);
	Reset();
}

void EdgeDetected::Reset(void)
{
	state=1;			// reset to first state
	done.Set(FUZZFALSE);
	detected.SetFalse();
}

void EdgeDetected::Abort(void)
{
	done.Set(FUZZTRUE);
}

Data *EdgeDetected::Step(void)
{
	// misc. constants
	const float max_range 	= 1.5;		// upper limit for accepted sonar data
	const float min_range 	= 0.45;		// lower limit for accepted sonar data
	const int	no_samples 	= 30;		// number of samples per frame
	const float thres_dist	= 0.3;		// distance threshold
	const float thres_ang	= PI8*1.5;	// angle threshold
	const byte  leftsonar	= 3;
	const byte	rightsonar	= 7;

	if (! robot) {done.SetTrue();return NULL;}
	
	float l,r;				// left and right sonar ranges
	float x,y;				// DR position
	float d,a;				// sonar data
	
	switch (state) {
		case	1:			// tying to lock onto target
			detected.SetFalse();
			l = robot->GetSonar(leftsonar);
			r = robot->GetSonar(rightsonar);
			if (l>max_range && r>max_range) break; // cannot aquire
			if (l<r) 	sonarno=leftsonar;
			else		sonarno=rightsonar;
printf("Selected Sonar %d for checking\n",sonarno);			
			curr_est.Init();
			sample_count=0;
			state++;
			break;
		case	2:			// acquiring first frame of data
			robot->GetDRPosition(x,y,a);
			d = robot->GetSonar(sonarno);
			curr_est.Update(x,y,d+HULL_RADIUS,a+((sonarno==leftsonar)?PI2:-PI2));
			if (d<min_range || d>max_range) break;
			sample_count++;
			if (sample_count>=no_samples) {
				prev_est.Set(&curr_est);
				sample_count=0;
				curr_est.Init();
				state++;
printf("\nFirst frame compledted, wall at distance %f\n",prev_est.Dist(x,y));				
			}
			break;
		case	3:			// acquiring data frame (general)
			detected.SetFalse();
			robot->GetDRPosition(x,y,a);
			d = robot->GetSonar(sonarno);
			curr_est.Update(x,y,d+HULL_RADIUS,a+((sonarno==leftsonar)?PI2:-PI2));
			if (d<min_range || d>max_range) break;
			sample_count++;
			if (sample_count>=no_samples) {
				// decide whether new wall was found
				float pa,pb,pc,pd,ca,cb,cc,cd;
				prev_est.Estimate(pa,pb,pc);	// get estimated data
				pd = prev_est.Dist(x,y);
				curr_est.Estimate(ca,cb,cc);
				cd = curr_est.Dist(x,y);
				float pang	=	atan2(pb,pa);	// compute orientation
				float cang	=	atan2(cb,ca);
				float adiff	=	pang-cang;		// compute "real" angle diff
				while (adiff<-PI2) adiff+=PI;
				while (adiff> PI2) adiff-=PI;
printf("\nNext frame completed:\n");
printf("		old wall: %f,%f,%f -> dist(%f),ang(%f)\n",pa,pb,pc,pd,pang/3.1415936*180);		
printf("		new wall: %f,%f,%f -> dist(%f),ang(%f)\n",ca,cb,cc,cd,cang/3.1415936*180);		
				if  ((fabs(fabs(pd)-fabs(cd))>=thres_dist) || // test for difference
					 (fabs(adiff)>=thres_ang)) {
					detected.SetTrue();
printf("\n##### Edge detected ######\n");					
				}
				// restart the process
				sample_count=0;
				prev_est.Set(&curr_est);
				curr_est.Init();
			}
			break;
		default:
			assert(-100<state);		// this statement must not be reached
			break;
	};
	
	done.Set(&detected);
	return &detected;
}



/////////////////////////////////////////////////////////////
//
//    Data containing estimate of location of wall
//

void WallEstimate::Update(float x, float y)
{
	if (x_est>ANYCOORD+1) if (fabs(x-x_est)>thres) {
printf("Rejected est point %f / %f \n",x,y);
		return; // reject ?
	}
	if (y_est>ANYCOORD+1) if (fabs(y-y_est)>thres) {
printf("Rejected est point %f / %f \n",x,y);
		return; // reject ?
	}

printf("Updated wall statistics with point %f / %f \n",x,y);
	// update statistics
	sx	+= x;
	sy	+= y;
	sxy	+= x*y;
	sx2 += x*x;
	sy2	+= y*y;
	n++;
printf("Params now: %f %f %f %f %f %d\n",sx,sy,sxy,sx2,sy2,n);	
}

void WallEstimate::Update(float x, float y, float d, float a)
{
	double ex = d * cos(a) + x;	// endpoint of sonar beam (x)
	double ey = d * sin(a) + y;	// endpoint of sonar beam (y)
	Update(ex,ey);
}

void WallEstimate::Init(void)
{
	n=0;
	sx=sy=sxy=sx2=sy2=0.0;
printf("Params inited to: %f %f %f %f %f %d\n",sx,sy,sxy,sx2,sy2,n);	
}

boolean WallEstimate::Estimate(float &a, float &b, float &c)
{
	// compute and return estimate
	// the line's equation is ax+by+c=0
	if (n>=MIN_WALL_SAMPLES) {
		double ax = sx/(double)n;	// average x
		double ay = sy/(double)n;	// average y
		a  = ea = (sxy - sx*ay);	// a of equation
		b  = eb = (sx*ax-sx2);		// b of equation
		c  = ec = (-b*ay - a*ax);	// c of equation
printf("Returning an estimate of a %f / b %f / c %f\n",a,b,c);		
printf("With final params: %f %f %f %f %f %d\n",sx,sy,sxy,sx2,sy2,n);	
		return TRUE;
	} else {
		return FALSE;
	}
}

float WallEstimate::Dist(float x_, float y_)
{
	float a,b,c;
	if (! Estimate(a,b,c)) return 999999.87654321;	// dangerous !!!!
	return (a*x_ + b*y_ + c)/sqrt(a*a + b*b);
}

void WallEstimate::Set(WallEstimate *template_)
{
	if (! template_) return;
	sx	=	template_->sx;
	sy	=	template_->sy;
	sxy	=	template_->sxy;
	sx2	=	template_->sx2;
	sy2	=	template_->sy2;
	n	=	template_->n;
	ea	=	template_->ea;
	eb	=	template_->eb;
	ec	=	template_->ec;
}

void Wall::Set(float x1_, float y1_, float x2_, float y2_)
{
								// line goes ..
	x1	=	x1_;				// .. from this point ..
	y1	=	y1_;
	x2	=	x2_;				// .. to this point.
	y2	=	y2_;
	a	=	y1 - y2;			// compute line parameters: normal vector...
	b	=	x2 - x1;
	c	=	-( a*x1 + b*y1 );	// .. and offset.
	
	assert(fabs(a*a+b*b)>0.01);	// sensible parameters?
}

void Wall::Set(Wall *template_)
{
	if (! template_) return;
	x1	=	template_->x1;
	y1	=	template_->y1;
	x2	=	template_->x2;
	y2	=	template_->y2;
	a	=	template_->a;
	b	=	template_->b;
	c	=	template_->c;
}


////////////////////////////////////////////////////////////////////////
//  Avoid using sonars
//    - returns a Fuzz indicating whether avoidance was necessary or not
//    - using a kind of braitenberg-style avoidance scheme

SonarAvoid::SonarAvoid(RobotProxy *robot_,float trans_)
	:robot(robot_),def_trans(trans_)
{
	assert(robot_);
	done.Set(FUZZFALSE);
	a_side	=	ParamA(0.45);	// side sonars	
	b_side	=	ParamB(0.45);
	a_front	=	ParamA(0.65);	// front sonars
	b_front	=	ParamB(0.65);
	a_mid	=	ParamA(0.7);	// mid front sonar
	b_mid	=	ParamB(0.7);
}

// calculate activation parameters
// formula is:   a = 1/(min-max)
//               b = max/(min-max)
float SonarAvoid::ParamA(float max_,float min_){return 1.0/(min_-max_);}
float SonarAvoid::ParamB(float max_,float min_){return -max_/(min_-max_);}

Data *SonarAvoid::Step(void)
{
	assert(def_trans>0.0 && def_trans<0.5);	// reasonable values
	if (!robot) return NULL;
	const float nosens=5.0; 	// number of sensors (as float)
	const float acthre=0.001;	// activation threshold
	const float	rotspe=0.5;		// rotation speed, when fully activated
	const float	trafac=1.0;		// factor for translation from activation

	// calculate activation of "virtual tentacles"
	float l		=	Limit(robot->GetSonar(3)*a_side+b_side,0.0,1.0); 
	float lf	=	Limit(robot->GetSonar(2)*a_front+b_front,0.0,1.0);
	float m		=	Limit(robot->GetSonar(1)*a_mid+b_mid,0.0,1.0);
	float rf	=	Limit(robot->GetSonar(8)*a_front+b_front,0.0,1.0);
	float r		=	Limit(robot->GetSonar(7)*a_side+b_side,0.0,1.0); 
	
	float sum	=	l + lf + m + rf + r;
		
	boolean triggered = (sum > acthre) ? TRUE : FALSE ;
	
	if (triggered) {
		float rotsig	=	((l+lf-rf-r)>0.0)?1.0:-1.0;		// determine sign of rotation
		float transact	=	Limit(1.0-sum*trafac,0.0,1.0);	// activation of translation
		float trans		=	transact*def_trans;				// translation speed
		float rot 		=	rotsig*rotspe;					// rotation speed
		
		robot->SetRobotVelocity(trans,rot);					// set speeds
	
		done.Set(FUZZFALSE);
		return &fuzz_true;
	} else {
		done.Set(FUZZTRUE);	// is done if not activated
		return &fuzz_false; // return false, because did not avoid
	}
}

////////////////////////////////////////////////////////////////////////
//  Avoid forbidden areas using "virtual antlers"
//    - returns a Fuzz indicating whether avoidance was necessary or not
//    - using a kind of braitenberg-style avoidance scheme

Data *AvoidForbidden::Step(void)
{
	assert(dist>0.0 && dist<2.0);	// reasonable values

	const float spread	=	0.4;	// angle to spread from axis
	const float rotspe	=	0.3;	// rotation speed to avoid
	const float traspe	=	0.1;	// translation speed to avoid

	// get the map
	if (!hiwaymap) return NULL;
	Data *map_data = hiwaymap->Step();
	if (!map_data) return NULL;
	if (map_data->type != DT_HiwayMap) return NULL;
	HiwayMap *map = (HiwayMap*)map_data;
	
	// get current position
	float x,y,a;
	robot->GetDRPosition(x,y,a);
	
	// compute position of virtual antlers
	float x_left	= x + cos(a+spread);	
	float y_left	= y + sin(a+spread);
	float x_right	= x + cos(a-spread);
	float y_right	= y + sin(a-spread);
	
	// check whether they are in forbidden areas
	boolean left 	= map->Forbidden(x_left,y_left);
	boolean right	= map->Forbidden(x_right,y_right);
	
	// compute motor activity
	if (left || right) {
		float t = (left && right)?0.0:traspe;	// stand still if both forbidden
		float r = (left)?rotspe:(-rotspe);		// turn away (right is default)
		robot->SetRobotVelocity(t,r);	// set speeds
		done.Set(FUZZFALSE);
		return &fuzz_true;
	} else {
		done.Set(FUZZTRUE);	// is done if not activated
		return &fuzz_false; // return false, because did not avoid
	}
}

////////////////////////////////////////////////////////////////////////////////////
// Do I have to recalibrate DR?

RecalibQ::RecalibQ(RobotProxy *robot_, Steppable *calib_loco_, float dist_thres_)
	:robot(robot_),calib_loco(calib_loco),dist_thres(dist_thres_)
{
	calib_loco=Getport(calib_loco_);
	assert(dist_thres>0.0);
	assert(robot);
	done.SetFalse();
}	

Data	*RecalibQ::Step(void)
{
	Waypoint *recalpt;
	GETVALRET(Waypoint,recalpt,calib_loco);
	float tot_dist,x,y,a,rx,ry,ra;
	boolean rs;
	
	// get the data
	robot->GetDRPosition(x,y,a);
	tot_dist = robot->GetDRDistT();
	recalpt->Get(rx,ry,ra,rs);
	
	// compute the propensity
	float dx = rx-x;
	float dy = ry-y;
	float dd = sqrt(dx*dx+dy*dy);
	
	float p = Limit((dd+tot_dist)/dist_thres,0.0,1.0);
	
	propensity.Set(p);
	return &propensity;
}
	
void RecalibQ::Reset(void)
{
	if (calib_loco) calib_loco->Reset();
}

void RecalibQ::Abort(void)
{
	if (calib_loco) calib_loco->Abort();
}

Fuzz *RecalibQ::IsDone(void)
{
	if (calib_loco) return calib_loco->IsDone();
	return &fuzz_true;
}

Data *RefuelQ::Step(void)
{
	if (!robot)return NULL;
	
	float b = robot->GetBattery();		// get battery state [0..1]
	float p = Limit(1.0-2*b,0.0,1.0);	// compute propensity
	propensity.Set(p);
	return &propensity;
}
	


