/*

DISCLAIMER 

InSitu is provided as is, without any kind of guarantees or warranty. There is 
no documentation. The author does not provide any kind of support to users of 
this source code and is not responsible for any kind of problems or damages 
resulting from its use. 

The source code is the result of a long iterative design process for a research 
prototype and was never intended to be used in any kind of product. In general, 
the code cannot be expected to compile in any straightforward manner on your 
computer system, since the robot on which it was developed, has a custom design. 

*/

#ifndef _Navsup_h_
#define _Navsup_h_

/* Rene Schaad                                                    */
/* Departement of Computer Science, Univ. of Zurich, Switzerland  */
/* schaad@ifi.unizh.ch                                            */

#include "PrgStruct.h"
#include "RobotProxy.h"
#include "Graph.h"

/////////////////////////////////////////////////////////////
//
//    D a t a   for     Navigation
//

#define ANYATTITUDE   -10000.0
#define ANYCOORD      -10000.0

class Waypoint : public Data {
public:
	Waypoint(float x_=0.0, float y_=0.0, float a_=ANYATTITUDE, boolean dostop_=FALSE)
		:x(x_),y(y_),a(a_),dostop(dostop_),Data(DT_Waypoint){};
	
	// manipulation
	void	Set(float x_, float y_, float a_=ANYATTITUDE, boolean dostop_=FALSE)
				{x=x_;y=y_;a=a_;dostop=dostop_;};
	void	Set(Waypoint *newwp)	{if (newwp) {x=newwp->x;y=newwp->y;a=newwp->a;dostop=newwp->dostop;};};
	void	Get(float &x_, float &y_, float &a_, boolean &dostop_)	
				{x_=x;y_=y;a_=a;dostop_=dostop;};
				
	virtual boolean	SetData(Data *org_){
		if (!org_) return FALSE;
		if (org_->type!=type) return FALSE;
		Set((Waypoint*)org_);
		return TRUE;
	};

	virtual Data	*Duplicate(void){return new Waypoint(x,y,a,dostop);};
	virtual Data	*Print(void){printf("Waypoint(%f,%f,%f,%s)",x,y,a,(dostop?"STOP":"NONSTOP"));return this;};
private:
	float	x,y,a; // cartesian coordinates of waypoint (with attitude)
	boolean dostop;   // stop at this waypoint?
};

class Route : public Data {
public:
	Route(char *filename_);
	Route(Route *template_=NULL);
	~Route(){ptlist.DeleteAll();}

	// manipulation (route _OWNS_ its waypoints!!)
	void	Set(Route *template_);
	void		ResetRoute(void);	// ;get the first waypoint
	Waypoint	*GetLast(void);		// get the last waypoint
	Waypoint	*GetCurrent(void);	// get the current
	Waypoint 	*Iterate(void);		// get the next waypoint
	Waypoint	*Reverse(void);		// reverse order of waypoints
	
	boolean		Append(Waypoint *npt){return ptlist.Append(npt);};	// append a new waypoint
	boolean		Prepend(Waypoint *npt){return ptlist.Prepend(npt);};// add a new wp to beginning
	void		DeleteAll(void){ptlist.DeleteAll();};		// clear all waypoints
	unsigned int Size(void) {return ptlist.Size();};
	
	
	virtual boolean	SetData(Data *org_){
		if (!org_) return FALSE;
		if (org_->type!=type) return FALSE;
		Set((Route*)org_);
		return TRUE;
	};

	virtual Data	*Duplicate(void){return new Route(this);};
	virtual Data	*Print(void);


private:
	boolean ReadRoute(void);
	List ptlist;
	char *filename;	// name of the file of the waypoints
};

/////////////////////////////////////////////////////////////
//
//    Hiwaymap   for   Navigation
//

class HiwayMap : public Data {
public:
	HiwayMap(char *filename_);
	HiwayMap(HiwayMap *template_);
	~HiwayMap(){delete the_map; delete the_route;delete random_pt;};

	// manipulation
	void	Set(HiwayMap *template_);
	
	virtual boolean	SetData(Data *org_){
		if (!org_) return FALSE;
		if (org_->type!=type) return FALSE;
		Set((HiwayMap*)org_);
		return TRUE;
	};

	virtual Data	*Duplicate(void){return new HiwayMap(this);};
	virtual Data	*Print(void);
	
	// Map handling
	Route		*GetRoute(Waypoint *wp_start, Waypoint *wp_goal);
	float		GetDist(void){return tot_dist;};	// get distance of route 
	Waypoint	*GetRandomPoint(void);				// get any point on map
	
	// Forbidden area handling
	boolean	Forbidden(float x, float y)		// is this point in a forbidden zone? (conservative!)
		{if (the_map) return the_map->Forbidden(x,y); else return TRUE;};

private:
	void ReadMap(void);  
	UnGraph	*the_map;		// my very own highwaymap of the environment
	Route	*the_route;		// the most recently generated route
	char *filename;			// name of the file of the waypoints
	float tot_dist;			// total length of the_route
	Waypoint *random_pt;	// random point to navigate to
};

class Wall : public Data {
public:
	Wall(float x1_, float y1_, float x2_, float y2_)
		:a(0.0),b(0.0),c(0.0),
		 x1(0.0),y1(0.0),x2(0.0),y2(0.0),
		 Data(DT_Wall)
		{Set(x1_,y1_,x2_,y2_);};
	Wall(Wall *org_=NULL)
		:a(0.0),b(0.0),c(0.0),
		 x1(0.0),y1(0.0),x2(0.0),y2(0.0),
		 Data(DT_Wall)
		{Set(org_);};
		
	// data manipulation
	void	Get(float &wa, float &wb, float &wc)
		{wa=a;wb=b;wc=c;};
	void	Set(float x1_, float y1_, float x2_, float y2_);
		
	// manipulation (route _OWNS_ its waypoints!!)
	void	Set(Wall *template_);
	virtual boolean	SetData(Data *org_){
		if (!org_) return FALSE;
		if (org_->type!=type) return FALSE;
		Set((Wall*)org_);
		return TRUE;
	};

	virtual Data	*Duplicate(void){return new Wall(this);};
	virtual Data	*Print(void){printf("Wall(%f,%f->%f,%f)",x1,y2,x2,y2);return this;};

private:
	// estimated wall parameters
	float		a,b,c;	// estimated a,b,c
	float		x1,y1;	// endpoint one
	float		x2,y2;	// endpoint two
};

class WallEstimate : public Data {
public:
	WallEstimate(float x_est_,float y_est_,float thres_,WallEstimate *org_=NULL)
		:sx(0.0),sy(0.0),sxy(0.0),sx2(0.0),sy2(0.0),n(0),
		 ea(0.0),eb(0.0),ec(0.0),
		 x_est(x_est_),y_est(y_est_),thres(thres_),
		 Data(DT_WallEstimate)
		{Set(org_);Init();};
		
	// update Internal variables
	void	Update(float x, float y);
	void	Update(float x, float y, float d, float a);
	void	Init(void);
	boolean Estimate(float &a, float &b, float &c);
	float	Dist(float x_, float y_);
	unsigned int NoSamples(void){return n;}

	// manipulation (route _OWNS_ its waypoints!!)
	void	Set(WallEstimate *template_);
	virtual boolean	SetData(Data *org_){
		if (!org_) return FALSE;
		if (org_->type!=type) return FALSE;
		Set((WallEstimate*)org_);
		return TRUE;
	};

	virtual Data	*Duplicate(void){return new WallEstimate(x_est,y_est,thres,this);};
	virtual Data	*Print(void){printf("WallEstimate(%d:%f,%f,%f)",n,ea,eb,ec);return this;};

private:
	// empirical statistics
	float		sx;			// sum of x's
	float		sy;			// sum of y's
	float		sxy;		// sum of x*y
	float		sx2;		// sum of x squared
	float		sy2;		// sum of y squared
	unsigned int n;			// number of data points
	
	// estimated wall parameters
	float		ea,eb,ec;	// estimated a,b,c
	
	// rejection parameters
	float	x_est,y_est,thres;
};

/////////////////////////////////////////////////////////////
//
//    A C T I O N   for   Navigation
//

class Here : public Action {
public:
	Here(RobotProxy *robot_, Steppable *wp_=NULL);
	~Here();
	virtual	Data	*Step(void); // gimme next waypoint
private:
	boolean Update(void);		// update location
	RobotProxy  *robot;			// robot pointer 
	Waypoint	*this_location;	// waypoint with current location
	Steppable	*wp;			// waypoint to set
};

class Goto : public Action {
public:
	Goto(RobotProxy *robot_, Steppable *waypoint_)
		:myrobot(robot_),waypoint(waypoint_)
		{assert(myrobot);done.Set(FUZZFALSE);waypoint=Getport(waypoint_);};
	virtual	Data	*Step(void); // gimme next waypoint
	virtual	void	Reset(void);
	virtual	void	Abort(void);
protected:	
	void  FloorVector(RobotProxy *robot, Waypoint *goal_waypoint,
			float &dist, float &heading, float &dattitude, boolean &dostop);
	boolean	Arrived(float dist,float dattitude,boolean dostop,
			float &dthres,float &athres);
	RobotProxy *myrobot;
	Steppable *waypoint;
};

class Turnto : public Goto {
public:
	Turnto(RobotProxy *robot_, Steppable *waypoint_)
		:Goto(robot_,waypoint_){};
	virtual	Data	*Step(void); 
};

class Moveby : public Action {
public:
	Moveby(RobotProxy *robot_, Steppable *dist_step_);
	Moveby(RobotProxy *robot_, float dist_);
 	virtual	Data	*Step(void); 
	virtual	void	Reset(void);
	virtual	void	Abort(void);
protected:
	RobotProxy *robot;
	Steppable  *dist_step;
	float	dist;		// move by this distance
	float	started_x;	// started at this 	x,y,a
	float	started_y;
	float	started_a;
};

class Turnby : public Moveby {
public:
	Turnby(RobotProxy *robot_, Steppable *da_step_);
	Turnby(RobotProxy *robot_, float da_);
	virtual	Data	*Step(void); 
};

class TurntoAttitude : public Turnby {
public:
	TurntoAttitude(RobotProxy *robot_, Steppable *da_step_):Turnby(robot_,da_step_){};
	TurntoAttitude(RobotProxy *robot_, float da_):Turnby(robot_,da_){};
	virtual	Data	*Step(void); 
};

class IterateRoute : public Goto {
public:
	IterateRoute(RobotProxy *robot_,Steppable *route_)
		:Goto(robot_,NULL),route(NULL) 
		{done.Set(FUZZFALSE);route=Getport(route_);};
	virtual	Data	*Step(void); // gimme next waypoint
	virtual	void	Reset(void); // restart route (goto start)
	virtual	void	Abort(void); // abort route (goto end)
private:
	Route *GetRoute(void);
	Steppable *route;	
};

class PlanARoute  : public Action {
public:
	PlanARoute(Steppable *hiwaymap_, Steppable *wp_start_, Steppable *wp_goal_)
		:hiwaymap(NULL),wp_start(NULL),wp_goal(NULL),the_route(NULL)
		{done.Set(FUZZTRUE);the_route = new Route;
		 hiwaymap=Getport(hiwaymap_);wp_start=Getport(wp_start_);wp_goal=Getport(wp_goal_);};
	virtual ~PlanARoute(){delete the_route;}
	virtual	Data	*Step(void); // generate a route from "from" to "to"
protected:
	Steppable *hiwaymap;
	Steppable *wp_start;
	Steppable *wp_goal;
	Route	*the_route;	// the generated route
};

class PlanARandomRoute  : public PlanARoute {
public:
	PlanARandomRoute(Steppable *hiwaymap_, Steppable *wp_start_)
		:PlanARoute(hiwaymap_,wp_start_,NULL){};
	virtual	Data	*Step(void); // generate a route from "from" to "to"
};

#define LEFT_WALL  0
#define RIGHT_WALL 1
#define NO_WALL    10

class FollowWall : public Action {
public:
	FollowWall(RobotProxy *robot_,float mspeed_,float dist_, byte prefside_=NO_WALL);
	virtual	Data	*Step(void); // gimme next waypoint
	virtual	void	Reset(void); // restart process
	virtual	void	Abort(void); // end process
private:
	byte 		wside;			// side of the wall (0=left, 1=right)
	byte		prefside;		// preferred side
	float		transspeed;		// maximal speed translating
	float		prefdist;		// preferred distance at which to cruise
	RobotProxy	*robot;
	Steppable 	*wall;	
	float 		olddist;
};


#define ED_BUFSIZE  10
#define ED_HBUFSIZE ED_BUFSIZE/2

class EdgeDetected : public Action {
public:
	EdgeDetected(RobotProxy *robot_);
	virtual	Data	*Step(void); // update the estimate
	virtual	void	Reset(void); // restart estimation process
	virtual	void	Abort(void); // end estimation process and fix estimate
private:
	RobotProxy	*robot;				// the robot
	byte		sonarno;			// the sonar to be used
	int			state;				// 0=locking on, 1=first est. 2=operate
	int			sample_count;		// counter for sonar samples
	WallEstimate curr_est;			// estimate currently being updated
	WallEstimate prev_est;			// previous estimate
	Fuzz		detected;			// was detected?
};

class ResetWallEstimate : public Action {
public:
	ResetWallEstimate(Steppable *est_):est(est_){done.SetTrue();};
	virtual	Data	*Step(void);
	virtual	void	Reset(void){}; // restart estimation process
	virtual	void	Abort(void){}; // end estimation process and fix estimate
private:
	Steppable 	*est;		// estimated wall position
};

class EstimateWall : public Action {
public:
	EstimateWall(RobotProxy *robot_, byte sonarno_, Steppable *est_);
	virtual	Data	*Step(void); // update the estimate
	virtual	void	Reset(void); // restart estimation process
	virtual	void	Abort(void); // end estimation process and fix estimate
private:
	byte		sonarno;	// the sonar to use
protected:
	RobotProxy	*robot;		// the robot
	boolean		inp_ok;		// are inputs ok?
	Steppable 	*est;		// estimated wall position
};

class EstimateBumpWall : public EstimateWall {
public:
	EstimateBumpWall(RobotProxy *robot_, Steppable *est_);
	virtual	Data	*Step(void); // update the estimate
};

class UpdateDRPWall : public Action {
public:
	UpdateDRPWall(RobotProxy *robot_, Steppable *est_, Steppable *wall_)
		:robot(robot_),est(NULL),wall(NULL)
		{done.Set(FUZZTRUE);est=Getport(est_);wall=Getport(wall_);};
	virtual	Data	*Step(void); 	// update my dead reckoning position
private:
	RobotProxy	*robot;		// the robot
	Steppable	*est;		// estimated wall position
	Steppable	*wall;		// world-model wall to compare to
};

//---- avoid stuff using sonar ---------------------------------------------

class SonarAvoid : public Action {
public:
	SonarAvoid(RobotProxy *robot_,float trans_=0.1);
	virtual Data	*Step(void);
private:
	float		ParamA(float max_,float min_=0.4);
	float		ParamB(float max_,float min_=0.4);
	RobotProxy *robot;		// pointer to robot proxy
	float		def_trans;	// default translational velocity
	float		a_side,b_side,a_front,b_front;	// activation parameters
	float		a_mid,b_mid;					// activation parameters
	float		m_side,m_front;
};

//---- avoid forbidden areas ---------------------------------------------

class AvoidForbidden : public Action {
public:
	AvoidForbidden(RobotProxy *robot_,Steppable *hiwaymap_,float dist_=0.6)
		:robot(robot_),dist(dist_),hiwaymap(NULL)
		{assert(robot_);done.Set(FUZZFALSE);hiwaymap=Getport(hiwaymap_);};
	virtual Data	*Step(void);
private:
	RobotProxy *robot;		// the controlled robot
	float		dist;		// distance to antler point
	Steppable  *hiwaymap;	// the hiwaymap used to navigate
};

//---- do I have to recalibrate? ---------------------------------------------

class RecalibQ : public Action {
public:
	RecalibQ(RobotProxy *robot_, Steppable *calib_loco_, float dist_thres_);
	virtual Data	*Step(void);	// return propensity to recalibrate
	virtual	void	Reset(void); 	
	virtual	void	Abort(void); 	
	virtual Fuzz	*IsDone(void);	
private:
	RobotProxy 	*robot;		// the controlled robot
	Steppable	*calib_loco;// location of calibration start
	float		dist_thres; // distance threshold
	Fuzz		propensity;	// how much do I want to recalibrate
};

//---- do I have to recalibrate? ---------------------------------------------

class RefuelQ : public Action {
public:
	RefuelQ(RobotProxy *robot_)
		:robot(robot_){assert(robot);done.SetTrue();propensity.SetFalse();};
	virtual Data	*Step(void);
private:
	RobotProxy 	*robot;		// the controlled robot
	Fuzz		propensity;	// how much do I want to recalibrate
};



#endif	/* _Navsup_h_ */
