/*

DISCLAIMER 

InSitu is provided as is, without any kind of guarantees or warranty. There is 
no documentation. The author does not provide any kind of support to users of 
this source code and is not responsible for any kind of problems or damages 
resulting from its use. 

The source code is the result of a long iterative design process for a research 
prototype and was never intended to be used in any kind of product. In general, 
the code cannot be expected to compile in any straightforward manner on your 
computer system, since the robot on which it was developed, has a custom design. 

*//* Rene Schaad                                                    */
/* Departement of Computer Science, Univ. of Zurich, Switzerland  */
/* schaad@ifi.unizh.ch                                            */


#include "OperatorProxy.h"
#include <stdlib.h>


//---- OperatorProxy ---------------------------------------------------------

#define TMP_BUF_SIZE 2000
#define IP_CMD_BUFLEN 5000

static char OperatorProxy::tmpbuf[TMP_BUF_SIZE];

void OperatorProxy::ReportError(char *file_name, int line_no, iserr_t e, char *txt)
{
	char *etxt;	// error text
	switch (e) {
		case iserr_unknown:		etxt = "Unknown error"; break;
		case iserr_fileopen:	etxt = "Opening file"; break;
		case iserr_parse:		etxt = "Parse error"; break;
		case iserr_bufsize:		etxt = "Buffer-size too small"; break;
		case iserr_stringsize:	etxt = "String too long for buffer"; break;
		case iserr_toktocons:	etxt = "Converting token to ConsCell"; break;
		case iserr_scanner:		etxt = "In scanner"; break;
		case iserr_syntax:		etxt = "Syntax error"; break;
		case iserr_noeval:		etxt = "Cannot evaluate this object"; break;
		case iserr_notyet:		etxt = "Not yet implemented"; break;
		case iserr_undeffct:	etxt = "Attempt to call undefined function"; break;
		case iserr_nonfct:		etxt = "Attempt to call non-function"; break;
		case iserr_lambda:		etxt = "Lambda evaluation error"; break;
		case iserr_argformat:	etxt = "Argument format wront"; break;
		case iserr_argnull:		etxt = "NULL argument"; break;
		case iserr_arg2many:	etxt = "Too many arguments"; break;
		case iserr_arg2few:		etxt = "Too few arguments"; break;
		case iserr_argtype:		etxt = "Wrong argument type"; break;
		case iserr_argnoex:		etxt = "Argument does not exist"; break;
		case iserr_div0:		etxt = "Divide by zero"; break;
		case iserr_dupdock:		etxt = "Duplicate dock"; break;
		case iserr_nodock:		etxt = "Dock does not exist"; break;
		default:				etxt = "Strange error"; break;
	};
	if (txt)
		sprintf(tmpbuf,"ERROR: %s, context: %s (Source file: %s, line: %d)\n",etxt,txt,file_name,line_no);
	else	
		sprintf(tmpbuf,"ERROR: %s (Source file: %s, line: %d)\n",etxt,file_name,line_no);
	PutString(tmpbuf);
}

void OperatorProxy::ReportError(char *file_name, int line_no, iserr_t e, int no)
{
	char notxt[15];
	sprintf(notxt,"%10d",no);
	ReportError(file_name,line_no,e,notxt);
}

OperatorProxy &OperatorProxy::operator << (double d)
{
	sprintf(tmpbuf,"%lf",d);
	PutString(tmpbuf);
	return *this;
}

OperatorProxy &OperatorProxy::operator << (int x)
{
	sprintf(tmpbuf,"%5d",x);
	PutString(tmpbuf);
	return *this;
}

OperatorProxy &OperatorProxy::operator << (char *t)
{
	PutString(t);
	return *this;
}


//---- OPfile ---------------------------------------------------------

OPfile::OPfile(char *file_name_) : infile_name(file_name_)
{
	outfile=NULL;
	sprintf(tmpbuf,"%s.rep",infile_name);
	if ((outfile = fopen(tmpbuf,"w"))==NULL) {
		fprintf(stderr,"WARNING: could not open file for writing: %s (in file \"%s\", line %d)\n",
		   tmpbuf,__FILE__,__LINE__);
	}
	infile_loaded = FALSE;
}

OPfile::~OPfile()
{
	if (outfile!=NULL) fclose(outfile);
}

char *OPfile::GetCommand(void)
{
	if (infile_loaded)
		return NULL;
	else {
		infile_loaded=TRUE;
		return LoadFile(infile_name);
	}
}

char *OPfile::LoadFile(char *filename)
{
	// here, determine file length, or never get more than buffer can
	// hold !!
	int c;
	FILE *loadfile;
	
	if ((loadfile = fopen(filename,"r"))==NULL) {
		ISERR3(iserr_fileopen,filename);
		return NULL;
	}
	
	char *retbuf = new char [IP_CMD_BUFLEN]; // create new buffer
	int cnt=0;
	
	char *bufptr=retbuf;					// read in the file
	while ((c=getc(loadfile))!=EOF) {
		if (++cnt>=IP_CMD_BUFLEN) {
			ISERR2(iserr_bufsize);
			*bufptr=0;
			fclose(loadfile);
			return retbuf;
		}
		*(bufptr++)=(char)c;
	}
	*bufptr=(char)0;
	fclose(loadfile);
	return retbuf;							// return the buffer
}

void OPfile::PutString(char *s)
{
	if (outfile==NULL) {
		fprintf(stderr,"%s",s);
	} else {
		fprintf(outfile,"%s",s);
	}
}


//---- OPprompt ---------------------------------------------------------

OPprompt::OPprompt():promptn(0)
{
	PutString("\n\n\nInSitu Skript Version 0.1, 30 July 1997\n");
	PutString("        (c) by Rene Schaad\n");
	PutString("---------------------------------------\n\n");
}

char *OPprompt::GetCommand(void)
{
	char *retbuf = new char [IP_CMD_BUFLEN]; // create new buffer
	char *bufptr = retbuf;
	int c, nlcount=0, bracketcount=0;
	
	printf("\n<%d> ",++promptn);

	while ((nlcount<20) && ((c=getchar()) != EOF )) {
		switch (c) {
			case  1: // start-of-line
					bufptr=retbuf;
					nlcount=0;
					printf("\n<%d> ",promptn);
					break;
			case  8: // backspace
					if (bufptr>retbuf) bufptr--;
					nlcount=0;
					break;
			case 10: 
			case 13: // create a new line
					*(bufptr++)=' ';    
					nlcount++;
					if (bracketcount==0) nlcount=1000;
					break;
			case '(': // handle opening brackets
					bracketcount++;
					*(bufptr++)=(char)c;    
					nlcount=0;
					break;
			case ')': // handle closing brackets
					if (bracketcount>0) bracketcount--;
					*(bufptr++)=(char)c;    
					nlcount=0;
					break;
			default: // add all other/printable characters
					*(bufptr++)=(char)c;    
					nlcount=0;
					break;
		};
	}
	*bufptr=(char)0;
	return retbuf;							// return the buffer
}

char *OPprompt::LoadFile(char *filename)
{
	// here, determine file length, or never get more than buffer can
	// hold !!
	int c;
	FILE *loadfile;
	
	if ((loadfile = fopen(filename,"r"))==NULL) {
		ISERR3(iserr_fileopen,filename);
		return NULL;
	}
	
	char *retbuf = new char [IP_CMD_BUFLEN]; // create new buffer
	int cnt=0;
	
	char *bufptr=retbuf;					// read in the file
	while ((c=getc(loadfile))!=EOF) {
		if (++cnt>=IP_CMD_BUFLEN) {
			ISERR2(iserr_bufsize);
			*bufptr=0;
			fclose(loadfile);
			return retbuf;
		}
		*(bufptr++)=(char)c;
	}
	*bufptr=(char)0;
	fclose(loadfile);
	return retbuf;							// return the buffer
}

void OPprompt::PutString(char *s)
{
	printf("%s",s);
}


//---- OPremote ---------------------------------------------------------

OPremote::OPremote()
{
	PutString("\n\n\nInSitu Skript Remote Version 0.1, 30 July 1997\n");
	PutString("        (c) by Rene Schaad\n");
	PutString("----------------------------------------------\n\n");
}

char *OPremote::GetCommand(void)
{
	word dummy,no_bytes,sendlen;
	
	///////////////////////////////////////////////////////
	// process IPC messages (e.g. instructions from operator)
	if ((senderpid = CReceiveLen(0,recbuf,MAX_PACKET_LEN,dummy))>-1) {
		
		if (*recbuf == cmd_noop) {
			sendlen=instr_rep_pack.Marshal(sendbuf);// prepare reply
			ReplyLen(senderpid,sendbuf,sendlen);	// send reply
			instr_rep_pack.Clear();					// clear reply buffer
		} else if (*recbuf == cmd_instr) {
			sendlen=instr_rep_pack.Marshal(sendbuf);// prepare reply
			ReplyLen(senderpid,sendbuf,sendlen);	// send reply
			instr_rep_pack.Clear();					// clear reply buffer

			no_bytes=instr_pack.Unmarshal(recbuf);	// unpack packet
			char *retbuf = new char [no_bytes+3];	// make new buffer for parser
			strncpy(retbuf,instr_pack.text,no_bytes);// copy to new buffer
			retbuf[no_bytes]=0;						// terminate buffer
			return retbuf;							// return new buffer (deallocated by parser)
		} else {
			cout << "Unknown packet type\n";
			return NULL;
		}
		
		
	}
	return NULL;
}


// for now, only Robot-local files can be loaded !!

char *OPremote::LoadFile(char *filename)
{
	// here, determine file length, or never get more than buffer can
	// hold !!
	int c;
	FILE *loadfile;
	
	if ((loadfile = fopen(filename,"r"))==NULL) {
		ISERR3(iserr_fileopen,filename);
		return NULL;
	}
	
	char *retbuf = new char [IP_CMD_BUFLEN]; // create new buffer
	int cnt=0;
	
	char *bufptr=retbuf;					// read in the file
	while ((c=getc(loadfile))!=EOF) {
		if (++cnt>=IP_CMD_BUFLEN) {
			ISERR2(iserr_bufsize);
			*bufptr=0;
			fclose(loadfile);
			return retbuf;
		}
		*(bufptr++)=(char)c;
	}
	*bufptr=(char)0;
	fclose(loadfile);
	return retbuf;							// return the buffer
}

void OPremote::PutString(char *s)
{
	instr_rep_pack.Append(s);
}

////////////////////////////////////////////////////////////////////////
/////////
/////////   Report to operator
/////////

Report::Report(OperatorProxy *op_, Steppable *datasrc_)
	:op(op_),datasrc(NULL)
{
	datasrc=Getport(datasrc_);
}

Data *Report::Step(void)
{
	if (!datasrc) return NULL;		// does data source exist
	Data *retval = datasrc->Step();	// get data
	if (!retval) return NULL;		// does retval exist?
	if (!op) return retval;			// does operator proxy exist?
	
	// send data to proxy
	DataType t = retval->type;
	
	if 		(t==DT_Text)	*op << ((Text*)retval)->Get();
	else if	(t==DT_Number)	*op << ((Number*)retval)->Get();
	else 					*op << "Report<unknown DataType>";
	
	return retval;
}

