/*

DISCLAIMER 

InSitu is provided as is, without any kind of guarantees or warranty. There is 
no documentation. The author does not provide any kind of support to users of 
this source code and is not responsible for any kind of problems or damages 
resulting from its use. 

The source code is the result of a long iterative design process for a research 
prototype and was never intended to be used in any kind of product. In general, 
the code cannot be expected to compile in any straightforward manner on your 
computer system, since the robot on which it was developed, has a custom design. 

*/

#ifndef _OperatorProxy_h_
#define _OperatorProxy_h_

/* Rene Schaad                                                    */
/* Departement of Computer Science, Univ. of Zurich, Switzerland  */
/* schaad@ifi.unizh.ch                                            */

#include <stdio.h>
#include "defines.h"
#include "PrgStruct.h"
#include "Packet.h"

#define ISERR1			  ReportError(__FILE__,__LINE__,iserr_inknown,(char*)NULL)
#define ISERR2(type)	  ReportError(__FILE__,__LINE__,type,(char*)NULL)
#define ISERR3(type, arg) ReportError(__FILE__,__LINE__,type,arg)

enum iserr_t {
	iserr_unknown,		// 0
	iserr_fileopen, 	// 1  error while opening file
	iserr_parse, 		// 2  parse error
	iserr_bufsize,		// 3  buffer size exceeded
	iserr_stringsize,	// 4  string too long for buffer
	iserr_toktocons,	// 5  token to cons cell conversion
	iserr_scanner,      // 6  error in scanner
	iserr_syntax,       // 7  syntax error reported by parser
	iserr_noeval,       // 8  cannot evaluate this object
	iserr_notyet,       // 9  this feature not yet implemented
	iserr_undeffct,     // 10 undefined function called
	iserr_nonfct,       // 11 non-function called
	iserr_lambda,       // 12 error evaluating lambda expression
	iserr_argformat,    // 13 arument list format
	iserr_argnull,      // 14 NULL argument list
	iserr_arg2many,     // 15 too many arguments
	iserr_arg2few,      // 16 not enough arguments
	iserr_argtype,      // 17 wrong argument type
	iserr_argnoex,      // 18 argument does not exist
	iserr_div0,         // 19 divide by zero
	iserr_dupdock,      // 20 duplicate dock encountered
	iserr_nodock,	    // 21 no such dock exists
	iserr_dummy		
	};

//---- OperatorProxy -----------------------------------------------------------

class OperatorProxy : public Action {
public:
	virtual Data	*Step(void) {return NULL;};

	virtual char	*GetCommand(void) =0;			// get a new buffer full of commands
	virtual char	*LoadFile(char *filename) =0;	// load a named file into a new buffer
	virtual void	PutString(char *s) =0;			// move this string out
	virtual void	Sync(void) =0;					// synchronize with operator

	void		  ReportError(char *file_name, int line_no, iserr_t e, char *txt);
	void		  ReportError(char *file_name, int line_no, iserr_t e, int no);
	OperatorProxy &operator << (double d);	// dump data to the operator
	OperatorProxy &operator << (int x);		// dump data to the operator
	OperatorProxy &operator << (char *t);	// dump data to the operator
protected:
	static char tmpbuf[];
};

class Report : public Action {
public:
	Report(OperatorProxy *op_, Steppable *datasrc_);
	virtual Data	*Step(void);
	virtual Fuzz	*IsDone(void)
		{if (datasrc) return datasrc->IsDone();else return &fuzz_true;};
private:
	OperatorProxy 	*op;		// pointer to operator proxy
	Steppable		*datasrc;	// pointer to data source
};

class OPfile : public OperatorProxy {
public:
	OPfile(char *file_name_);
	~OPfile();
	virtual char	*GetCommand(void) ;				// get a new buffer full of commands
	virtual char	*LoadFile(char *filename);		// load a named file into a new buffer
	virtual void	PutString(char *s);				// move this string out
	virtual void	Sync(void){}; 					// synchronize with operator
private:	
	FILE *outfile;			// file for output of replies, etc.
	char *infile_name;		// the name of the file
	boolean infile_loaded;	// was infile already loaded?
};

class OPprompt : public OperatorProxy {
public:
	OPprompt();
	virtual char	*GetCommand(void) ;				// get a new buffer full of commands
	virtual char	*LoadFile(char *filename);		// load a named file into a new buffer
	virtual void	PutString(char *s);				// move this string out
	virtual void	Sync(void){}; 					// synchronize with operator
private:
	int promptn;		// input prompt number	
};

class OPremote : public OperatorProxy {
public:
	OPremote();
	virtual char	*GetCommand(void) ;				// get a new buffer full of commands
	virtual char	*LoadFile(char *filename);		// load a named file into a new buffer
	virtual void	PutString(char *s);				// move this string out
	virtual void	Sync(void){}; 					// synchronize with operator
private:
	byte  		sendbuf[MAX_PACKET_LEN],// buffers for send/receive data
				recbuf[MAX_PACKET_LEN];	
	pid_t 		senderpid;				// process id of sender of a IPC message
	AckPacket   ack_pack;
	InstructionReplyPacket	instr_rep_pack; //  replies generated by instrucions
	InstructionPacket 		instr_pack;		//  instrucionts to robot
	NoopPacket		 		noop_pack;		//  just return replies
};

#endif	/* _OperatorProxy_h_ */
