/*

DISCLAIMER 

InSitu is provided as is, without any kind of guarantees or warranty. There is 
no documentation. The author does not provide any kind of support to users of 
this source code and is not responsible for any kind of problems or damages 
resulting from its use. 

The source code is the result of a long iterative design process for a research 
prototype and was never intended to be used in any kind of product. In general, 
the code cannot be expected to compile in any straightforward manner on your 
computer system, since the robot on which it was developed, has a custom design. 

*/

#ifndef _Packet_h_
#define _Packet_h_

/* Rene Schaad                                                    */
/* Departement of Computer Science, Univ. of Zurich, Switzerland  */
/* schaad@ifi.unizh.ch                                            */

#include "RufusPhys.h"
#include "defines.h"
#include <sys/types.h>

// IPC names of all deamons
#define MODEMD_IPCN			"plascom/modemd"
#define ROBOTD_IPCN 		"plascom/robotd"
#define BEHAVIORD_IPCN		"plascom/behaviord"
#define INTERPRETER_IPCN	"plascom/interpreter"


#define MAX_PACKET_LEN 5000 
#define MAX_PROTOCOL_LEN (MAX_PACKET_LEN-100)
#define SPK_TXTLEN     700

#define LOBYTE(A)	((byte)((A) & 0xFF))
#define HIBYTE(A)	((byte)(((A) & 0xFF00) >> 8))
#define ADDIBUF(A)  {*buf++=LOBYTE(A);*buf++=HIBYTE(A);len+=2;}
#define ADDWBUF(A)  {*buf++=LOBYTE(A);*buf++=HIBYTE(A);len+=2;}
#define ADDBBUF(A)  {*buf++=(A);len++;}
#define MAKEINT(A)  ((signed short int)(*A)+(signed short int)(((signed short int)(*(A+1)))<<8))
#define MAKEWORD(A) ((word)((*A)+((*(A+1))<<8)))
#define GETIBUF(A,B) {A=MAKEINT(B); B++;B++;len+=2;}
#define GETWBUF(A,B) {A=MAKEWORD(B); B++;B++;len+=2;}
#define GETBBUF(A,B) {(A)=*(B);(B)++;len++;};

enum proc_ids {	robotd_proc,
				behavior_proc,
				interpreter_proc,
				modemd_proc,
				any_proc,
				all_proc,
				error_proc
				};
enum packet_ids  {
				cmd_noop,			// 0 does nothing, carries reply code
				cmd_list_term,		// 1 terminates list in composite
				cmd_composite,		// 2 collection of other cmd packets
				cmd_set_pid,		// 3 set PID parameters of motor control
				cmd_set_sound,		// 4 switch sound on/off
				cmd_set_speech,		// 5 say this text in synthesizer
				cmd_set_velocity,	// 6 go this fast
				cmd_set_drpos,		// 7 reset dead reckoning to this position
				cmd_set_campos,		// 8 set camera to this position
				cmd_set_dockpos,	// 9 set docking mechanism to this position (in %)
				cmd_set_mode,		// 10 switch robot to this mode (no toggle)
				cmd_error,			// 11 indicates error
				cmd_start_stop,		// 12 start or stop something (item)
				
				cmd_instr,			// instruction packet

				data_ack,
				data_status,
				data_protocol,
				
				dummy_packet_id
			  };
enum reply_ids {
				rly_send_nothing,
				rly_send_ack,
				rly_send_status
				};

/////////////////////////////////////////////////////////////////////
// Wrapper IPC functions that add/extract packet length information

#ifdef __QNX__

int SendLen(pid_t pid, void *smsg, void *rmsg, 
			word snbytes, word rnbytes, word &arnbytes);
int ReceiveLen(pid_t pid, void *msg,word nbytes, word &anbytes);
int CReceiveLen(pid_t pid, void *msg,word nbytes, word &anbytes);
int ReplyLen(pid_t pid, void *msg, word nbytes);

#endif

/////////////////////////////////////////////////////////////////////
// Packets from robot to base station

class AckPacket {
public:
	word Marshal(byte *buf);
	word Unmarshal(byte *buf);
}; 

class InstructionReplyPacket  {
public:
	InstructionReplyPacket();
	~InstructionReplyPacket();
	word	Marshal(byte *buf);
	word	Unmarshal(byte *buf);
	
	void	Clear(void);
	void	Append(char *newtxt);
	
	char *text;
	char *next;
	word nchars;
};

class StatPacket  {
public:
	word	Marshal(byte *buf);
	word	Unmarshal(byte *buf);
	
	float	sonar[NO_SONARS];		// sonar readings
	float	pos_x,pos_y,attitude;	// dead reckoning results
	float	tspeed, aspeed;			// translational & rotational speeds
	float	compass;				// compass heading (body coordinates)
	boolean radar;					// motion detected by radar
	boolean pir;					// motion detected by passive IR sensor
	word	light[NO_LIGHTS];		// ambient light sensors
	word	light50;				// modulated light sensor
	word	bump_stat;				// bumper status (16bits)
	float	camera_pan,	camera_tilt;// pan and tilt angle of camera
	sound_t	soundno;				// sound that is currently being played
	byte	mode;					// mode of robotd
	float	dock_pos;				// position of docking device (0=retracted, 1=extended)
	boolean	behavior_running;		// is behavior system running
};

/////////////////////////////////////////////////////////////////////
// Packets from base station to robot


class CmdPacket {
friend class CompositePacket;
public:	
	CmdPacket();
	virtual	word	Marshal(byte *buf);
	virtual	word	Unmarshal(byte *buf);
	
	byte	reply_code;	// what to reply with
	byte	cmd;		// command to execute
	
protected:
	CmdPacket * next;	// used for linked list, not marshaled
};

word UnmarshalAny(byte *buf, CmdPacket *&newp);

class InstructionPacket : public CmdPacket {
public:
	InstructionPacket();
	virtual	word	Marshal(byte *buf);
	virtual	word	Unmarshal(byte *buf);
	
	word 	Store(const char *txt);
	void	Clear(void);
	
	word	textlen;
	char	text[SPK_TXTLEN];
};

class SpeechPacket : public CmdPacket {
public:
	SpeechPacket();
	virtual	word	Marshal(byte *buf);
	virtual	word	Unmarshal(byte *buf);
	
	word 	Store(const char *txt);
	void	Clear(void);
	
	word	textlen;
	char	text[SPK_TXTLEN];
};

class PidPacket : public CmdPacket  {
public:
	PidPacket();
	virtual	word Marshal(byte *buf);
	virtual	word Unmarshal(byte *buf);
	
	float pout;
	byte p,i,d,cnt;
};

class SoundPacket : public CmdPacket {
public:
	SoundPacket();
	virtual	word Marshal(byte *buf);
	virtual	word Unmarshal(byte *buf);
	
	sound_t sound_no;
};	

class SpeedPacket : public CmdPacket {
public:
	SpeedPacket();
	virtual	word Marshal(byte *buf);
	virtual	word Unmarshal(byte *buf);
	
	float t,a;
	boolean joymode;
};	

class DRPosPacket : public CmdPacket {
public:
	DRPosPacket();
	virtual	word Marshal(byte *buf);
	virtual	word Unmarshal(byte *buf);
	
	float x,y,a;
};	

class ModePacket : public CmdPacket {
public:
	ModePacket();
	virtual	word Marshal(byte *buf);
	virtual	word Unmarshal(byte *buf);
	
	byte	mode;
};	

class CamPacket : public CmdPacket {
public:
	CamPacket();
	virtual	word Marshal(byte *buf);
	virtual	word Unmarshal(byte *buf);
	
	float pan,tilt;
};	

class DockPacket : public CmdPacket {
public:
	DockPacket();
	virtual	word Marshal(byte *buf);
	virtual	word Unmarshal(byte *buf);
	
	float amount;
};	

class StartStopPacket : public CmdPacket {
public:
	StartStopPacket();
	virtual	word Marshal(byte *buf);
	virtual	word Unmarshal(byte *buf);
	
	byte item,start;
};	

class NoopPacket : public CmdPacket {
public:
	NoopPacket();
	virtual	word Marshal(byte *buf);
	virtual	word Unmarshal(byte *buf);
};	

class CompositePacket : public CmdPacket {
public:
	CompositePacket();
	
	virtual	word Marshal(byte *buf);
	virtual	word Unmarshal(byte *buf);
	
	void		Clear(void);	// erase list but do not delete objects
	void		Delete(void);	// erase list but and delete objects
	boolean		AddUnique(CmdPacket *new_pack);	// replaces cmd of same kind, or adds
	CmdPacket	*First(void);
	CmdPacket	*Next(void);
	word		NoItems(void);

private:
	word		no_items;
	CmdPacket	* first;  // first element of linked list
	CmdPacket	* last;	  // last list element
	CmdPacket	* current;// current list element
};	



#endif	/* _Packet_h_ */
