/*

DISCLAIMER 

InSitu is provided as is, without any kind of guarantees or warranty. There is 
no documentation. The author does not provide any kind of support to users of 
this source code and is not responsible for any kind of problems or damages 
resulting from its use. 

The source code is the result of a long iterative design process for a research 
prototype and was never intended to be used in any kind of product. In general, 
the code cannot be expected to compile in any straightforward manner on your 
computer system, since the robot on which it was developed, has a custom design. 

*//* Rene Schaad                                                    */
/* Departement of Computer Science, Univ. of Zurich, Switzerland  */
/* schaad@ifi.unizh.ch                                            */


#include "PrgStruct.h"
#include "SupMgr.h"
#include <assert.h>


#define MAX_NAME_LEN 200

////////////////////////////////////////////////////////////////////////////
//
//           T E S T E R    S T U F F 
//

Fuzz fuzz_true(FUZZTRUE);
Fuzz fuzz_false(FUZZFALSE);
Fuzz fuzz_unknown(FUZZUNKNOWN);
Fuzz fuzz_evidence(FUZZEVIDENCE);
Fuzz fuzz_in_progress(FUZZEVIDENCE);
Fuzz chain_exit(FUZZFALSE);

////////////////////////////////////////////////////////////////////////////
//
//           D A T A
//


Fuzz	*Data::IsDone(void) {return &(fuzz_true);}

////////////////////////////////////////////////////////////////////////////
//
//           S T E P P A B L E
//

Steppable::~Steppable()
{
	delete chain_dock;
}

void	Steppable::SetChain(Steppable *chain)
{
	if (NULL==chain_dock) chain_dock = new Dock;
	chain_dock->SetExit(chain);
}

Steppable	*Steppable::GetChain(void)
{
	if (chain_dock) return chain_dock->GetExit();
	else return NULL;
}

Steppable *Steppable::Getport(Steppable *p)
{
	// Getport checks whether the pointer p is to indicate that a 
	// chained argument must be used. 
	// If p is CHAIN, this indicates that the SUP must be chained and
	// that the chain_dock must be used as an argument at this point.
	
	if (p==CHAIN) {				// if the argument is CHAIN token
		if (NULL==chain_dock) chain_dock = new Dock; // create chain_dock
		return chain_dock;		// return the chain_dock as the argument
	} else {
		return p;				// if not token: return the argument itself
	}
}

boolean	Steppable::Test(Steppable *tester, boolean &determining, double thres)
{
	// Test the tester
	// returns FALSE when: NULL tester, NULL result, not Fuzz result, 
	//						< thres, determining
	
	determining = FALSE;
	fuzz_res	= NULL;
	Data *res=NULL;							// set the result to NULL
	if (tester!=NULL) res=tester->Step();   // do the test, tester exists
	
	if (res!=NULL)	{		
		if (res->type==DT_Fuzz) {
			fuzz_res = (Fuzz*)res;	
			if (fuzz_res->Gt(thres)) {		
				return TRUE;
			} else if (fuzz_res->Determining()) { // is still trying to find answer
				determining = TRUE;
			}
		}
	}
	
	return FALSE;
}


////////////////////////////////////////////////////////////////////////////
//
//           T R E E  B A S E
//

TreeBase::TreeBase(
		Steppable *before_,	// step these when committing the tree
		Steppable *while_  ,	// step this on each update
		Steppable *after_  ,	// step this upon removal
		Parameter *p1 ,		// these are the parameters
		Parameter *p2 ,
		Parameter *p3 ,
		Parameter *p4 ,
		Parameter *p5 ,
		Parameter *p6 ,
		Parameter *p7 ,
		Parameter *p8 ,
		Parameter *p9 )
		:param_list(NULL)
{
	SetBeforeTree(before_);
	SetWhileTree(while_);
	SetAfterTree(after_);
	AppendParameters(p1,p2,p3,p4,p5,p6,p7,p8,p9);
}

void TreeBase::AppendParameters(
			Parameter *p1,
			Parameter *p2,
			Parameter *p3,
			Parameter *p4,
			Parameter *p5,
			Parameter *p6,
			Parameter *p7,
			Parameter *p8,
			Parameter *p9)
{
	AppendParameter(p1);
	AppendParameter(p2);
	AppendParameter(p3);
	AppendParameter(p4);
	AppendParameter(p5);
	AppendParameter(p6);
	AppendParameter(p7);
	AppendParameter(p8);
	AppendParameter(p9);
}


Parameter *TreeBase::AppendParameter(Parameter* p)
{
	if (! param_list) param_list = new List;
	if (param_list->Append(p)) return p;
	else return NULL;
}

Steppable	*TreeBase::GetParameter(unsigned int pos)
{
	if (! param_list) return NULL;
	Parameter *par = (Parameter*)param_list->Get(pos);	// get element # pos
	if (! par) return NULL;
	return par->Get();
}

Steppable	*TreeBase::SetParameter(unsigned int pos, Steppable *p)
{
	if (! param_list) return NULL;
	Parameter *par = (Parameter*)param_list->Get(pos);
	if (! par) return NULL;
	par->Set(p);
	return p;
}

unsigned int TreeBase::GetNoParameters(void)
{
	return ((param_list)?param_list->Size():0);
}

////////////////////////////////////////////////////////////////////////////
//
//           T E X T
//

Text::Text(const char *text_):Data(DT_Text),textlen(0),text(NULL)
{
	Set(text_);
}

void Text::Set(const char *text_)
{
	if (!text_) return;
	
	word newtextlen=0;
	if (text_) newtextlen=(word)strlen(text_);
	else newtextlen=0;

	if (newtextlen>textlen) {
		delete [] text;
		if (newtextlen>MAX_TEXT_LEN) newtextlen=MAX_TEXT_LEN-1;
		text = new char [newtextlen+4];
		textlen=newtextlen;
	} 
	
	if (textlen>0) {
		strncpy(text,text_,textlen);
		text[textlen]=0;
	} else {
		delete [] text;
		text=NULL;
	}
}

void Text::Clear(void)
{
	textlen=0;
	delete [] text;
}

////////////////////////////////////////////////////////////////////////////
//
//           F C O U N T D O W N
//

Ramp::Ramp(long msecs_)
{
	msecs		=	msecs_;
	Reset();
}

void Ramp::Reset(void)
{
	retval.Set(FUZZFALSE);
	done.Set(FUZZFALSE);
	start_time	=	rt.CurrentTime();
}

void Ramp::Abort(void)
{
	retval.Set(FUZZTRUE);
	done.Set(FUZZTRUE);
	start_time	=	rt.CurrentTime() - msecs - 10L;
}

Data	*Ramp::Step(void)
{
	double progress = Limit((float)(rt.CurrentTime() - start_time)/(float)msecs,
							FUZZFALSE,FUZZTRUE);
	done.Set(progress);
	retval.Set(progress);
	return &retval;
}

////////////////////////////////////////////////////////////////////////////
//
//           F P U L S E T R A I N
//

Pulsetrain::Pulsetrain(time_t msecs_hi_,time_t msecs_lo_)
{
	assert(msecs_hi_>=0L);
	assert(msecs_lo_>=0L);
	
	retval.Set(FUZZFALSE);
	start_time	=	rt.CurrentTime();
	msecs_hi		=	msecs_hi_;
	msecs_lo		=	msecs_lo_;
	msecs_to		=	msecs_lo_+msecs_hi_;
}

void Pulsetrain::Reset(void)
{
	retval.Set(FUZZFALSE);
	start_time	=	rt.CurrentTime();
}

void Pulsetrain::Abort(void)
{
	Reset();
}

Data	*Pulsetrain::Step(void)
{
	long diff = rt.CurrentTime() - start_time;
	if ((diff%msecs_to)<=msecs_hi)	retval.Set(FUZZTRUE);
	else							retval.Set(FUZZFALSE);
	return &retval;
}

////////////////////////////////////////////////////////////////////////////
//   LogOp 

LogOp::LogOp(	Steppable *t1,
			Steppable *t2,
			Steppable *t3,
			Steppable *t4,
			Steppable *t5,
			Steppable *t6,
			Steppable *t7,
			Steppable *t8,
			Steppable *t9)
{
	Append(t1);
	Append(t2);
	Append(t3);
	Append(t4);
	Append(t5);
	Append(t6);
	Append(t7);
	Append(t8);
	Append(t9);
}

void LogOp::Append(Steppable *t)
{
	if (t==NULL) return;
	tlist.Append(Getport(t));
}

void LogOp::Abort(void)
{
	Steppable *current=NULL;
	tlist.Reset();
	while (NULL!=(current=(Steppable*)tlist.Iterate())) {
		current->Abort();
	}
}

void LogOp::Reset(void)
{
	Steppable *current=NULL;
	tlist.Reset();
	while (NULL!=(current=(Steppable*)tlist.Iterate())) {
		current->Reset();
	}
}

////////////////////////////////////////////////////////////////////////////
//   And 

And::And(Steppable *t1,Steppable *t2,Steppable *t3,Steppable *t4,Steppable *t5,
		 Steppable *t6,Steppable *t7,Steppable *t8,Steppable *t9) 
		 : LogOp(t1,t2,t3,t4,t5,t6,t7,t8,t9)
{}

Data *And::Step(void)
{
	retval.Set(FUZZTRUE);
	Steppable *current=NULL;
	if (tlist.Size()<=0) return &fuzz_false;
	
	tlist.Reset();
	while (NULL!=(current=(Steppable*)tlist.Iterate())) {
		Data *dr = current->Step();
		Fuzz *r;
		if (dr!=NULL) {
			if (dr->type==DT_Fuzz) {
				r = (Fuzz*)dr;
				if (r->Unknown()) {
					retval.Set(FUZZUNKNOWN);
					return &retval;
				} else if (r->Determining()) {
					retval.Set(FUZZDETERMINING);
					return &retval;
				} else {
					retval.And(r->Get());
				}
			} else
				return &fuzz_false;
		} else {
			return &fuzz_false;
		}
	}
	return &retval;
}

////////////////////////////////////////////////////////////////////////////
//   Or 

Or::Or  (Steppable *t1,Steppable *t2,Steppable *t3,Steppable *t4,Steppable *t5,
		 Steppable *t6,Steppable *t7,Steppable *t8,Steppable *t9) 
		 : LogOp(t1,t2,t3,t4,t5,t6,t7,t8,t9)
{}

Data *Or::Step(void)
{
	retval.Set(FUZZFALSE);
	Steppable *current=NULL;
	if (tlist.Size()<=0) return &fuzz_false;
	
	tlist.Reset();
	while (NULL!=(current=(Steppable*)tlist.Iterate())) {
		Data *dr = current->Step();
		Fuzz *r;
		if (dr!=NULL) {
			if (dr->type==DT_Fuzz) {
				r = (Fuzz*)dr;
				if (r->Unknown()) {
					retval.Set(FUZZUNKNOWN);
					return &retval;
				} else if (r->Determining()) {
					retval.Set(FUZZDETERMINING);
					return &retval;
				} else {
					retval.Or(r->Get());
				}
			} 
		}
	}
	return &retval;
}

////////////////////////////////////////////////////////////////////////////
//   Not 

Not::Not  (Steppable *t) 
		 : LogOp(t)
{}

Data *Not::Step(void)
{
	retval.Set(FUZZTRUE);
	Steppable *current=NULL;
	if (tlist.Size()<=0) return &fuzz_false;
	
	tlist.Reset();
	if (NULL!=(current=(Steppable*)tlist.Iterate())) {
		Data *dr = current->Step();
		Fuzz *r;
		if (dr!=NULL) {
			if (dr->type==DT_Fuzz) {
				r = (Fuzz*)dr;
				if (r->Unknown()) {
					retval.Set(FUZZUNKNOWN);
					return &retval;
				} else if (r->Determining()) {
					retval.Set(FUZZDETERMINING);
					return &retval;
				} else {
					retval.Set(r->Get());
					retval.Not();
				}
			} 
		}
	}
	return &retval;
}

////////////////////////////////////////////////////////////////////////////
//
//           S C H E D U L E R
//

Scheduler::Scheduler(word interval_):interval(interval_)
{ 
	producer_list	=	new List[MAX_PP];
	current_pp=(ProducerPriority)0;   // set to first, whatever its name
	done.Set(FUZZFALSE);
}

Scheduler::~Scheduler()
{
	delete [] producer_list;
}

void	Scheduler::SetPriority(ProducerPriority pp)
{
	if (pp<MAX_PP) current_pp=pp;
}

void	Scheduler::Remove(Steppable *oldprod)
{
	Remove(current_pp,oldprod);
}

void	Scheduler::Remove(ProducerPriority pp, Steppable *oldprod)
{
	if (pp>=MAX_PP) return;
	producer_list[pp].Remove(oldprod);
}

void	Scheduler::Add(Steppable *newprod)
{
	Add(current_pp,newprod);
}

void	Scheduler::Add(ProducerPriority pp, Steppable *newprod)
{
	if (pp<MAX_PP) {
		if (newprod) producer_list[pp].Append(newprod);
		else printf("WARNING: attempt to add NULL Sup to scheduler rejected\n");
	}
}

void	Scheduler::WaitSlice(void)
{
	time_t rem = cnt.Remaining();
	if (rem>0) cnt.Sleep(rem);
	else cout << "WARNING: time slices too small \n";
	cnt.Count(interval);
}

void	Scheduler::Run(void)
{
	cnt.Count(interval);
	while (1) {
		(void)Step();
		WaitSlice();
	}
}

Data	*Scheduler::Step(void)
{
	for (int prio=0;prio<MAX_PP;prio++) {
		// run propensities first
		Steppable *prod_ptr;
		producer_list[prio].Reset();
		while (NULL!=(prod_ptr = (Steppable*) producer_list[prio].Iterate())) {
			(void)prod_ptr->Step();
		}
	}
	
	// do some consistency checking
	// on every pass (switched off after debugging)
	// the ifndef is there just in case
	#ifndef NDEBUG
	assert(fuzz_true.True());		
	assert(fuzz_false.False());
	assert(fuzz_unknown.Unknown());
	assert(fuzz_evidence.Evidence());
	assert(fuzz_in_progress.Get()>0.0);
	assert(chain_exit.False());
	#endif
	
	return NULL;
}

////////////////////////////////////////////////////////////////////////////
//
//           E V E R Y
//

Every::Every(unsigned long interval_,Steppable *action_)
	:interval(interval_),counter(0),last_retval(NULL)
{
	action		=	Getport(action_);
}

Data *Every::Step(void)
{
	// evaluate every "interval" steps, in between just return the 
	// most recently acquired value (!! NOT BUFFERED, just POINTER is passed)
	if (!action) return NULL;
	
	if ((counter++%interval)==0)return (last_retval=action->Step());
	else {		
		if (last_retval) return last_retval;
		else return NULL;
	}
}

void Every::Reset(void)
{
	if (action) action->Reset();
}

void Every::Abort(void)
{
	if (action) action->Abort();
}

Fuzz *Every::IsDone(void)
{
	if (action) return action->IsDone();
	else return &fuzz_true;
}

////////////////////////////////////////////////////////////////////////////
//
//           E V E R Y T I M E
//

EveryTime::EveryTime(unsigned long interval_,Steppable *action_):Every(interval_,action_)
{
	next_time	=	time_base.CurrentTime()-1;
}

Data *EveryTime::Step(void)
{
	// evaluate every "interval" milliseconds, in between just return the 
	// most recently acquired value (!! NOT BUFFERED, just POINTER is passed)
	
	time_t now;
	if (!action) return NULL;
	
	if ((now=time_base.CurrentTime())>=next_time) {
		next_time = now + interval;
		return (last_retval=action->Step());
	} else {		
		if (last_retval) return last_retval;
		else return NULL;
	}
}

////////////////////////////////////////////////////////////////////////////
//
//           S E T Q
//

Data	*Setq::Step(void)
{
	if (var && action) {var->Set(action->Step());return var->Step();}
	if (par && action) {par->Set(action->Step());return par->Step();}
	return NULL;
}

////////////////////////////////////////////////////////////////////////////
//
//           S E T P A R A M
//

Data	*Setparam::Step(void)
{
	done.SetTrue();	
	if (!par) return NULL;
	if (action) par->Set(action);
	return &fuzz_true;
}

////////////////////////////////////////////////////////////////////////////
//
//           W I T H P A R A M
//

Data	*Withparam::Step(void)
{
	if (!par1 || !tree) return NULL;
	if (par1 && act1) par1->Set(act1);
	if (par2 && act2) par2->Set(act2);
	if (par3 && act3) par3->Set(act3);
	if (par4 && act4) par4->Set(act4);
	if (par5 && act5) par5->Set(act5);
	return tree->Step();
}

////////////////////////////////////////////////////////////////////////////
//
//           I F
//

If::If(		Steppable	*cond_, 
			Steppable 	*action_, 
			Steppable 	*else_action_,
			double	threshold_)
{
	done.Set(FUZZEVIDENCE);

	action		=	Getport(action_);
	cond		=	Getport(cond_);
	else_action = 	Getport(else_action_);
	threshold	=	Limit(threshold_,FUZZFALSE,FUZZTRUE);
}

Data *If::Step(void)
{
	if ((!action)||(!cond)) return NULL;
	
	// action is executed if TESTER >= threshold 
	// a test result of NULL, or of the wrong type is == FALSE
	boolean tst_determining;
	if (Test(cond,tst_determining,threshold)) {
		Data *retval = action->Step();
		Fuzz *adone  = action->IsDone();
		if (adone) done.Set(adone->Get());
		return retval;
	} else {
		if (tst_determining) {
			done.Set(FUZZFALSE);
			return NULL;
		} else {
			if (else_action!=NULL)	{
				Data *retval = else_action->Step();
				Fuzz *adone  = else_action->IsDone();
				if (adone) done.Set(adone->Get());
				return retval;
			} else {
				done.Set(FUZZTRUE);
				return NULL;
			}
		}
	}
}

void If::Reset(void)
{
	if (cond) cond->Reset();
	if (action) action->Reset();
	if (else_action) else_action->Reset();
	done.Set(FUZZFALSE);
}

void If::Abort(void)
{
	if (cond) cond->Abort();
	if (action) action->Abort();
	if (else_action) else_action->Abort();
	done.Set(FUZZTRUE);
}


////////////////////////////////////////////////////////////////////////////
//
//           D O I N T E R V A L
//

Data *DoInterval::Step(void)
{
	if (time_base.CurrentTime()>end) done.Set(FUZZTRUE);
	return &done;
}

void DoInterval::Reset(void)
{
	done.Set(FUZZFALSE);
	end = time_base.CurrentTime() + len;
}

void DoInterval::Abort(void)
{ 
	done.Set(FUZZTRUE);
	end = time_base.CurrentTime() - 1;
}

////////////////////////////////////////////////////////////////////////////
//
//           T I M E O U T  
//

Timeout::Timeout(Steppable *step_action_, time_t msecs, Steppable *t_action_)
	:s_action(NULL),interval(msecs),t_action(NULL)
{
	assert(msecs>0);
	s_action=Getport(step_action_);
	t_action=Getport(t_action_);
	last_time=0;
}

Data *TimeoutAbort::Step(void)
{
	if (!s_action) return NULL;
	time_t now = time_base.CurrentTime();
	if (last_time!=0) {						// the first time it's 0
		if (last_time+interval<now) {
			if (t_action) t_action->Abort();
			else s_action->Abort();
		}
	}
	last_time = now;
	return s_action->Step();
}

Data *TimeoutReset::Step(void)
{
	if (!s_action) return NULL;
	time_t now = time_base.CurrentTime();
	if (last_time!=0) {						// the first time it's 0
		if (last_time+interval<now) {
			if (t_action) t_action->Reset();
			else s_action->Reset();
		}
	}
	last_time = now;
	return s_action->Step();
}

Data *TimeinAbort::Step(void)
{
	assert(interval>0 && recover>0);
	if (!s_action) return NULL;
	time_t now = time_base.CurrentTime();

	if (last_time+recover<now || last_time==0) start_time=now;
	if (start_time+interval<now) {
		if (t_action) t_action->Abort();
		else s_action->Abort();
		start_time=now;
	}
	
	last_time = now;
	return s_action->Step();
}

Data *TimeinReset::Step(void)
{
	assert(interval>0 && recover>0);
	if (!s_action) return NULL;
	time_t now = time_base.CurrentTime();

	if (last_time+recover<now || last_time==0) start_time=now;
	if (start_time+interval<now) {
		if (t_action) t_action->Reset();
		else s_action->Reset();
		start_time=now;
	}
	
	last_time = now;
	return s_action->Step();
}


////////////////////////////////////////////////////////////////////////////
//
//           S T I C K Y I F
//

// semantics if a test succeedes, step "action" until it IsDone

StickyIf::StickyIf(	Steppable *cond_,Steppable *action_, 
				Steppable  *else_action_,double	threshold_)
	:cond(NULL),action(NULL),else_action(NULL),threshold(threshold_),sticking(FALSE)
{
	done.Set(FUZZEVIDENCE);
	action		=	Getport(action_);
	cond		=	Getport(cond_);
	else_action = 	Getport(else_action_);
	threshold	=	Limit(threshold_,FUZZFALSE,FUZZTRUE);
}

Data *StickyIf::Step(void)
{
	if ((!action)||(!cond)) return NULL;
	
	if (sticking) {
		Data *retval = action->Step();
		Fuzz *adone  = action->IsDone();
		if (adone) {
			if (adone->True())	{
				sticking = FALSE;
				if (else_action!=NULL) else_action->Reset();
			} else 				sticking = TRUE;
			done.Set(adone->Get());
		} else {
			sticking = FALSE;
			done.Set(FUZZTRUE);
		}
		return retval;
	} else {
		boolean tst_determining;
		if (Test(cond,tst_determining,threshold)) {
			action->Reset();						// reset before starting
			Data *retval = action->Step();
			Fuzz *adone  = action->IsDone();
			if (adone) {
				if (adone->True())	{
					sticking = FALSE;
					if (else_action!=NULL) else_action->Reset();
				} else 				sticking = TRUE;
				done.Set(adone->Get());
			} else {
				sticking = FALSE;
				done.Set(FUZZTRUE);
			}
			return retval;
		} else {
			sticking = FALSE;
			if (tst_determining) {
				done.Set(FUZZFALSE);
				return NULL;
			} else {
				if (else_action!=NULL)	{
					Data *retval = else_action->Step();
					Fuzz *adone  = else_action->IsDone();
					if (adone) done.Set(adone->Get());
					return retval;
				} else {
					done.Set(FUZZTRUE);
					return NULL;
				}
			}
		}
	}
}

void StickyIf::Reset(void)
{
	if (cond) cond->Reset();
	if (action) action->Reset();
	if (else_action) else_action->Reset();
	sticking=FALSE;
	done.Set(FUZZFALSE);
}

void StickyIf::Abort(void)
{
	if (cond) cond->Abort();
	if (action) action->Abort();
	if (else_action) else_action->Abort();
	sticking=FALSE;
	done.Set(FUZZTRUE);
}


////////////////////////////////////////////////////////////////////////////
//
//           T I M E D  I F
//

TimedIf::TimedIf(		
			Steppable	*cond_, 
			Steppable 	*action_, 
			time_t		msecs,
			Steppable 	*else_action_,
			boolean	retrigger_,
			double	threshold_)
			:interval(msecs),retrigger(retrigger_)
{
	done.Set(FUZZEVIDENCE);
	action		=	Getport(action_);
	cond		=	Getport(cond_);
	else_action = 	Getport(else_action_);
	threshold	=	Limit(threshold_,FUZZFALSE,FUZZTRUE);
	last_test	=	TRUE;	// any init will do
	unbounce	=	time_base.CurrentTime()-1;
}

Data *TimedIf::Step(void)
{
	if ((!action)||(!cond)) return NULL;
	
	// a test result of NULL, or of the wrong type is == FALSE
	boolean tst_determining;
	
	time_t now = time_base.CurrentTime();
	
	if (retrigger) {
		// retrigger time interval if condition evaluates to true
		if (Test(cond,tst_determining,threshold)) {
			last_test = TRUE;
			unbounce = now + interval;
			// action->Reset(); do not reset on each positive cond
			Data *retval = action->Step();
			Fuzz *adone  = action->IsDone();
			if (adone) done.Set(adone->Get());
			return retval;
		} else {
			if (now<=unbounce) {
				Data *retval = action->Step();
				Fuzz *adone  = action->IsDone();
				if (adone) done.Set(adone->Get());
				return retval;
			} else {
				if (tst_determining) return NULL;
				else {
					last_test = FALSE;
					if (else_action!=NULL)	{
						Data *retval = else_action->Step();
						Fuzz *adone  = else_action->IsDone();
						if (adone) done.Set(adone->Get());
						return retval;
					} else 					return NULL;
				}
			}
		}
	} else {
		// do not retest the condition while interval is running
		if (now<=unbounce) {
			Data *retval = action->Step();
			Fuzz *adone  = action->IsDone();
			if (adone) done.Set(adone->Get());
			return retval;
		} else {
			if (Test(cond,tst_determining,threshold)) {
				last_test = TRUE;
				unbounce = now + interval;
				// action->Reset(); do not reset on each positive
				Data *retval = action->Step();
				Fuzz *adone  = action->IsDone();
				if (adone) done.Set(adone->Get());
				return retval;
			} else {
				if (tst_determining) return NULL;
				else {
					last_test = FALSE;
					if (else_action!=NULL)	{
						Data *retval = else_action->Step();
						Fuzz *adone  = else_action->IsDone();
						if (adone) done.Set(adone->Get());
						return retval;
					} else 					return NULL;
				}
			}
		}
	}
	
}

void TimedIf::Reset(void)
{
	if (cond) cond->Reset();
	if (action) action->Reset();
	if (else_action) else_action->Reset();
	done.Set(FUZZFALSE);
	unbounce	=	time_base.CurrentTime()-1;
}

void TimedIf::Abort(void)
{ 
	// use test result of last Step to determine which to abort
	if (cond) cond->Abort();
	if (action) action->Abort();
	if (else_action) else_action->Abort();
	done.Set(FUZZTRUE);
	unbounce	=	time_base.CurrentTime()-1;
}

////////////////////////////////////////////////////////////////////////////
//
//           D E B O U N C E
//

Debounce::Debounce(			
			Steppable	*action_,
			time_t		msecs,
			Steppable	*alt_action_)	// time in milliseconds
			:interval(msecs),unbounce(0)
{
	done.Set(FUZZTRUE);
	action		=	Getport(action_);
	alt_action	=	Getport(alt_action_);
	unbounce	=	time_base.CurrentTime()-1;
}

Data *Debounce::Step(void)
{
	if (action)	{
		time_t now = time_base.CurrentTime();
		if (now>unbounce) {
			unbounce = now + interval;
			return (action->Step());
		} else {
			if (alt_action) return alt_action->Step();
			else return NULL;
		}
	} else {
		if (alt_action) return alt_action->Step();
		else return NULL;
	}
}

void Debounce::Reset(void)
{
	unbounce	=	time_base.CurrentTime()-1;
	if (action) action->Reset();
}

void Debounce::Abort(void)
{ 
	unbounce	=	time_base.CurrentTime()-1;
	if (action) action->Abort();
}

Fuzz *Debounce::IsDone(void)
{
	// using last test result to determine done-test	
	if (action) return action->IsDone();
	else return &fuzz_true;
}

////////////////////////////////////////////////////////////////////////////
//
//           R E S E T S T E P
//

ResetStep::ResetStep(Steppable	*action_){action=Getport(action_);};		

Data *ResetStep::Step(void)
{
	if (action) {
		action->Reset();
		return action->Step();
	} else return NULL;
}

void ResetStep::Reset(void)
{
	if (action) action->Reset();
}

void ResetStep::Abort(void)
{ 
	if (action) action->Abort();
}

Fuzz *ResetStep::IsDone(void)
{
	// using last test result to determine done-test	
	if (action) return action->IsDone();
	else return &fuzz_true;
}

////////////////////////////////////////////////////////////////////////////
//
//           A B O R T S T E P
//

AbortStep::AbortStep(Steppable	*action_){action=Getport(action_);};		

Data *AbortStep::Step(void)
{
	if (action) {
		action->Abort();
		return action->Step();
	} else return NULL;
}

void AbortStep::Reset(void)
{
	if (action) action->Reset();
}

void AbortStep::Abort(void)
{ 
	if (action) action->Abort();
}

Fuzz *AbortStep::IsDone(void)
{
	// using last test result to determine done-test	
	if (action) return action->IsDone();
	else return &fuzz_true;
}

////////////////////////////////////////////////////////////////////////////
//
//           R E S E T S T E P
//

IsDoneStep::IsDoneStep(Steppable	*action_){action=Getport(action_);};		

Data *IsDoneStep::Step(void)
{
	return IsDone();
}

void IsDoneStep::Reset(void)
{
	if (action) action->Reset();
}

void IsDoneStep::Abort(void)
{ 
	if (action) action->Abort();
}

Fuzz *IsDoneStep::IsDone(void)
{
	// using last test result to determine done-test	
	if (action) return action->IsDone();
	else return &fuzz_true;
}

////////////////////////////////////////////////////////////////////////////
//
//           R E M O V E  S U P
//

Data *RemoveSup::Step(void)
{
	if (supmgr && action_ptr) 
		if (supmgr->RemoveBlock(*action_ptr)) return &fuzz_true;
	return &fuzz_false;
}

////////////////////////////////////////////////////////////////////////////
//
//           D O C K
//

unsigned int Dock::next_id=0; // common id counter for all docks

Dock::Dock(SupMgr *mgr, char *name, Steppable *new_exit, boolean is_default):achieved_cond(NULL)
{
	id=next_id++;
	done.Set(FUZZTRUE);			// set a default done value
	SetExit(new_exit);			// set the exit
	my_name=NULL;					// set the name
	SetName(name);
	assert(mgr);
	supmgr=mgr;
	if (mgr) mgr->AddDock(this, my_name, is_default);	// report MYSELF to the manager
}

Dock::Dock():achieved_cond(NULL),my_name(NULL)
{
	id=next_id++;
	done.Set(FUZZTRUE);		// set a default done value
	SetExit(NULL);			// set the exit
	SetName(NULL);
	//printf("WARNING: dock created without link to dock manager (%s)\n",my_name);
}

Dock::~Dock()
{
	delete my_name;
}

Data *Dock::Step(void)
{
	boolean determining;
	if (! achieved_cond) {	// if NULL achieved condition -> use IsDone to terminate
		if (action) {
			if (action->IsDone()->True()) supmgr->RemoveBlock(action);
		}
	}
	else if (Test(achieved_cond,determining,FUZZTRUE)) // remove my action from sup mgr
		supmgr->RemoveBlock(action);			// also removes/replaces action of ME !!
	if (action)	return action->Step();
	else return NULL;
}

void Dock::Reset(void)
{
	if (action) action->Reset();
	if (achieved_cond) achieved_cond->Reset();	
}

void Dock::Abort(void)
{ 
	if (action) action->Abort();	
	if (achieved_cond) achieved_cond->Abort();	
}

Fuzz *Dock::IsDone(void)
{
	if (action)	 return action->IsDone();	
	else return &done;
}

void	Dock::SetChain(Steppable *chain)
{
	SetExit(chain);
}

Steppable	*Dock::GetChain(void)
{
	return GetExit();
}



Steppable *Dock::GetExit(void)
{
	return action;
}

void Dock::SetExit(Steppable *new_exit)
{
	action=new_exit;
}

void Dock::ClearExit(void)
{
	SetExit(NULL);
}

void Dock::SetAchievedCond(Steppable *cond)
{
	achieved_cond=cond;
}

void Dock::SetName(char *txt)
{
	word len;
	if (txt!=NULL) {
		if ((len=(word)strlen(txt))<=MAX_NAME_LEN) {
			delete my_name;
			my_name = new char [len+1];
			strcpy(my_name,txt);
		} else {
			delete my_name;
			my_name = new char [12];
			strncpy(my_name,txt,10);
			my_name[10]=my_name[11]=0;
			printf("WARNING: dock name length exceeds max.; string truncated (%s)\n",my_name);
			return;
		}
	} else {
		delete my_name;
		my_name = new char [13];
		sprintf(my_name,"dock_%d\0\0",id);
		//printf("WARNING: created automatic dock name (%s)\n",my_name);
	}
}

boolean Dock::IsName(char *txt)
{
	return (strcmp(txt,my_name)==0);
}

////////////////////////////////////////////////////////////////////////////
//
//           P A R A L L E L
//

Parallel::Parallel(	Steppable *a1, 
					Steppable *a2,
					Steppable *a3,
					Steppable *a4,
					Steppable *a5,
					Steppable *a6,
					Steppable *a7,
					Steppable *a8,
					Steppable *a9)
{
	done.Set(FUZZEVIDENCE);
	Append(a1);
	Append(a2);
	Append(a3);
	Append(a4);
	Append(a5);
	Append(a6);
	Append(a7);
	Append(a8);
	Append(a9);
}

void Parallel::Append(Steppable *action_)
{
	if (action_==NULL) return;	// refuse to append NULL actions
	sup_list.Append(Getport(action_));
}

Data *Parallel::Step(void)
{
	if (sup_list.Size()<=0) return NULL;
	
	// step them all and test for all done
	Steppable	*current=NULL;
	boolean all_done=TRUE;
	sup_list.Reset();
	while (NULL!=(current=(Steppable*)sup_list.Iterate())) {
		current->Step();							// step them all
		all_done = all_done && current->IsDone()->True();	// AND them together
	}
	if (all_done) done.Set(FUZZTRUE);	// done if ALL totally done
	else done.Set(FUZZEVIDENCE);
	
	// cannot use parallel to return a marker, returns NULL, always
	return NULL;
}

void Parallel::Reset(void)
{
	// reset all contained actions
	Steppable	*current;
	sup_list.Reset();
	while (NULL!=(current=(Steppable*)sup_list.Iterate())) { // reset all
		current->Reset();
	}
	done.Set(FUZZEVIDENCE);
}

void Parallel::Abort(void)
{ 
	// abort all contained actions
	Steppable	*current;
	sup_list.Reset();
	while (NULL!=(current=(Steppable*)sup_list.Iterate())) { // abort all
		current->Abort();
	}
	done.Set(FUZZTRUE);
}


////////////////////////////////////////////////////////////////////////////
//
//           S E Q U E N C E
//
//
//	Testers can either be NULL, or evaluate to 
//		TRUE, FALSE, DETERMINING, EVIDENCE, or UNKNOWN (only one of these)
//		so, there are 6 outcomes for a test
//
//		A NULL test is treated as if returning a FALSE value
//

Sequence::Sequence(
			 Steppable *done_action_,	
			 Steppable *restart_tst_,
			 Steppable *kill_tst_,
			 Steppable *action1, Steppable	*tc1,
			 Steppable *action2, Steppable	*tc2, 
			 Steppable *action3, Steppable	*tc3, 
			 Steppable *action4, Steppable	*tc4, 
			 Steppable *action5, Steppable	*tc5, 
			 Steppable *action6, Steppable	*tc6, 
			 Steppable *action7, Steppable	*tc7, 
			 Steppable *action8, Steppable	*tc8, 
			 Steppable *action9, Steppable	*tc9)
{
	current = NULL;
	done.Set(FUZZFALSE);
	done_action = done_action_;
	restart_tst	= restart_tst_;
	kill_tst	= kill_tst_;
	steps		= 0;
	first_done	= FALSE;
	Append(action1,tc1);
	Append(action2,tc2);
	Append(action3,tc3);
	Append(action4,tc4);
	Append(action5,tc5);
	Append(action6,tc6);
	Append(action7,tc7);
	Append(action8,tc8);
	Append(action9,tc9);
}

void Sequence::Append(Steppable *act_, Steppable *tc_)
{
	if ((tc_==NULL)&&(act_!=NULL))
		printf("WARNING: tc was NULL in Sequence::Append (NULL blocks Sequence)\n");
	if ((tc_!=NULL)||(act_!=NULL))
		sup_list.Append(new SubUP(Getport(act_),Getport(tc_)));
}

Data *Sequence::Step(void)
{	
	// NULL test pointers are interpreted to evaluate to FALSE
	boolean tst_determining=FALSE;
	
	// TEST: should sequence be restarted ?
	if ((first_done)&&(sup_list.Size()>0)) {	// test for restart only if first step done
		if (Test(restart_tst,tst_determining)) {
			this->Reset();
			return this->Step();					// RECURSION ! Wow!
		} else 
			if (tst_determining) return NULL;
	}

	if (done.True()) {							// sequence totally done
		if (done_action!=NULL) return done_action->Step();
		else return NULL;
	} else {									// not totally done yet
	
		if (current!=NULL) {					// already initialized
			// TEST: should sequence be aborted/killed ?
			if (Test(kill_tst,tst_determining)) {
				done.Set(FUZZTRUE);
				first_done = TRUE;
				if (done_action!=NULL) {
					done_action->Reset();
					return done_action->Step();
				}
				else return NULL;
			} else {
				if (tst_determining) return NULL;
			}


			if (Test(current->tst,tst_determining)) {
				first_done	=	TRUE;
				current=(SubUP*)sup_list.Iterate();
				if (current!=NULL) {			// more steps, goto next 
					steps++;
					done.Set(((float)steps-FUZZEVIDENCE)/(float)sup_list.Size());
					if (current->act!=NULL) current->act->Reset();
					if (current->tst!=NULL) current->tst->Reset();
					return this->Step();		// RECURSION for next step
				} else {						// no more steps, done
					done.Set(FUZZTRUE);
					if (done_action!=NULL) {
						done_action->Reset();
						return done_action->Step();
					}
					else return NULL;
				}
			} else {
				if (tst_determining) {
					return NULL;
				} else {
					if (current->act!=NULL) return current->act->Step();
					else return NULL;
				}
			}
		} else {								// current is NULL							
			if (sup_list.Size()>0) {			// initialize first step
				sup_list.Reset();
				current=(SubUP*)sup_list.Iterate();
				assert(current!=NULL);
				steps = 1;
				first_done = FALSE;
				done.Set((1.0-FUZZEVIDENCE)/(float)sup_list.Size());
				if (current->act!=NULL) current->act->Reset();
				if (current->tst!=NULL) current->tst->Reset();
				if (kill_tst!=NULL)     kill_tst->Reset();
				if (restart_tst!=NULL)  restart_tst->Reset();
				return this->Step();			// RECURSION for next step
			} else {							// ex. no steps
				first_done = TRUE;
				done.Set(FUZZTRUE);
				if (done_action!=NULL) {
					done_action->Reset();
					return done_action->Step();
				}
				else return NULL;
			}
		}
	}
}

void Sequence::Abort(void)
{ 
	if (!done.True())
		if (done_action!=NULL) done_action->Reset();
	done.Set(FUZZTRUE);
	first_done	=	TRUE;
}

void Sequence::Reset(void)
{
	current = NULL;		// causes to Reset on next step
	done.Set(FUZZFALSE);
	steps		= 0;
	first_done	=	FALSE;
}

////////////////////////////////////////////////////////////////////////////
//
//           P R O G
//
//

Prog::Prog(
			 Steppable *done_action_,	
			 Steppable *action1,
			 Steppable *action2,
			 Steppable *action3,
			 Steppable *action4,
			 Steppable *action5,
			 Steppable *action6,
			 Steppable *action7,
			 Steppable *action8,
			 Steppable *action9,
			 Steppable *action10,
			 Steppable *action11,
			 Steppable *action12,
			 Steppable *action13,
			 Steppable *action14,
			 Steppable *action15,
			 Steppable *action16,
			 Steppable *action17,
			 Steppable *action18,
			 Steppable *action19,
			 Steppable *action20)
{
	current = NULL;
	done.Set(FUZZFALSE);
	done_action = done_action_;
	steps		= 0;
	Append(action1);
	Append(action2);
	Append(action3);
	Append(action4);
	Append(action5);
	Append(action6);
	Append(action7);
	Append(action8);
	Append(action9);
	Append(action10);
	Append(action11);
	Append(action12);
	Append(action13);
	Append(action14);
	Append(action15);
	Append(action16);
	Append(action17);
	Append(action18);
	Append(action19);
	Append(action20);
}

void Prog::Append(Steppable *act_)
{
	if (act_!=NULL)
		sup_list.Append(Getport(act_));
}

Data *Prog::Step(void)
{	
	if (done.True()) {							// sequence totally done
		if (done_action!=NULL) return done_action->Step();
		else return NULL;
	} else {									// not totally done yet
		if (current!=NULL) {					// already initialized
			Data *result=current->Step();
			if (current->IsDone()->True()) {
				steps++;
				current=(Steppable*)sup_list.Iterate();	// goto next step
				if (current!=NULL) {			// more steps, goto next 
					current->Reset();
					done.Set(((float)steps-FUZZEVIDENCE)/(float)sup_list.Size());
				} else {						// no more steps, done
					done.Set(FUZZTRUE);
					if (done_action!=NULL) done_action->Reset();
				}
			}
			return result;
		} else {								// current is NULL							
			if (sup_list.Size()>0) {			// initialize first step
				sup_list.Reset();
				current=(Steppable*)sup_list.Iterate();
				assert(current!=NULL);
				steps = 1;
				current->Reset();
				done.Set(((float)steps-FUZZEVIDENCE)/(float)sup_list.Size());
				return this->Step();			// RECURSION for next step
			} else {							// ex. no steps
				done.Set(FUZZTRUE);
				if (done_action!=NULL) {
					done_action->Reset();
					return done_action->Step();
				}
				else return NULL;
			}
		}
	}
}

void Prog::Abort(void)
{ 
	if (!done.True())
		if (done_action!=NULL) done_action->Reset();
	done.Set(FUZZTRUE);
}

void Prog::Reset(void)
{
	current = NULL;		// causes to Reset on next step
	done.Set(FUZZFALSE);
	steps		= 0;
}

////////////////////////////////////////////////////////////////////////////
//
//           O N C E
//
//

Once::Once(
			 Steppable *done_action_,	
			 Steppable *action1,
			 Steppable *action2,
			 Steppable *action3,
			 Steppable *action4,
			 Steppable *action5,
			 Steppable *action6,
			 Steppable *action7,
			 Steppable *action8,
			 Steppable *action9)
{
	done.Set(FUZZFALSE);
	done_action = done_action_;
	Append(action1);
	Append(action2);
	Append(action3);
	Append(action4);
	Append(action5);
	Append(action6);
	Append(action7);
	Append(action8);
	Append(action9);
}

void Once::Append(Steppable *act_)
{
	if (act_!=NULL)
		sup_list.Append(Getport(act_));
}

Data *Once::Step(void)
{	
	if (done.True()) {							// sequence totally done
		if (done_action!=NULL) return done_action->Step();
		else return NULL;
	} else {				
		if (sup_list.Size()>0) {			// initialize first step
			Steppable 	*current;
			sup_list.Reset();
			while (NULL!=(current=(Steppable*)sup_list.Iterate())) {
				current->Reset();
				(void)current->Step();
			}
		}
		done.Set(FUZZTRUE);
		if (done_action!=NULL) {
			done_action->Reset();
			return done_action->Step();
		}
		else return NULL;
	}
}

void Once::Abort(void)
{ 
	if (!done.True())
		if (done_action!=NULL) done_action->Reset();
	done.Set(FUZZTRUE);
}

void Once::Reset(void)
{
	done.Set(FUZZFALSE);
}

Fuzz *Once::IsDone(void)
{
	if (done_action!=NULL) return done_action->IsDone();
	else return NULL;
}

////////////////////////////////////////////////////////////////////////////
//
//           L O O P
//

Loop::Loop(	Steppable *action, 
			Steppable *cond, 
			Steppable *done_action,
			float  threshold_)
{
	assert(action!=NULL);
	act			= Getport(action);
	done_act	= Getport(done_action);
	tst			= Getport(cond);
	threshold	= threshold_;
	done.Set(FUZZEVIDENCE);
	if ((cond==NULL)&&(done_action!=NULL))
		printf("WARNING: done_action of Loop will never be executed (Loop::Loop)\n");
}

Data *Loop::Step(void)
{
	boolean tst_determining=FALSE;
	
	// test for done
	if (done.True()) {
		if (done_act!=NULL)	return done_act->Step();
		else return NULL;
	}

	// should loop be killed ?
	if (Test(tst,tst_determining,threshold)) {
		done.Set(FUZZTRUE);
		if (done_act!=NULL)	{
			done_act->Reset();
			return done_act->Step();
		}
		else return NULL;			// no else-action -> do nut'n'
	} else {
		if (tst_determining) return NULL;
	}
	
	// step the loop action and recycle when done
	// CAUTION: done test not tested for NULL result
	Data *actret = act->Step();
	if (act->IsDone()->True()) act->Reset();  // reset to create loop
	return actret;
}

void Loop::Reset(void)
{
	act->Reset();
	if (tst!=NULL) tst->Reset();
	done.Set(FUZZEVIDENCE);
}

void Loop::Abort(void)
{ 
	if (!done.True())
		if (done_act!=NULL) done_act->Reset();
	done.Set(FUZZTRUE);
}


//////////////////////////////////////////////////////////////////////
//
//             S W I T C H
//

Switch::Switch(
		float	threshold_,			// winner must pass this
		Steppable  *default_action_,
		Steppable *p1, Steppable *a1,	// candidates
		Steppable *p2, Steppable *a2,			
		Steppable *p3, Steppable *a3,			
		Steppable *p4, Steppable *a4,			
		Steppable *p5, Steppable *a5,			
		Steppable *p6, Steppable *a6,			
		Steppable *p7, Steppable *a7,			
		Steppable *p8, Steppable *a8,			
		Steppable *p9, Steppable *a9)
		:old_tco(NULL)
{
	threshold=Limit(threshold_,FUZZFALSE,FUZZTRUE);
	default_action=default_action_;
	done.Set(FUZZFALSE);
	Append(p1,a1);
	Append(p2,a2);
	Append(p3,a3);
	Append(p4,a4);
	Append(p5,a5);
	Append(p6,a6);
	Append(p7,a7);
	Append(p8,a8);
	Append(p9,a9);
}		

void	Switch::Append(Steppable *p, Steppable *a)
{
	if (p==NULL) return;	// refuse to append NULL propensity tester
	sup_list.Append(new SubUP(Getport(a),Getport(p)));
}

Steppable	*Switch::Selector(void)
{
	assert((sup_list.Size()>0)||(default_action!=NULL));
	SubUP	*current=NULL;
	Data	*tstr=NULL;
	boolean tst_determining=FALSE;
	
	// select winner based on current propensitiy testers
	//    (propensities are updated by scheduler, i.e. not here)
	float		current_max=threshold;	// init best prop so far
	Steppable	*the_chosen_one=default_action;	// init best action so far
	sup_list.Reset();
	
printf("Propensities: %f ",threshold);	
	
	while (NULL!=(current=(SubUP*)sup_list.Iterate())) {
		if (Test(current->tst,tst_determining, FUZZFALSE)) {
			double	p = fuzz_res->Get();	// handle TRUE, FALSE, EVIDENCE, UNKNOWN
printf("%f ",p);			
			assert(p<=FUZZTRUE);
			if (p>current_max) {		// best score ?
				current_max		=	p;	
				the_chosen_one	=	current->act;
			}
		} else {
			if (tst_determining) return NULL;
		}
	}
	
printf("\n");	
	// step the winner, NOOP if no winner and no default
	return the_chosen_one;
}

Data *Switch::Step(void)
{
	Steppable *tco = Selector();
	
	if (tco!=old_tco) if (tco) tco->Reset();
	old_tco=tco;
	
	if (tco!=NULL)	return tco->Step();
	else			return NULL;
}

void Switch::Reset(void)
{
	if (default_action!=NULL) default_action->Reset();
	
	SubUP *current=NULL;
	sup_list.Reset();
	while (NULL!=(current=(SubUP*)sup_list.Iterate())) {
		if (current->act!=NULL) current->act->Reset();
		if (current->tst!=NULL) current->tst->Reset();
	}
}

void Switch::Abort(void)
{
	if (default_action!=NULL) default_action->Abort();
	
	SubUP *current=NULL;
	sup_list.Reset();
	while (NULL!=(current=(SubUP*)sup_list.Iterate())) {
		if (current->act!=NULL) current->act->Abort();
		if (current->tst!=NULL) current->tst->Abort();
	}
}

Fuzz	*Switch::IsDone(void)
{
	// CAUTION: IsDone results are NOT tested for NULL (core dump danger)
	// NULL actions are always done
	
	boolean all_done=TRUE;
	
	if (default_action!=NULL) all_done = all_done && default_action->IsDone()->True();
	
	SubUP *current=NULL;
	sup_list.Reset();
	while (NULL!=(current=(SubUP*)sup_list.Iterate())) {
		if (current->act!=NULL) 
			all_done = all_done && current->act->IsDone()->True();
	}
	
	if (all_done) done.Set(FUZZTRUE);
	else done.Set(FUZZEVIDENCE);
	
	return &done;
}


//////////////////////////////////////////////////////////////////////
//
//             S T A T E   T R A N S I T I O N   G R A P H
//

StateTGraph::StateTGraph(
				Steppable *start_, 
				Steppable *stop_,
				Steppable *a1,
				Steppable *a2,
				Steppable *a3,
				Steppable *a4,
				Steppable *a5,
				Steppable *a6,
				Steppable *a7,
				Steppable *a8,
				Steppable *a9)
{
	next_stateid	=	1;
	next_transid	=	1;

	done.Set(FUZZFALSE);
	current_state	=	NULL;
	
	AppendState(a1);
	AppendState(a2);
	AppendState(a3);
	AppendState(a4);
	AppendState(a5);
	AppendState(a6);
	AppendState(a7);
	AppendState(a8);
	AppendState(a9);
	states.Append(new STGstate(STG_START_STATEID,start_));
	states.Append(new STGstate(STG_STOP_STATEID,stop_));
}		

void	StateTGraph::AppendState(Steppable *news_)
{
	if (news_==NULL) return;		// don't accept NULL actions, use NOOP
	states.Append(new STGstate(next_stateid,Getport(news_)));
	next_stateid++;
}

void	StateTGraph::AppendTrans(Steppable *tst_, state_id from_id, state_id to_id)
{
	if (tst_==NULL) return;			// refuse to append NULL test
	if (from_id==STG_STOP_STATEID) {// refuse to add links FROM stop state
		printf ("WARNING: cannot add link FROM stop state (StateTGraph::AppendTrans)\n",from_id);
		return;
	}
	
	if (from_id==STG_GLOBAL_STATEID) {
		if ((to_id==STG_START_STATEID)||(to_id==STG_STOP_STATEID)||((to_id>0)&&(to_id<next_stateid))) {
			STGstate *tstate = FindState(to_id);
			if (tstate!=NULL) {
				STGtrans *newtrans = new STGtrans(next_transid,Getport(tst_),tstate);
				glob_trans.Append(newtrans);
				next_transid++;
			} else {
				printf ("WARNING: invalid state id (%d) in (StateTGraph::AppendTrans/global)\n",to_id);
			}
		} else {
				printf ("WARNING: invalid state id RANGE in (StateTGraph::AppendTrans/global)\n");
		}
	} else {
		if (((to_id==STG_START_STATEID)||(to_id==STG_STOP_STATEID)||((to_id>0)&&(to_id<next_stateid))) 
			&& ((from_id==STG_START_STATEID)||((from_id>0)&&(from_id<next_stateid)))){
			STGstate *fstate = FindState(from_id);
			STGstate *tstate = FindState(to_id);
			if ((fstate!=NULL)&&(tstate!=NULL)) {
				STGtrans *newtrans = new STGtrans(next_transid,Getport(tst_),tstate);
				loc_trans.Append(newtrans);
				fstate->exits.Append(newtrans);
				next_transid++;
			} else {
				printf ("WARNING: invalid state id (%d) in (StateTGraph::AppendTrans/local)\n",from_id);
			}
		} else {
				printf ("WARNING: invalid state id RANGE in (StateTGraph::AppendTrans/local)\n");
		}
	}
	
	printf("Added transition from %d to %d \n",from_id, to_id);
}

STGstate	*StateTGraph::FindState(state_id id_)
{
	// linear search (maybe a dynamic array would be better, but who cares, its not in the loop
	STGstate *curstate;
	states.Reset();
	while (NULL!=(curstate = (STGstate*)states.Iterate())) {
		printf("Searching state %d \n",curstate->id);
		if (curstate->id==id_) return curstate;
	}
	return NULL;
}

void	StateTGraph::ListStates(void)
{
	STGstate *curstate;
	states.Reset();
	printf("The current states are: ");
	while (NULL!=(curstate = (STGstate*)states.Iterate())) {
		printf("%d, ",curstate->id);
	}
	printf("\n");
}

Data	*StateTGraph::Step(void)
{
	// check for done state
	if (done.True()) {
		if (current_state!=NULL) return current_state->act->Step();
		else return NULL;
	}

	// check for startup condition
	if (current_state==NULL) {
		current_state	=	FindState(STG_START_STATEID);
		if (current_state==NULL) {
			done.Set(FUZZTRUE);
			return NULL;			// there's no start state
		} else {
			done.Set(FUZZEVIDENCE);
			STGtrans *t;
			glob_trans.Reset();
			while (NULL!=(t=(STGtrans*)glob_trans.Iterate()))			// reset global transitions
				t->tst->Reset();
			current_state->exits.Reset();						// reset local transitions
			while (NULL!=(t=(STGtrans*)current_state->exits.Iterate()))
				t->tst->Reset();
			current_state->act->Reset();						// reset action
			return this->Step();					// RECURSIVELY step the action
		}
	} else {
		STGtrans	*tran;
		boolean 	tst_determining;

		// check global transitions
		glob_trans.Reset();
		while (NULL!=(tran=(STGtrans*)glob_trans.Iterate())) {
			if (tran->next==current_state) continue; // reject looping transitions
			if (Test(tran->tst,tst_determining)) {
				return Transit(tran->next);
			} else {
				if (tst_determining) return NULL;
			}
		}
		
		// check local transitions
		current_state->exits.Reset();
		while (NULL!=(tran=(STGtrans*)current_state->exits.Iterate())) {
			if (tran->next==current_state) continue; // reject looping transitions
			if (Test(tran->tst,tst_determining)) {
				return Transit(tran->next);
			} else {
				if (tst_determining) return NULL;
			}
		}
		
		// step the current step if no transitions apply
		return current_state->act->Step();
	}
}

Data	*StateTGraph::Transit(STGstate *next_s)
{
	if (next_s==NULL) {
		Abort();
		return this->Step();
	} else if (next_s->id==STG_STOP_STATEID) {
		Abort();
		return this->Step();
	} else {
		STGtrans *t;
		current_state	= next_s;
		current_state->exits.Reset();						// reset local transitions
		while (NULL!=(t=(STGtrans*)current_state->exits.Iterate()))
			t->tst->Reset();
		current_state->act->Reset();						// reset action
		return this->Step();					// RECURSIVELY step the action
	}
}

void	StateTGraph::Reset(void)
{
	done.Set(FUZZFALSE);
	current_state	=	NULL;
}

void	StateTGraph::Abort(void)
{
	if (!done.True()) {
		done.Set(FUZZTRUE);
		current_state	=	FindState(STG_STOP_STATEID);
		if (current_state!=NULL)
			current_state->act->Reset();
		// no need to reset exits, because there must not be any !
	}
}


//////////////////////////////////////////////////////////////////////////////
//
//    V A R I A B L E
//

Variable::Variable(const char *name_,Data &content_)
	:Data(DT_Variable),content(NULL)
{
	name.Set(name_);
	Set(&content_);
}

Variable::Variable(const char *name_,Data *content_)
	:Data(DT_Variable),content(NULL)
{
	name.Set(name_);
	Set(content_);
}

void Variable::Set(Data *content_)
{
	// this function has some funny semantics ;-/
	if (!content_) return;
	Data *thedata = content_->Step(); // evaluate the "expression"
	if (!thedata) return;
	if (!content) {
		delete content;					// then delete content
		content = thedata->Duplicate();	// make new content from other
	} else if (!content->SetData(thedata)) {		// try to assign, but if types are unequal
		delete content;					// then delete content
		content = thedata->Duplicate();	// make new content from other
	}
}

Data	*Variable::Duplicate(void)
{
	if (content) return content->Duplicate(); // duplicate the content, not the variable
	else return NULL;
}


