/*

DISCLAIMER 

InSitu is provided as is, without any kind of guarantees or warranty. There is 
no documentation. The author does not provide any kind of support to users of 
this source code and is not responsible for any kind of problems or damages 
resulting from its use. 

The source code is the result of a long iterative design process for a research 
prototype and was never intended to be used in any kind of product. In general, 
the code cannot be expected to compile in any straightforward manner on your 
computer system, since the robot on which it was developed, has a custom design. 

*//* Rene Schaad                                                    */
/* Departement of Computer Science, Univ. of Zurich, Switzerland  */
/* schaad@ifi.unizh.ch                                            */

#include "RobotProxy.h"
#include "../include/RufusPhys.h"
#include "Packet.h"
#include <sys/name.h>
#include <sys/kernel.h>
#include <string.h>
#include <assert.h>
#include <math.h>

//---- RobotProxy ---------------------------------------------------------


RobotProxy::RobotProxy(void) 
:timebase(TRUE),log_cnt(0)
{
	if (-1==qnx_name_attach(0,BEHAVIORD_IPCN)){	// not checking for errors !
		printf("Error %d attaching to QNX name\n");
	}
	(void)Connect();
	fgrabber = new FrameGrabber(80,60);

	SetCameraPos(0,0);					// initialize actuators
	SetRobotVelocity(0,0);
	SetSound(NOSOUND);
	free_valid	= FALSE;
	timebase.Start();
}

RobotProxy::~RobotProxy(void)
{
	delete fgrabber;
}


	
boolean RobotProxy::UpdateSensors(void) 
{	
	word dummy;
	p_noop.reply_code	=	rly_send_status;
	word len = p_noop.Marshal(sendbuf);
	
	if (-1==SendLen(robotd_pid,sendbuf,recbuf,		// get data
							len,    MAX_PACKET_LEN,dummy)) {
		if (!Connect()) return FALSE;						// try to reconnect
		if (-1==SendLen(robotd_pid,sendbuf,recbuf,		// get data
								len,    MAX_PACKET_LEN,dummy))
			return(FALSE);									// give up
	}
		
	if ((len = sensor_vals.Unmarshal(recbuf))<1) return FALSE;
	
	return TRUE;			// phew, it worked!
}

void	RobotProxy::UpdateCameraImage(void)	// capture picture and store internally for postprocessing
{
	fgrabber->capture(camera_image,TRUE);	// grab from camera
	camera_image.MakeGray5(camera_image,5); // convert to 5 bit grayscale
}

void	RobotProxy::UpdateFreeSpace(void)	// calc freespace 
{
	free_valid	= FALSE;

	// check for valid tilt angle (upper camera edge must be < horizon)
	// or even better project to about 8 meters in front of robot
	if (sensor_vals.camera_tilt<CAMERA_MIN_TILT) return;
		
	// process images
	// 1. compute variance in window that slides across image -> edges
	// 2. compute distance to first edge from bottom in each column
	// 3. compute distance to obstacles to left and right of projected
	//    path of robot, and in the path of the robot
	variance_image.WindowVariance(camera_image,8,8,90);
	freespace_image.ObstDist(variance_image,14);

	// compute some reusable constants
	float a = CAMERA_V_AP_CONST-sensor_vals.camera_tilt*CAMERA_TILT_UNIT;
	
	// compute robot path parameters from camera tilt angle
	float t = CAMERA_FLOOR_HEIGHT*tan(a+CAMERA_V_APERTURE); // image top distance
	float b = CAMERA_FLOOR_HEIGHT*tan(a); // image bottom distance
	
	// apparent with of robot path in pixels at bottom and at top of image
	path_width_top		= atan(HULL_RADIUS/t) / CAMERA_H_APERTURE * 2 *
						(float)freespace_image.GetImageWidth();
	path_width_bottom	= atan(HULL_RADIUS/b) / CAMERA_H_APERTURE * 2 *
						(float)freespace_image.GetImageWidth();

	// do the freespace computation of the ObstDist image
	FreeSpace f = freespace_image.FreePath(path_width_top,path_width_bottom);
	
	// compute distances to the left and right and in path of robot
	// in meters
	//
	// Formula: 
	// dist = 	height_of_camera * Tan[PI/2-aperture/2-tilt*tilt_unit
	//			+aperture*freespace/image_height]
	float h 	= CAMERA_V_APERTURE/(float)freespace_image.GetImageHeight();
	free_left 	= CAMERA_FLOOR_HEIGHT*tan(a+((float)f.l)*h);
	free_mid	= CAMERA_FLOOR_HEIGHT*tan(a+((float)f.m)*h);
	free_right	= CAMERA_FLOOR_HEIGHT*tan(a+((float)f.r)*h);
	free_valid	= TRUE;
}

void	RobotProxy::UpdateHallAxis(void)	// use Horswill to calc axis of hallway
{
	axis_image.VanishHorswill(camera_image, hall_axis, hall_axis_variance);
}
 
// Get Sensors
boolean	RobotProxy::GetFreeSpace(float &left, float &mid, float &right)
{
	left	= free_left;
	mid 	= free_mid;
	right	= free_right;
	return free_valid;
}

void	RobotProxy::GetHallAxis(int &axis, int &variance)
{
	axis	= hall_axis;
	variance= hall_axis_variance;
}

word 	RobotProxy::GetAmbient(byte no)  // sensors numbered 1..6 (1==front, counter clock wise)
{
	assert((no<=NO_LIGHTS)&&(no>0));
	return sensor_vals.light[no-1]; // internally numbered 0..5
}

word 	RobotProxy::Get50Hz(void)              // detects fast changing light (eg. 50Hz light bulb)
{
	return sensor_vals.light50;
}

float 	RobotProxy::GetBattery(void)              // detects fast changing light (eg. 50Hz light bulb)
{
	return sensor_vals.battery;
}

float 	RobotProxy::GetCompass(void)		  	 // returns heading degree
{
	return sensor_vals.compass;
}

boolean RobotProxy::GetRadar(void)			 // true if motion detected
{
	return sensor_vals.radar;
}

boolean RobotProxy::GetPIR(void)				 // true if motion detected
{
	return sensor_vals.pir;
}

word 	RobotProxy::GetBumpers(void)			 // pattern corresponds to bumper activation
{
	return sensor_vals.bump_stat;
}

void  	RobotProxy::GetRobotVelocity(float  &translation,
						 float	&rotation)  	/* get translational speed (unit: m/s) */
{
	translation = sensor_vals.tspeed;
	rotation	= sensor_vals.aspeed;
}

void  	RobotProxy::GetDRPosition(float &x, float &y, float &a) /* get xy pos, attitude */
{
	x = sensor_vals.pos_x;
	y = sensor_vals.pos_y;
	a = sensor_vals.attitude;
}

float  	RobotProxy::GetDRDistT(void)
{
	return sensor_vals.drt;
}

float  	RobotProxy::GetDRDistA(void)
{
	return sensor_vals.dra;
}

float 	RobotProxy::GetSonar(byte no)      		/* get sonar distance measure (unit: m) */
{
	assert((no<=NO_SONARS)&&(no>0));  // numbered 1..8
	return sensor_vals.sonar[no-1];   // but 0..7 internally
}

void  	RobotProxy::GetCameraPos(float &pan, float &tilt)    /* set servo angle (unit: radians) */
{
	pan = sensor_vals.camera_pan;
	tilt= sensor_vals.camera_tilt;
}

sound_t 	RobotProxy::GetSound(void)
{
	return sensor_vals.soundno;
}

byte 	RobotProxy::GetMode(void)
{
	return sensor_vals.mode;
}

float 	RobotProxy::GetDockPos(void)
{
	return sensor_vals.dock_pos;
}

////////////////////////////////////////////////////////////////////
// Set Actuators

boolean	RobotProxy::SetOperation(CmdPacket &p_cmd)
{
	word dummy;
	p_cmd.reply_code	=	rly_send_nothing;
	word len = p_cmd.Marshal(sendbuf);
	
	if (-1==SendLen(robotd_pid,sendbuf,NULL,len,0,dummy)) {	// send it
		if (!Connect()) return FALSE;			// try to reconnect
		if (-1==SendLen(robotd_pid,sendbuf,NULL,len,0,dummy)) 		// try again
			return FALSE;						// give up
	}
	return TRUE;						// phew, worked!
}

boolean 	RobotProxy::SetSound(sound_t soundno)
{
	p_sound.sound_no=soundno;
	return SetOperation(p_sound);
}

boolean	RobotProxy::SetSpeech(const char *txt) // timing: 2.5ms plus 0.34 per character
{
	p_speak.Store(txt);
	return SetOperation(p_speak); // send null char, too
}

boolean  	RobotProxy::SetRobotVelocity(float  translation,
						 float	rotation,
						 boolean joymode)  /* set speeds (unit: m/s) */
{
	p_speed.t = translation;
	p_speed.a = rotation;
	p_speed.joymode = joymode;
	return SetOperation(p_speed);
}

boolean  	RobotProxy::SetDRPosition(float x_, float y_, float a_) /* set position */
{
	p_drpos.x	=	x_;
	p_drpos.y	=	y_;
	p_drpos.a	=	a_;
	return SetOperation(p_drpos); 
}

boolean  	RobotProxy::SetDRDist(float drt, float dra) /* set position */
{
	p_drd.drt	=	(drt<-0.0001)?sensor_vals.drt:drt;
	p_drd.dra	=	(dra<-0.0001)?sensor_vals.dra:dra;
	return SetOperation(p_drd); 
}

boolean  	RobotProxy::SetCameraPos(float pan, float tilt)    /* set servo angles (unit: radians) */
{
	p_cam.pan	=	pan;
	p_cam.tilt	=	tilt;
	return SetOperation(p_cam); 
}

boolean  	RobotProxy::MoveCameraPos(float dpan, float dtilt)    /* move servo angles (unit: radians) */
{
	p_cam.pan	=	sensor_vals.camera_pan + dpan;
	p_cam.tilt	=	sensor_vals.camera_tilt + dtilt;
	return SetOperation(p_cam); 
}

boolean	RobotProxy::SetCameraImageSize(word width, word height)
{
	fgrabber->SetMVVResolution(width,height);
	return TRUE;
}

boolean	RobotProxy::SetMode(byte mode_)
{
	p_mode.mode	=	mode_;
	return SetOperation(p_mode);
}

boolean	RobotProxy::SetDockPos(float amount_)
{
	if (amount_>1.0) amount_=1.0;
	else if (amount_<0.0) amount_=0.0;
	p_dock.amount	=	amount_;
	return SetOperation(p_dock);
}

boolean	RobotProxy::SetBattery(float level_)
{
	if (level_>1.0) level_=1.0;
	else if (level_<0.0) level_=0.0;
	p_batt.level	=	level_;
	return SetOperation(p_batt);
}

boolean	RobotProxy::SetPID(byte p, byte i, byte d, byte cnt)
{
	p_pid.p		=	p;
	p_pid.i		=	i;
	p_pid.d		=	d;
	p_pid.cnt	=	cnt;
	return SetOperation(p_pid);
}


///////////////////////////////////////////////////////////////		
// Connection management functions

boolean		RobotProxy::Connect(void)
{
	if (-1==(robotd_pid	= qnx_name_locate(0,ROBOTD_IPCN,0,0))) return FALSE;
	else return TRUE;
	
}
