/*

DISCLAIMER 

InSitu is provided as is, without any kind of guarantees or warranty. There is 
no documentation. The author does not provide any kind of support to users of 
this source code and is not responsible for any kind of problems or damages 
resulting from its use. 

The source code is the result of a long iterative design process for a research 
prototype and was never intended to be used in any kind of product. In general, 
the code cannot be expected to compile in any straightforward manner on your 
computer system, since the robot on which it was developed, has a custom design. 

*/

#ifndef _RobotProxy_h_
#define _RobotProxy_h_

/* Rene Schaad                                                    */
/* Departement of Computer Science, Univ. of Zurich, Switzerland  */
/* schaad@ifi.unizh.ch                                            */

#include "utils.h"
#include <sys/types.h>
#include "FrameGrabber.h"
#include "PrgStruct.h"
#include "Packet.h"
#include "CImage.h"

//---- RobotProxy -----------------------------------------------------------

class RobotProxy : public Action {
public:
	RobotProxy();
	~RobotProxy();
	
	virtual Data *Step(void) {
		(void)UpdateSensors(); 
		UpdateCameraImage();
		if (log_cnt<=0) {
			log_cnt=10;
			printf("Pos: %4.2f %4.2f %5.3f %d\n",
				sensor_vals.pos_x,sensor_vals.pos_y,sensor_vals.attitude,timebase.Lap());
		} else log_cnt--;
//		UpdateFreeSpace();
//		UpdateHallAxis();
		return NULL;
	};

	// Retrieve sensor data
	boolean	UpdateSensors(void);
	void	UpdateCameraImage(void);	// capture picture and store internally for postprocessing
	
	// Process sensor data
	void	UpdateFreeSpace(void);		// calc freespace 
	void	UpdateHallAxis(void);		// use Horswill to calc axis of hallway
	
	// Access Sensor Data
	float	GetBattery(void);		// battery state [0..1]
	word 	GetAmbient(byte no);  // sensors numbered 1..6 (1==front, counter clock wise)
	word 	Get50Hz(void);              // detects fast changing light (eg. 50Hz light bulb)
	float 	GetCompass(void);		  	 // returns heading degree
	boolean GetRadar(void);			 // true if motion detected
	boolean GetPIR(void);				 // true if motion detected
	word 	GetBumpers(void);			 // pattern corresponds to bumper activation
	void  	GetRobotVelocity(float  &translation,
							 float	&rotation);  	/* get translational speed (unit: m/s) */
	void  	GetDRPosition(float &x, float &y, float &a); /* get xy pos, attitude */
	float  	GetDRDistT(void); /* get integrated stuff */
	float  	GetDRDistA(void); /* get integrated stuff */
	float 	GetSonar(byte no);      		/* get sonar distance measure (unit: m) */
	void  	GetCameraPos(float &pan, float &tilt);    /* set servo angle (unit: radians) */
	sound_t	GetSound(void);			// which sound is currently played?
	byte	GetMode(void);			// which mode is robotd in?
	float	GetDockPos(void);		// position of docking device (0=retracted, 1=extended)
	boolean	GetFreeSpace(float &left, float &mid, float &right);
	void	GetHallAxis(int &axis, int &variance);
	
	// Access processed sensor data 
	CImage&	GetCameraImage(void){return camera_image;} 		// return stored camera image
	CImage&	GetVarianceImage(void){return variance_image;}	// return image of window variances
	CImage&	GetFreeSpaceImage(void){return freespace_image;}// return image of free space (l,m,r)
	CImage&	GetAxisImage(void){return axis_image;}			// return image of hall axis
		
	// Set Actuators
	boolean	SetBattery(float level_);
	boolean SetSound(sound_t soundno);
	boolean	SetSpeech(const char *txt); // timing: 2.5ms plus 0.34 per character
	boolean SetRobotVelocity(float  translation,
							 float	rotation,
							 boolean joymode=FALSE);  /* set speeds (unit: m/s) */
	boolean SetDRPosition(float x_, float y_, float a_); /* set position */
	boolean SetDRDist(float drt, float dra); /* set integrals */
	boolean SetCameraPos(float pan, float tilt);    /* set servo angles (unit: radians) */
	boolean MoveCameraPos(float dpan, float dtilt);    /* set servo angles (unit: radians) */
	boolean	SetCameraImageSize(word width, word height); // set image size
	boolean	SetMode(byte mode_);
	boolean	SetDockPos(float amount_);
	boolean	SetPID(byte p, byte i, byte d, byte cnt);

private:
	// Connection management
	boolean	Connect(void);		
	boolean	SetOperation(CmdPacket &p_cmd);
	StopWatch timebase;	// real time for logging
	int		log_cnt;
	
	FrameGrabber  	*fgrabber;
	CImage			camera_image;
	CImage			variance_image;
	CImage			axis_image;
	CImage			freespace_image;
	
	StatPacket		sensor_vals;
	pid_t			robotd_pid;
	byte			sendbuf[MAX_PACKET_LEN],
					recbuf[MAX_PACKET_LEN];
					
	// Processed sensor data
	float			free_left,free_mid,free_right;	// from FreeSpace
	boolean			free_valid;	// was last tilt angle valid?
	int				hall_axis, hall_axis_variance;	// from Horswill
	unsigned int	path_width_top,	path_width_bottom; // path with in image
	
	// Buffer for command packets
	BatteryPacket	p_batt;
	SpeechPacket	p_speak;
	PidPacket		p_pid;
	SoundPacket		p_sound;
	SpeedPacket		p_speed;
	DRPosPacket		p_drpos;
	DRDistPacket	p_drd;
	ModePacket		p_mode;
	DockPacket		p_dock;
	CamPacket		p_cam;
	NoopPacket		p_noop;
	CompositePacket p_composite;
};

#endif	/* _RobotProxy_h_ */
