/*

DISCLAIMER 

InSitu is provided as is, without any kind of guarantees or warranty. There is 
no documentation. The author does not provide any kind of support to users of 
this source code and is not responsible for any kind of problems or damages 
resulting from its use. 

The source code is the result of a long iterative design process for a research 
prototype and was never intended to be used in any kind of product. In general, 
the code cannot be expected to compile in any straightforward manner on your 
computer system, since the robot on which it was developed, has a custom design. 

*//* Rene Schaad                                                    */
/* Departement of Computer Science, Univ. of Zurich, Switzerland  */
/* schaad@ifi.unizh.ch                                            */

#include <sys/irqinfo.h>
#include <sys/proxy.h>
#include <signal.h>
#include <unistd.h>
#include <errno.h>
#include <stdlib.h>
#include <i86.h>
#include "Timing.h"

static pid_t Timing::timer_proxy=-1;
static boolean Timing::do_tick=FALSE;

//---- Timing ---------------------------------------------------------


Timing::Timing() 
{
	watchdog_secs 	=  1;
	timer_proxy		= -1;
	c_watchdog_proxy= -1;
} 

Timing::~Timing()
{
	// Dog stuff
	StopDog();
	StopHeartbeat();
	StopCommWatchdog();
}


//////////////////////////////////////////////////////////////////////////////////////////
// software watchdog  (uses "alarm" system call)

void Timing::StartDog(unsigned int secs_)
{
	watchdog_secs = secs_;
	alarm(watchdog_secs);
}

void Timing::ResetDog(void)
{
	alarm(watchdog_secs);
}

void Timing::StopDog(void)
{
	alarm(0);
}

void Timing::RegisterDogFunction(alarm_cb_t alarm_cb)
{
  signal( SIGALRM , alarm_cb); /* catch SIGALARM */
}

/////////////////////////////////////////////////////////////////
// heartbeat is created thru 50 ms timer interrupt
//    timer_create was strangely inaccurate, occasional 1s delays

#pragma off(check_stack);
pid_t far timer_int()
{
  if (Timing::do_tick) return (Timing::timer_proxy);
  else return(0);
}
#pragma on(check_stack);

pid_t Timing::CreateHeartbeat(void)
{
	do_tick=TRUE;
	
	/* attach a proxy for the timer */
	if ( (timer_proxy = qnx_proxy_attach(0,0,0,-1))  == -1 ) {
		cerr << "Timing: unable to attach timer proxy: " << errno << "\n";
		exit(1);
	}
	
    /* attach interrupt handler to 50ms interrupt */
    if ( ( intid = qnx_hint_attach(-1, timer_int, 
		   FP_SEG((void far *) &do_tick))) == -1 ) {
       cerr << "Unable to attach timer interrupt: ";
	   cerr << errno << "\n";
	   exit(1);
	}
	return timer_proxy;
}

void Timing::SuspendHeartbeat(void) 
{
	do_tick=FALSE;
}

void Timing::ResumeHeartbeat(void) 
{
	do_tick=TRUE;
}

void Timing::StopHeartbeat(void)
{
	qnx_hint_detach(intid);
	qnx_proxy_detach(timer_proxy);
}

//////////////////////////////////////////////////////////////////////////////////////////
// Communications timeout watchdog (via a timer proxy)

pid_t Timing::CreateCommWatchdog(long comm_timeout_secs, long comm_timeout_nsecs)
{
	// Sends a proxy every COMM_TIMEOUT seconds
	/* attach a proxy for communications watchdog */
	if ( (c_watchdog_proxy = qnx_proxy_attach(0,0,0,0))  == -1 ) {
		cerr << "Timing::CreateCommWatchdog: unable to attach c_watchdog_proxy proxy: " << errno << "\n";
		exit(1);
	}
  
	// create the timer
	struct sigevent c_event_p;
	c_event_p.sigev_signo = -c_watchdog_proxy;
	if ((comm_watchdog = timer_create(CLOCK_REALTIME,&c_event_p)) == -1) {
		cerr << "Timing::CreateCommWatchdog: could not create comm_watchdog: " << errno << "\n";
		exit(1);
	}
	
	// arm the timer
	c_timer_val_stop.it_value.tv_nsec    = 0;
	c_timer_val_stop.it_interval.tv_nsec = 0;
	c_timer_val_stop.it_value.tv_sec     = 0;
	c_timer_val_stop.it_interval.tv_sec  = 0;
	c_timer_val_start.it_value.tv_nsec    = comm_timeout_nsecs;
	c_timer_val_start.it_interval.tv_nsec = 0; // no repeat
	c_timer_val_start.it_value.tv_sec     = comm_timeout_secs;
	c_timer_val_start.it_interval.tv_sec  = 0;
	if (timer_settime(comm_watchdog,0,&c_timer_val_start,NULL)==-1) {
		cerr << "MCdriver: could not arm timer: " << errno << "\n";
		exit(1);
	}
	return c_watchdog_proxy;
}

void Timing::ResetCommWatchdog(void)
{
	if (timer_settime(comm_watchdog,0,&c_timer_val_start,NULL)==-1) {
		cerr << "robotd: could not retrigger comm watchdog: " << errno << "\n";
		exit(1);
	}		
}

void Timing::StopCommWatchdog(void)
{
	qnx_proxy_detach(c_watchdog_proxy);
	timer_delete(comm_watchdog);
}



